 /*****************************************************************************
 *
 * Copyright 1999-2000 Digi International (www.digi.com)
 *    Zhong Deng <Zhong_Deng@digi.com>
 *    Jeff Randall <Jeff_Randall@digi.com>
 *    Bob Rubendunst <Robert_Rubendunst@digi.com>
 *
 * SCSI code Copyright 2000-2001 Kees Cook <cook@cpoint.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY, EXPRESS OR IMPLIED; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **************************************************************************** */

/* to build cdscsid, define "STS_CODE" during a gcc run, as in the Makefile
	#define STS_CODE
   -Kees 01/19/01
*/

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <sys/termios.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/syslog.h>
#include <linux/serial.h>
#include <netinet/in.h>
#include "ctio.h"
#include "fas.h"
#include "cdutils.h"
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#ifdef STS_CODE
# include <scsi/sg.h>
# define __USE_XOPEN
# define __USE_GNU
# include <stdlib.h>
# include "sts.h"
#endif


#ifdef STS_CODE
# define MY_LONGNAME	"scsiTerminal Server"
# define MY_SHORTNAME	"cdscsid"
# define MY_SOCKNAME    "SCSI device"
#else
# define MY_LONGNAME	"EtherLite"
# define MY_SHORTNAME	"cdetherd"
# define MY_SOCKNAME    "socket"
#endif

#define ELS_TCP_PORT    10001

/*
 * MAXSLEEPTIME is the maximum sleep time in seconds to wait between
 * ELS connection attempts. MAXQUICKTRYS is the number of quick attempts
 * before we slow down requests fo rthe resource.
 */
#define MAXSLEEPTIME 120
#define MAXQUICKTRYS 120
#define MAXDEVNAMELEN 7

/*
 * "where" for log messages.
 * SYSCON - send to system console in addition to log file.
 */
#define SYSCON  0x2
#define NOSTAMP 0x4		/* don't print time & datestamp */

/*
 * DEFINE max path length.
 * the rest of the path is /tmp/ + .older + null byte
 */
#define RESTOFPATH 12

#undef MAX
#define MAX(a, b) (((a) > (b)) ? (a) : (b))

void els_terminate();
void bitmash();
void err_exit();
#ifdef STS_CODE
int make_socket_connection(char * sts_device);
#else
int make_socket_connection();
#endif

extern  char    *optarg;        /* for getopt function */

/* Globals */
char Logfilename[MAXDEVNAMELEN+RESTOFPATH]; 
int log_fd = 0;
int sock_fd = -1;
int els_fd;
long curr_time, prev_time;
/* struct ct_unit unit; */
struct sockaddr_in  cli_addr;
struct timeval sock_time;
char data_buf[4096];
char log_file[40];
els_config els_cfg;
#ifdef STS_CODE
sts_state * sts;
#endif

void els_closelog();
void els_log_err(int where, const char *format, ...);
void els_log_msg(int where, const char *fmt, ...);
void els_log_renamer(char *unit);
void els_terminate();
int sock_read(int sock_fd,unsigned char *buf, int wanted);

req_unit_unlock unlock_req;
rsp_unit_unlock unlock_rsp;
req_unit_inquiry inquiry_req;
rsp_unit_inquiry inquiry_rsp;
unsigned char inquiry_data[128];

/* 
 * deletes the newname file, and renames oldname to newname
 * Returns 0 if all went well, or errno if not.
 */
int name_shuffle(char *oldname, char *newname)
{
	struct stat buf;
	if (!stat(newname, &buf))
		unlink(newname);
	if (!stat(oldname, &buf)) {
		if (!link(oldname, newname)) {
			return unlink(oldname);
		}
	}
	return errno;
}


/*
 * delete (unit).older, if it exits.
 * rename (unit).old to (unit).older
 * rename (unit).log to (unit).old
 * remove (unit).log.
 */
void els_log_renamer(char *unit)
{
	char tmpnam[] = "/tmp/";
	char name2[MAXDEVNAMELEN+RESTOFPATH];
	char name1[MAXDEVNAMELEN+RESTOFPATH];
	struct stat buf;
   
	if (strrchr(unit,'/'))
		unit = strrchr(unit,'/') + 1;
	strcpy(name2,tmpnam);
	strcat(name2, unit);
	strcat(name2, ".older");	/* +7 */
	strcpy(name1,tmpnam);
	strcat(name1, unit);
	strcat(name1, ".old");
	name_shuffle(name1, name2);
	strcpy(name2,tmpnam);
	strcat(name2, unit);
	strcat(name2, ".log");
	name_shuffle(name2, name1);
	strcpy(Logfilename, name2);	/*  store this Global */
	if (!stat(Logfilename, &buf))
		unlink(Logfilename);
}

/* close the log file */
void els_closelog()
{
	if (log_fd)
		close(log_fd);
	log_fd = 0;
}

/*
 * common routine to print the timestamp.
 * Routine 0 if all went well, errno if log file could not
 * be opened.
 */
void print_timestamp()
{
	time_t cur_time;
	struct tm *tp;
	char timout[40];
  
	/*
	 * Print the datestamp to the log file;
	 */
	
	cur_time = time(0);
	tp = localtime(&cur_time);
	sprintf(timout, "%02d/%02d/%02d %02d:%02d:%02d: ",
		tp->tm_mon+1, tp->tm_mday, tp->tm_year%100,
		tp->tm_hour, tp->tm_min, tp->tm_sec);
	
	fprintf(stderr,timout);
}


/* print standard error msg to log file.
 */
void els_log_msg(int where, const char *fmt, ...)
{
	va_list vaptr;

	va_start(vaptr, fmt);
	if (!(where&NOSTAMP))
		print_timestamp();
	vfprintf(stderr,fmt,vaptr);
	va_end(vaptr);
	fflush(stderr);
	/* It may go to the system console/syslog as well. */
	if (where&SYSCON) {
		char buff[120];
		va_start(vaptr, fmt);
		vsprintf(buff,fmt,vaptr);
		va_end(vaptr);
		syslog(LOG_INFO, buff);
	}
}


/*
 * print standard error msg to log file.
 */
void els_log_err(int where, const char *fmt, ...)
{
	va_list vaptr;
	
	va_start(vaptr, fmt);
	if (!(where&NOSTAMP))
		print_timestamp();
	vfprintf(stderr,fmt,vaptr);
	va_end(vaptr);
	fflush(stderr);
	/* It may go to the system console/syslog as well. */
	if (where&SYSCON) {
		char buff[120];
		va_start(vaptr, fmt);
		vsprintf(buff,fmt,vaptr);
		va_end(vaptr);

		/* syslog has datestamp */
		syslog(LOG_ERR, buff);
	}
}


int main(int argc, char **argv)
{
	int n, pid;
	char dev_name[20];
	char log_buf[80];
	int max_fds, nfds;
	fd_set rd_fdset;
	int first_conn = 1;
	unsigned char *ptr;
	int sleeptime = 2;	/* used to prevent syslog from filling up */
	int attempts;
#ifdef DBG
	int i;
#endif

	openlog(MY_SHORTNAME, LOG_PID, LOG_DAEMON);

	els_log_renamer(argc >=2 ? argv[1] : "els");
	/* Remap stderr to log_fd. */
	if ((log_fd = open(Logfilename, O_WRONLY | O_CREAT,
			   (mode_t)0644)) < 0) {
		perror(Logfilename);
		exit(1);
	}
#ifndef STDERR_DEBUG
	dup2(log_fd, STDERR_FILENO);
#endif
	
	if (argc != 3)  {
#ifdef STS_CODE
		printf("Usage: %s els_device sts_device\n", argv[0]);
		els_log_err(SYSCON|NOSTAMP, "Usage: %s els_device sts_device\n",
#else
		printf("Usage: %s els_device ip_addr\n", argv[0]);
		els_log_err(SYSCON|NOSTAMP, "Usage: %s els_device ip_addr\n",
#endif
			    argv[0]);
		els_closelog();
		exit(1);
	}

	signal (SIGHUP, SIG_IGN);
	signal (SIGINT, SIG_IGN);
	signal (SIGQUIT, SIG_IGN);
	signal (SIGTERM, els_terminate);

#ifndef STDERR_DEBUG
	if ((pid = fork()) < 0) {
		els_log_msg(SYSCON, "Cannot fork child process\n");
		els_closelog();
		exit(1);
	}
	else if (pid > 0)
		exit(0);
	
	close(0);
	close(1);
/*      close(2); */
#endif
	
	chdir("/");
	umask(0);
	
	els_log_msg(SYSCON, MY_LONGNAME " daemon running\n");
	sleeptime = 2;
	attempts = 0;
 again:
	sprintf(dev_name, "%s/%s", DEV_PATH, argv[1]);
	
	if((els_fd = open(dev_name, O_RDWR)) < 0)  {
		els_log_err(SYSCON, "open ELS(%s):%s\n", 
			    dev_name, strerror(errno));
		/* if device isn't there, check less frequently for it. */
		sleep(sleeptime);
		if (attempts > MAXQUICKTRYS) {
			if (sleeptime < MAXSLEEPTIME)
				sleeptime += 1;
		}
		else
			attempts += 1;
		goto again;
	}
	sleeptime = 2;		/* reset on success */
	/* can leave this because I'm using h/c/id/lun with dots... ;) -Kees */
	els_cfg.netaddr = inet_addr(argv[2]);

 main_loop:
	attempts = 0;
	sleeptime = 2;
#ifdef STS_CODE
	/* I don't want to loop forever.  If there isn't a scsi device
           then we should shutdown immediately -Kees */
	if (make_socket_connection(argv[2]) < 0) {
		err_exit();
#else
	while (make_socket_connection() < 0) {
		/* loop until connection is made */
		if (attempts > MAXQUICKTRYS) {
			if (sleeptime < MAXSLEEPTIME)
				sleeptime += 2;
		}
		else
			attempts += 1;
		sleep(sleeptime);
#endif
	}
	
	if (first_conn) {
		ptr = inquiry_data + 16;	/* start of inquiry_data */
		strcpy(els_cfg.product_id, 
		       (char *) ptr + strlen((char *) ptr) + 1);
		strcpy(els_cfg.fw_ver, (char *) ptr + strlen((char *) ptr) 
		       + strlen(els_cfg.product_id) + 2);
		sprintf(log_buf, "Vendor ID: %s, Firmware Version: %s\n", 
			ptr, els_cfg.fw_ver);
		els_log_msg(SYSCON, log_buf);
		sprintf(log_buf, "Model Name: %s, %s Address: %s\n", 
			els_cfg.product_id,
#ifdef STS_CODE
			"SCSI",
#else
			"IP",
#endif
			argv[2]);
		els_log_msg(SYSCON|NOSTAMP, log_buf);
		if (ioctl(els_fd, ELSADDUNIT, &els_cfg) < 0) {
			els_log_err(SYSCON, "ELSADDUNIT:%s\n",strerror(errno));
			err_exit();
		}

		first_conn = 0;
	}
	
	els_log_msg(SYSCON, "Unit Online\n");
	if (ioctl(els_fd, ELSONLINE, 0) < 0) {
		els_log_err(SYSCON, "ELSONLINE:%s\n",strerror(errno));
		err_exit();
	}
	
	sock_time.tv_sec = 120;
	sock_time.tv_usec = 0;
	curr_time = time(0);
	max_fds = MAX(sock_fd, els_fd) + 1;
	
	while (1)  {
		FD_ZERO(&rd_fdset);
		FD_SET(sock_fd, &rd_fdset);
		FD_SET(els_fd, &rd_fdset);
		if ((nfds = select(max_fds, &rd_fdset, (fd_set *)0, 
				   (fd_set *)0, &sock_time)) < 0) {
			els_log_err(SYSCON, "select:%s\n", strerror(errno));
			break;
		}
		if (nfds == 0)  {
			els_log_msg(SYSCON, "select timed out\n");
			break;
		}
#ifdef SELECT_DEBUG
		fprintf(stderr,"select finished\n");
#endif

		prev_time = curr_time;
		curr_time = time(0);
		
		if (FD_ISSET(els_fd, &rd_fdset))  {
#ifdef SELECT_DEBUG
			fprintf(stderr,"els_fd readable\n");
#endif
#if 0
			/*
			 * N.B. - Linux's select already adjusts tv.sec for
			 * us, so we don't need to adjust it the way we do
			 * for other O/Ses -rpr
			 */
			sock_time.tv_sec -= (curr_time - prev_time);
#endif
			if ((n = read(els_fd, data_buf, 4096)) > 0)  {
#ifdef DBG
				els_log_msg(SYSCON, "data received "
					    "from els: ");
				ptr = (unsigned char *) data_buf;
				strcpy(log_buf, "    ");
				for (i = 0; i < n; i++) {
					sprintf(log_buf + 4 + (i % 16) * 3,
						"%02x ", *ptr++);
					if ((i % 16) == 15 || i == n - 1) {
						strcat(log_buf,"\n");
						els_log_msg(SYSCON|NOSTAMP,
							    log_buf);
					}
				}
#endif
#ifdef STS_CODE
				if (sts_write(sts, data_buf, n) < 0)  {
#else
				if (write(sock_fd, data_buf, n) < 0)  {
#endif
					els_log_err(SYSCON, 
						    "write on %s:%s\n",
						    MY_SOCKNAME,
						    strerror(errno));
					break;
				}
			} else {
				els_log_err(SYSCON,"read on ELS kernel interface:%s\n",
					    strerror(errno));
				err_exit();
			}
		}
		if (FD_ISSET(sock_fd, &rd_fdset)) {
#ifdef SELECT_DEBUG
			fprintf(stderr,"sock_fd readable\n");
#endif
			sock_time.tv_sec = 120;
#ifdef STS_CODE
			if ((n = sts_read(sts, data_buf, 4096)) > 0) {
#else
			if ((n = read(sock_fd, data_buf, 4096)) > 0) {
#endif
#ifdef DBG
				els_log_msg(SYSCON, "data received from %s: ",
					MY_SOCKNAME);
				ptr = (unsigned char *) data_buf;
				strcpy(log_buf, "    ");
				for (i = 0; i < n; i++) {
					sprintf(log_buf + 4 + (i % 16) * 3,
						"%02x ", *ptr++);
					if ((i % 16) == 15 || i == n - 1)
						els_log_msg(SYSCON|NOSTAMP,
							    log_buf);
				}
#endif
				if (write(els_fd, data_buf, n) < 0)  {
					els_log_err(SYSCON,
						    "write on %s:%s\n",
						    MY_LONGNAME,
						    strerror(errno));
					err_exit();
				}
#ifdef STS_CODE
			} else if (n < 0) {
#else
			} else {
#endif
				els_log_err(SYSCON, "read on %s:%s\n",
					MY_SOCKNAME,strerror(errno));
				break;
			}
		}
	}
#ifdef STS_CODE
	sts=sts_shutdown(sts);
#else
	close(sock_fd);
#endif
	
	els_log_msg(SYSCON, "Unit Offline\n");
	if (ioctl(els_fd, ELSOFFLINE, 0) < 0) {
		els_log_err(SYSCON, "ELSOFFLINE:%s\n",strerror(errno));
		err_exit();
	}
	goto main_loop;
}

#ifdef STS_CODE
int make_socket_connection(char * sts_device) {
	int host, chan, id, lun;
	unsigned char extra_len;
	unsigned char * response;
	unsigned char * ptr;
	char * space;

	if (!sts_device) {
		els_log_err(SYSCON,"make_sock_conn: null sts_device\n");
		return(-1);
	}
	/* FIXME: detect re-initialization calls, and deal with them */
	if (sts) {
		els_log_err(SYSCON,"make_sock_conn: already attached\n");
		return(-1);
	}
	/* alloc with cleared memory */
	if (!(sts=(sts_state *)calloc(1,sizeof(sts_state)))) {
		els_log_err(SYSCON,"calloc: %s\n",strerror(errno));
		return(-1);
	}

	/* process the incoming dot-separated text as host/chan/id/lun */
	if (sscanf(sts_device,"%d.%d.%d.%d",&host,&chan,&id,&lun)!=4) {
		els_log_err(SYSCON,"need SCSI address as 'HOST.CHAN.ID.LUN'\n");
		return(-1);
	}

	/* find the first lun */
	if (!(sts->senddev=scsi_open_id(host,chan,id,lun))) {
		els_log_err(SYSCON,"could not open Send LUN on SCSI host %d "
				"chan %d id %d lun %d!\n",host,chan,id,lun);
		return(-1);
	}
	/* grab the silly extra text */
	response = sts->senddev->fullinq;
	extra_len = response[EXTRA_OFF];
	ptr=response+VENDOR_START+VENDOR_LEN+PRODUCT_LEN+REV_LEN;
	extra_len -= ((ptr-response) - 1);
	if (extra_len > 0)
			memcpy(sts->sendextra, ptr, extra_len);

	/* store regular device info */
	strncpy(sts->vendor,sts->senddev->vendor,VENDOR_LEN);
	strncpy(sts->product,sts->senddev->product,PRODUCT_LEN);
	strncpy(sts->rev,sts->senddev->version,REV_LEN);

	/* report first one */
	els_log_msg(SYSCON,"vendor: [%s]  product: [%s]  "
			"rev: [%s%s%s%s]\n",
			sts->vendor, sts->product,sts->rev,
				*sts->sendextra ? " (" : "",
				sts->sendextra,
				*sts->sendextra ? ")"  : ""
			);
	els_log_msg(SYSCON,"  host: %d chan: %d id: %d lun: %d\n",
			host, chan, id, lun);

	/* check that this is at least a communications device */
	if (sts->senddev->sginfo.scsi_type!=TYPE_COMM) {
		els_log_msg(SYSCON,"Send LUN is not a 'Communication' device!\n");
		scsi_close(sts->senddev);
		sts->senddev=NULL;
		return(-1);
	}
	/* look if medium is loaded */
	if (!scsi_unitready(sts->senddev)) {
			els_log_err(SYSCON,"Send LUN is not ready -- "
				"aborting...\n");
			scsi_close(sts->senddev);
			return(-1);
	}

	/* find the second lun */
	host=sts->senddev->sginfo.host_no;
	chan=sts->senddev->sginfo.channel;
	id=sts->senddev->sginfo.scsi_id;
	lun=sts->senddev->sginfo.lun+1;
	if (!(sts->recvdev=scsi_open_id(host,chan,id,lun))) {
		els_log_err(SYSCON,"could not open Recv LUN on SCSI host %d "
				"chan %d id %d lun %d!\n",host,chan,id,lun);
		scsi_close(sts->senddev);
		return(-1);
	}
	/* grab the silly extra text */
	response = sts->recvdev->fullinq;
	extra_len = response[EXTRA_OFF];
	ptr=response+VENDOR_START+VENDOR_LEN+PRODUCT_LEN+REV_LEN;
	extra_len -= ((ptr-response) - 1);
	if (extra_len > 0)
			memcpy(sts->recvextra, ptr, extra_len);
	/* report second one */
	els_log_msg(SYSCON,"vendor: [%s]  product: [%s]  "
			"rev: [%s%s%s%s]\n",
			sts->vendor, sts->product,sts->rev,
				*sts->recvextra ? " (" : "",
				sts->recvextra,
				*sts->recvextra ? ")"  : ""
			);
	els_log_msg(SYSCON,"  host: %d chan: %d id: %d lun: %d\n",
			host, chan, id, lun);

	/* look if medium is loaded */
	if (!scsi_unitready(sts->recvdev)) {
			els_log_err(SYSCON,"Recv LUN is not ready -- "
				"aborting...\n");
			scsi_close(sts->senddev);
			scsi_close(sts->recvdev);
			return(-1);
	}




	/* FIXME: do any STS inits here, like FAS_GLOBAL */





	/* MIMIC EtherLite */
	ptr=inquiry_data + 16;
	/* store vendor info */
	memcpy(ptr,sts->vendor,VENDOR_LEN);
	/* terminate the scsi text */
	ptr[VENDOR_LEN]='\0';
	/* null terminate at the first space */
	if((space=index(ptr,' '))!=0) *space='\0';
	/* advance to next area */
	ptr+=strlen(ptr)+1;

	/* store product info */
	memcpy(ptr,sts->product,PRODUCT_LEN);
	/* terminate the scsi text */
	ptr[PRODUCT_LEN]='\0';
	/* null terminate at the first space */
	if((space=index(ptr,' '))!=0) *space='\0';
	/* advance to next area */
	ptr+=strlen(ptr)+1;
 
	/* store firmware version info */
	memcpy(ptr,sts->rev,REV_LEN);
	/* terminate the scsi text */
	ptr[REV_LEN]='\0';
	/* null terminate at the first space */
	if((space=index(ptr,' '))!=0) *space='\0';

	/* pretend we have sock_fd */
	sock_fd=sts->recvdev->fd;

	/* start getting info */
	recv_fas_block(sts);
    
	return 1;
}
#else
int make_socket_connection()
{
	int inquiry_len;
	
	if ((sock_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0)  {
		els_log_err(SYSCON, "socket:%s\n", strerror(errno));
		return(-1);
	}


#if 0
	/* options disabled by default */
	{	
		int on = 1;
		setsockopt(sock_fd, SOL_SOCKET, SO_KEEPALIVE, &on, sizeof(on));
# if 0
		setsockopt(sock_fd, SOL_SOCKET, SO_LINGER, &on, sizeof(on));
# endif
	}
#endif

	bzero((char *)&cli_addr, sizeof(cli_addr));
	cli_addr.sin_family = AF_INET;
	cli_addr.sin_addr.s_addr = els_cfg.netaddr;
	cli_addr.sin_port = htons(ELS_TCP_PORT);

	if (connect(sock_fd, (struct sockaddr *) &cli_addr, 
		    sizeof(cli_addr)) < 0) {
		close(sock_fd);
		return(-1);
	}
	unlock_req.opcode = FAS_UNIT_UNLOCK;
	if (write(sock_fd, &unlock_req, FASCMDSZ) < 0) {
		els_log_err(SYSCON, "write FAS_UNIT_UNLOCK:%s\n",
			    strerror(errno));
		close(sock_fd);
		return(-1);
	}
	if ((sock_read(sock_fd, (unsigned char *) &unlock_rsp, FASCMDSZ) <0) ||
	    unlock_rsp.opcode != FAS_UNIT_UNLOCK) {
		els_log_err(SYSCON,"read FAS_UNIT_UNLOCK:%s\n",
			    strerror(errno));
		close(sock_fd);
		return(-1);
	}
	bitmash(unlock_rsp.challenge, inquiry_req.key);
	inquiry_req.opcode = FAS_UNIT_INQUIRY;
	if (write(sock_fd, &inquiry_req, FASCMDSZ) < 0) {
		els_log_err(SYSCON, "write FAS_UNIT_INQUIRY:%s\n",
			    strerror(errno));
		close(sock_fd);
		return(-1);
	}
	if ((sock_read(sock_fd, (unsigned char *)&inquiry_rsp, FASCMDSZ) <0) ||
	    inquiry_rsp.opcode != FAS_UNIT_INQUIRY) {
		els_log_err(SYSCON, "read FAS_UNIT_INQUIRY:%s\n",
			    strerror(errno));
		close(sock_fd);
		return(-1);
	}
	inquiry_len = inquiry_rsp.cnt_lo + (inquiry_rsp.cnt_hi << 8);
	if (sock_read(sock_fd, inquiry_data, inquiry_len) < 0) {
		els_log_err(SYSCON, "read inquiry data:%s\n",strerror(errno));
		close(sock_fd);
		return(-1);
	}
	return(0);
}
#endif


int sock_read(int sock_fd, unsigned char *buf, int wanted)
{
	int n;
	
	while (wanted)  {
		if ((n = read(sock_fd, buf, wanted)) <= 0) {
			return(-1);
		}
		buf += n;
		wanted -= n;
	}
	return(0);
}


/*
 * bitmash function
 *
 *  Accepts an 6 byte challenge and returns a 3 byte key
 *  which has been bitmashed in a moderately inscrutable way.
 *  Prevents inadvertant access of Etherlites.
 */
void bitmash(unsigned char *challenge, unsigned char *key)
{
	int i, j, k, n, churns;
	
	/* mash the bits together */
	for( i=0, j=1, k=2, churns=0; churns < 31;
	     i = (i + 1) % 6, j = (j + i) % 6, k = (k + j) % 6, churns++ ) {
		/* XOR a pair together */
		challenge[i] ^= challenge[j];
		/* rotate one byte each time */
		n = challenge[k] << 1;
		if( n & 0x100 )
			n |= 1;
		challenge[k] = (unsigned char) n;
	}
	/* build a key */
	key[0] = challenge[0] + challenge[4];
	key[1] = challenge[2] + challenge[1];
	key[2] = challenge[3] + challenge[5] + challenge[0];
}


void els_terminate()
{
	err_exit();
}


void els_log(char *msg, int err_flag)
{
	time_t cur_time;
	struct tm *tp;
	char buf[100];
	char err_buf[20];
  
	cur_time = time(0);
	tp = localtime(&cur_time);
	sprintf(buf, "%d/%d/%d %02d:%02d:%02d  %s.", tp->tm_mon+1,
		tp->tm_mday, tp->tm_year, tp->tm_hour, tp->tm_min,
		tp->tm_sec, msg);
  
	if (err_flag) {
		sprintf(err_buf, " ERRNO %d\n", errno);
		strcat(buf, err_buf);
	}
	fprintf(stderr, "%s\n", buf);
}


void err_exit()
{
	els_log_msg(SYSCON, "Daemon process terminated\n");
	els_log_msg(SYSCON, "Unit offline\n");
	if (sock_fd!=-1) {
	    if(ioctl(els_fd, ELSOFFLINE, 0) < 0) 
		els_log_err(SYSCON, "ELSOFFLINE:%s\n",strerror(errno));
#ifdef STS_CODE
	    sts=sts_shutdown(sts);
#else
	    close(sock_fd);
#endif
        }
	close(els_fd);
	close(log_fd);
	exit(1);
}
