/* 
 * ctio.h - ioctl calls for EtherLite(R) and SCSI Terminal Server
 * module drivers
 *
 *****************************************************************************
 *
 * Copyright 1999-2000 Digi International (www.digi.com)
 *    Zhong_Deng <Zhong_Deng@digi.com>
 *    Jeff Randall <Jeff_Randall@digi.com>
 *    Bob Rubendunst <Robert_Rubendunst@digi.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY, EXPRESS OR IMPLIED; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *****************************************************************************/

#ifndef _ctio_h_
#define _ctio_h_


/* private ioctls and defines for extensions, development and debug
 *
 * note that existing digi driver uses args 250-255. We will reserve
 * args 128-143 for Etherlite.
 */
#define STS_TIOC	(('d' << 8))
#define	STSSETEXTCTL	(STS_TIOC | 128)	/* set various tmrs/ctls */
#define	STSGETEXTCTL	(STS_TIOC | 129) 	/* get various tmrs/ctls */
#define	STSFLUSH	(STS_TIOC | 130) 	/* flush a line */
#define	ELSADDUNIT	(STS_TIOC | 131) 	/* add/config ELS unit */
#define	ELSRMUNIT	(STS_TIOC | 132) 	/* remove ELS unit */
#define	ELSGETCFGDATA	(STS_TIOC | 133) 	/* get ELS config data */
#define	ELSONLINE	(STS_TIOC | 134) 	/* els unit online */
#define	ELSOFFLINE	(STS_TIOC | 135) 	/* els unit offline */
#define STSGETTRCBUF	(STS_TIOC | 136)	/* retrieve trace buf */
#define STSCLRTRCBUF	(STS_TIOC | 137)	/* clear trace buf */
#define STSLINEDIAG	(STS_TIOC | 138)	/* retrieve ct_line structure*/
/* STSERRORLEVEL defines (one integer param passed) */
#define STSERR_QUIET	0			/* fewest errors are printed */
#define STSERR_DEFAULT	1			/* normal verbosity */
#define STSERR_VERBOSE	2			/* lots of info */
#define STSSETERRALL	0x80			/* OR'd in with error level */

/* length fields for product ID stuff */
#define	NAMLEN			16
#define	VERLEN			5

/* structure for setting/getting delay values */
typedef
struct stsextctl {
  int x_flags;			/* various ctl flags */
  int local_timeout;		/* local close watchdog (secs) */
  int nonloc_timeout;		/* non-local close watchdog (secs) */
  int dtr_delay;		/* min DTR deassertion (ticks) (unused) */
  int guard_delay;		/* min inter-session (ticks) (unused) */
  int local_waitout;		/* local output wait watchdog (secs) */
  int nonloc_waitout;		/* non-local output wait watchdog (secs) */
  char product_id[NAMLEN];	/* ASCII name */
  char fw_version[VERLEN];	/* firmware rev level */
  u_char scsi_id;		/* target ID on SCSI bus */
  u_char bus_id;		/* bus ID */
  int rsrvd[4];			/* future reserved (zeroed) */
} stsextctl;

/* defs for 'x_flags' field above */
#define	XC_ERRLEV		0x000f	/* error level bits (leave extra) */
#define	XC_LOOP			0x0100	/* flag for loopback */
#define	XC_BAUDEXT		0x00f0	/* mask for baudrate extension factor*/
#define	XC_BAUDSHIFT		4

/* defaults for 'sane' */
#define	XC_DFLT_XFLAGS	STSERR_DEFAULT
#define XC_DFLT_LOCAL_TIMEOUT	0		/* in seconds */
#define	XC_DFLT_NONLOC_TIMEOUT	60		/* in seconds */
#define	XC_DFLT_DTR_DELAY	200		/* in ticks unused  */
#define	XC_DFLT_GUARD_DELAY	20		/* in ticks unused  */
#define	XC_DFLT_LOCAL_WAITOUT	0		/* in seconds */
#define	XC_DFLT_NONLOC_WAITOUT	0		/* in seconds */

/* EtherLite stuff */

typedef struct els_config {
	char		product_id[NAMLEN];
	char		fw_ver[VERLEN];
	unsigned int	netaddr;
} els_config;

#endif /*_ctio_h_*/
