/* 
 * sts_unittab.h - defines for the product capability routines.
 *
 ******************************************************************************
 *
 * Copyright 1999-2000 Digi International (www.digi.com)
 *    Jeff Randall <Jeff_Randall@digi.com>
 *    Zhong Deng <Zhong_Deng@digi.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY, EXPRESS OR IMPLIED; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *****************************************************************************/


#ifndef sts_unit_h			/* prevent multiple defines */
# define	sts_unit_h

#ifndef NULL
# define NULL	0
#endif

/* Note: These should match the defines in fas.h */
#ifndef FAS38400
# define FAS38400		0xc
#endif

#ifndef FAS57600
# define FAS57600		0xd
#endif

#ifndef FAS115200
# define FAS115200		0xf
#endif

#ifndef FAS230400
# define FAS230400		0x22
#endif

#ifndef FAS460800
# define FAS460800		0x2f
#endif

#ifndef FASNOBAUD
# define FASNOBAUD		0xff
#endif


/* unit information record */
typedef
struct ctunitinfo {
	char *name;			/* ASCII product ID */
	int num_lines;			/* maximum lines supported by unit */
	int max_pkt;			/* maximum packet size */
	int max_line_out;		/* maximum output for each line */
	int port_cap_code;		/* port capability personality code */
	int spec_func_code;		/* code for special hardware types */
} ctunitinfo;

/*
 * absolute maximum # of lines at a unit- allow for 2001 w/ 7 pods
 * of 16 serial, 2 parallel
 */
#define	STS_MAX_LINES_PER_UNIT	126


/*
 * Used to define the various port codes used.
 *
 * Port codes are defined as such:
 *
 *	Bits	Function
 *	0-7	FAS code for max speed serial ports can do
 *	8-15	Number of Parallel ports on unit
 *	16-23	Number of Serial ports on unit
 *	24-31	Unique identifier if needed
 */
#define MAKE_PCC_CODE(s_ports, p_ports, speed, identifier)		\
	(((identifier)	<< 24)	|					\
	 ((s_ports)	<< 16)	|					\
	 ((p_ports)	<< 8)	|					\
	 speed)

#define	PCC_0_3_57600		MAKE_PCC_CODE(	0,	3, FAS57600,  0)
#define	PCC_2_0_230400		MAKE_PCC_CODE(	2,	0, FAS230400, 0)
#define	PCC_2_0_460800		MAKE_PCC_CODE(	2,	0, FAS460800, 0)
#define	PCC_2_1_57600		MAKE_PCC_CODE(	2,	1, FAS57600,  0)
#define	PCC_2_1_115200		MAKE_PCC_CODE(	2,	1, FAS115200, 0)
#define	PCC_2_1_230400		MAKE_PCC_CODE(	2,	1, FAS230400, 0)
#define	PCC_2_1_460800		MAKE_PCC_CODE(	2,	1, FAS460800, 0)
#define	PCC_4_0_115200		MAKE_PCC_CODE(	4,	0, FAS115200, 0)
#define	PCC_4_0_230400		MAKE_PCC_CODE(	4,	0, FAS230400, 0)
#define	PCC_4_0_460800		MAKE_PCC_CODE(	4,	0, FAS460800, 0)
#define	PCC_8_0_38400		MAKE_PCC_CODE(  8,	0, FAS38400,  0)
#define	PCC_8_0_57600		MAKE_PCC_CODE(	8,	1, FAS57600,  0)
#define	PCC_8_0_115200		MAKE_PCC_CODE(	8,	0, FAS115200, 0)
#define	PCC_8_0_230400		MAKE_PCC_CODE(	8,	0, FAS230400, 0)
#define	PCC_8_0_460800		MAKE_PCC_CODE(	8,	0, FAS460800, 0)
#define	PCC_8_1_57600		MAKE_PCC_CODE(	8,	1, FAS57600,  0)
#define	PCC_8_1_115200		MAKE_PCC_CODE(	8,	1, FAS115200, 0)
#define	PCC_8_1_230400		MAKE_PCC_CODE(	8,	1, FAS230400, 0)
#define	PCC_8_1_460800		MAKE_PCC_CODE(	8,	1, FAS460800, 0)
#define	PCC_16_0_38400		MAKE_PCC_CODE( 16,	0, FAS38400,  0)
#define	PCC_16_0_57600		MAKE_PCC_CODE( 16,	0, FAS57600,  0)
#define	PCC_16_0_115200		MAKE_PCC_CODE( 16,	0, FAS115200, 0)
#define	PCC_16_0_230400		MAKE_PCC_CODE( 16,	0, FAS230400, 0)
#define	PCC_16_0_460800		MAKE_PCC_CODE( 16,	0, FAS460800, 0)
#define	PCC_32_0_57600		MAKE_PCC_CODE( 32,	0, FAS57600,  0)
#define	PCC_32_0_115200		MAKE_PCC_CODE( 32,	0, FAS115200, 0)
#define	PCC_32_0_230400		MAKE_PCC_CODE( 32,	0, FAS230400, 0)
#define	PCC_32_0_460800		MAKE_PCC_CODE( 32,	0, FAS460800, 0)

#define	PCC_SPEED(code)		(code & 0x000000ff) /* mask off ident part */
#define	PCC_MATCH(code)		(code & 0xffffff00) /* mask off speed part */

#define	PCC_0_3			PCC_MATCH(PCC_0_3_57600)
#define PCC_2_0			PCC_MATCH(PCC_2_0_230400)
#define PCC_2_1			PCC_MATCH(PCC_2_1_115200)
#define PCC_4_0			PCC_MATCH(PCC_4_0_115200)
#define PCC_8_0			PCC_MATCH(PCC_8_0_115200)
#define PCC_8_1			PCC_MATCH(PCC_8_1_115200)
#define PCC_16_0		PCC_MATCH(PCC_16_0_115200)
#define PCC_32_0		PCC_MATCH(PCC_32_0_115200)
#define	PCC_2001		PCC_MATCH(PCC_2001_57600)
#define	PCC_2126		PCC_MATCH(PCC_2126_115200)

#define PCC_MAX_BAUD_EXT(code)	((PCC_SPEED(code) == FAS460800) ? 7 : \
				((PCC_SPEED(code) == FAS230400) ? 5 : \
								  3))




/* special function codes (bit masks) */
/*#define	SFC_STD		0x00*/	/* normal units */
#define	SFC_2000	0x01	/* 2000 series units - Not supported */
#define	SFC_1003	0x02	/* SP-1003 fast parallel unit */
#define	SFC_NO_SPLIT	0x04	/* can't support split baudrates */
#define SFC_ETH		0x08	/* EtherLite products */
#define SFC_PCI		0x10	/* PCI products */
#define SFC_MILSPEC	0x80	/* Special firmware for military apps */
/*
   until the els code admits to knowing what a SCSI unit is, we'll just
   keep on tricking it -Kees 01/19/2000
*/
#define SFC_STD		SFC_ETH



/* port type codes returned by 'ctf_port_type' */

#define PT_SERIAL	0x70000000	/* top bit set for serial */
#define PT_PARALLEL	0x00000000	/* top bit clear for parallel */

#define	PT_38400_SERIAL		(PT_SERIAL | FAS38400)	/* serial to 38.4 */
#define	PT_57600_SERIAL		(PT_SERIAL | FAS57600)	/* serial to 57.6 */
#define	PT_115200_SERIAL	(PT_SERIAL | FAS115200)	/* serial to 115.2 */
#define	PT_230400_SERIAL	(PT_SERIAL | FAS230400)	/* serial to 230.4 */
#define	PT_460800_SERIAL	(PT_SERIAL | FAS460800)	/* serial to 460.8 */
#define	PT_STD_PARALLEL		(PT_PARALLEL | FAS57600)/* Standard parallel */
#define	PT_FAST_PARALLEL	(PT_PARALLEL | FAS115200)/* fast parallel */
#define	PT_NODEV		(-1)

#define PT_SPEED(port)		(port & 0x000000ff)
#define PT_IS_SERIAL(port)	(  (port & PT_SERIAL)  && (port != PT_NODEV))
#define PT_IS_PARALLEL(port)	((!(port & PT_SERIAL)) && (port != PT_NODEV))
#define PT_USES_NEW_SET_PARAMS(port) (PT_SPEED(port) > FAS115200)
#define PT_MAX_BAUD_EXT(code)	((PT_SPEED(code) == FAS460800) ? 7 : \
				((PT_SPEED(code) == FAS230400) ? 5 : \
								 3))


#ifndef OK
# define OK		0
#endif

#ifdef _STDC_
int ctf_port_type( ctunitinfo *uip, int port_num );
ctunitinfo *ctm_get_unitinfo( char *name );
int ctf_check_baud( ctunitinfo *uinfo, int port_num, int fascode );
#else
int ctf_port_type( /* ctunitinfo *uip, int port_num */ );
ctunitinfo *ctm_get_unitinfo( /* char *name */ );
int ctf_check_baud( /* ctunitinfo *uinfo, int port_num, int fascode */ );
#endif

#endif
