/*
 * This is the primary driver file for the Digi International EtherLite(R)
 * Terminal Server for Linux.
 *
 *****************************************************************************
 *
 * Copyright 1999-2000 Digi International (www.digi.com)
 *    Zhong Deng <Zhong_Deng@digi.com>
 *    Jeff Randall <Jeff_Randall@digi.com>
 *    Bob Rubendunst <Robert_Rubendunst@digi.com>
 *    Kees Cook <cook@cpoint.net>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY, EXPRESS OR IMPLIED; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *****************************************************************************/


#include </usr/src/linux-2.4.20-8/include/linux/autoconf.h> /* retrieve the CONFIG_* macros */
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#   define MODVERSIONS /* force it on */
#endif

#ifdef MODVERSIONS
#  include <linux/modversions.h>
#endif
#ifndef LINUX_VERSION_CODE
#include <linux/version.h>
#endif
#include <linux/module.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,2,0)
#  error "This module needs Linux 2.2.0 or newer"
#endif

/* disable additional testing printk debugging messages */
#define printz(fmt, args...)

#include <linux/sched.h>
#ifndef set_current_state
#define set_current_state(a) do { current->state = a; } while (0)
#endif
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/config.h>
#include <linux/slab.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/timer.h>
#include <linux/tty.h>
#include <linux/serial.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/delay.h>
#include <linux/major.h>
#include <linux/mm.h>
#include <linux/tty_flip.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/bitops.h>
#include <asm/uaccess.h>

#include <linux/kernel.h>
#if !defined(KERNEL_VERSION)
# define KERNEL_VERSION(a,b,c) (((((a)<<8)+(b))<<8)+(c))
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
#include </usr/include/linux/spinlock.h>       /* no spinlocks in 2.0 */
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,23)
#include <linux/poll.h>
#endif

#include "fas.h"
#include "ctio.h"
#include "els.h"
/*
 * limit maximum receive message so payload + FAS bytes
 * plus FAS termination character fits in the unit write buffer
 */
#if (CQSIZE-9) < TTY_FLIPBUF_SIZE
#define MAX_RECV_CHUNK (CQSIZE-9)
#else
#define MAX_RECV_CHUNK TTY_FLIPBUF_SIZE
#endif

/* supported products characteristics module */
#include "sts_unittab.h"

#if defined(TRACER)
# include "tracer.c"
/* Tracer call recognizes the following single char format specs:
 *	tilda (~)		print %02x
 *	percent (%)		print %08lx full integer
 *	dollarsign ($)		print string
 *	carat (^)		dump 8 byte FAS command
 */

#endif

/* Module Documentation Stuff */
MODULE_AUTHOR("Zhong Deng / Jeff Randall / Bob Rubendunst / Kees Cook");
MODULE_DESCRIPTION("Digi EtherLite(R) and SCSI Terminal Server TTY driver");

/* Global timeouts -Kees */
LOCAL int unit_wait=60;
#define ELS_UNIT_WAIT unit_wait
MODULE_PARM(unit_wait,"i");
MODULE_PARM_DESC(unit_wait,"Seconds to wait during opens & reads for unit to come online");

LOCAL int unit_gone=60;
#define ELS_UNIT_GONE unit_gone
MODULE_PARM(unit_gone,"i");
MODULE_PARM_DESC(unit_gone,"Seconds until unit offlined from lack of UNIT_POLLs");

/* global unit array */
LOCAL ct_unit		*els_units[MAX_MODULE][MAX_ELS_DEVS];

/* global diag/util line structure */
LOCAL ct_line		*ct_diag;

/* time keeping, etc. */
LOCAL int		 ct_systime = 1000;
LOCAL int		 ct_err_level = STSERR_DEFAULT;

static struct timer_list ct_timer;
LOCAL int kill_timer = 0;	/* cleanup sets to 1 to end timer */

LOCAL struct tty_driver els_callout_driver[MAX_MODULE];
LOCAL struct tty_driver els_serial_driver[MAX_MODULE];
LOCAL char		 serial_name[MAX_MODULE][6];
LOCAL char		 callout_name[MAX_MODULE][6];
LOCAL char		 els_name[MAX_MODULE][6];

LOCAL int		 serial_refcount[MAX_MODULE];
LOCAL struct tty_struct *serial_table[MAX_MODULE][NR_PORTS];
LOCAL struct termios	*serial_termios[MAX_MODULE][NR_PORTS];
LOCAL struct termios	*serial_termios_locked[MAX_MODULE][NR_PORTS];
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
spinlock_t		 ct_pktlock;
#endif


LOCAL struct ct_pkt	*ct_pkt_free;	/* head of ct_pkt free list */
LOCAL int		 ct_pkt_cnt;

LOCAL int global_buy = 0;	/* count # of times we alloc more ct_pkts */
/**************************** function protos ****************************/

/* OS interface routines */
GLOBAL	int	els_init(void);
LOCAL	int	els_open(struct tty_struct *tty, struct file *filp);
LOCAL	void	els_close(struct tty_struct *tty, struct file *filp);
LOCAL	int	els_write(struct tty_struct *tty, int from_user,
			  const u_char *buf, int count);
LOCAL	void	els_put_char(struct tty_struct *tty, unsigned char ch);
LOCAL	void	els_send_xchar(struct tty_struct *tty, char ch);
LOCAL	void	els_flush_chars(struct tty_struct *tty);
LOCAL	int	els_write_room(struct tty_struct *tty);
LOCAL	int	els_ioctl(struct tty_struct *tty, struct file *file,
			  u_int cmd,u_long arg);
LOCAL	void	els_throttle(struct tty_struct *tty);
LOCAL	void	els_unthrottle(struct tty_struct *tty);
LOCAL	void	els_stop(struct tty_struct *tty);
LOCAL	void	els_start(struct tty_struct *tty);
LOCAL	void	els_flush_buffer(struct tty_struct *tty);
LOCAL	void	els_hangup(struct tty_struct *tty);
LOCAL	void	els_set_termios(struct tty_struct *tty, struct termios * old);
LOCAL	void	els_wait_until_sent(struct tty_struct *tty, int timeout);
LOCAL	void	els_break_ctl(struct tty_struct *tty, int state);
LOCAL	int	els_chars_in_buffer(struct tty_struct *tty);

LOCAL	int	unit_read(struct tty_struct *tty, struct file *file, 
			  __u8 *buf, size_t count);
LOCAL	int	unit_write(struct tty_struct *tty, struct file *file, 
			   const __u8 *buf, size_t count);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,23)
LOCAL	int	unit_select(struct tty_struct *tty, struct inode *inode,
                            struct file *filp,int sel_type,select_table *wait);
#else
LOCAL	u_int	unit_poll(struct tty_struct *tty, struct file *filp,
			  poll_table *wait);
#endif

LOCAL	int	unit_ioctl(struct tty_struct *tty, struct file *file,
			   u_int cmd,u_long arg);

/* global ioctl support routine */

/* misc. support routine */
LOCAL	int	ctf_wait_for_carrier(ct_line *lp,  struct tty_struct *tty,
				     struct file *filp);
LOCAL	void	ctf_clear_flow_control(ct_line *lp);
LOCAL	int	dev_to_mod(dev_t major);

/* FAS layer routines */
LOCAL	int	ctf_param(ct_line *ct, struct termios *tiosp, int flag);
LOCAL	void	ctf_handle_recv(ct_line *ct, rsp_recv *rrsp);
LOCAL	void	ctf_handle_send(ct_line *ct);
LOCAL	int	ctf_wait_send(ct_line *ct, int timer);
LOCAL	void	ctf_offline(ct_line *ct);
LOCAL	void	ctf_timer(unsigned long unused);
LOCAL	int	ctf_block_ioctl(ct_line *ct, int flag, int timer);
LOCAL	int	ctf_next_fas(u_char *s, int n);
LOCAL	void	ctf_startin(ct_line *lp);
LOCAL	void	ctf_suspend_output(ct_line *lp);
LOCAL	void	ctf_resume_output(ct_line *lp);
LOCAL	int	ctf_init_line(ct_line *lp);
LOCAL	void	ctf_deinit_line(ct_line *lp);
 
LOCAL	void	ctm_unit_offline(ct_unit *up);
LOCAL	void	ctm_sendfas(ct_unit *up);
LOCAL	void	ctm_reset_circular_queue(circle_q *cq);
LOCAL	int	ctm_circular_queue_space(circle_q *cq);

/* utility routines */
LOCAL	int	get_module(struct tty_struct *tty);
LOCAL	void	bzero(void *ptr, int count);
LOCAL	void	bcopy(void *s, void *d, int count);
LOCAL	int	ct_cmp(u_char *p1, u_char *p2, int cnt);
LOCAL	void	ct_delay(int t, int how);
LOCAL	void   *ct_malloc(int size);
LOCAL	void	ct_free(void *ptr);
LOCAL	void	ctlogerr(u_int err_level, ct_unit *up, ct_line *ct, 
			 char *err_string, int opt_arg);

/*
 * Defines for easy building and unbuilding of FAS commands.
 */
#define	ctm_putfascmd(up)	ctm_putfas((up), FASCMDSZ)
#define	ctm_putfasclrcmd(up)	ctm_putfasclr((up), FASCMDSZ)

#define	ctm_unputfascmd(up)	ctm_putfas((up),-FASCMDSZ)
#define	ctm_getfascmd(up)	ctm_getfas((up), FASCMDSZ)
#define rcvrsp			((rsp_recv *)(cmdp))



/*
 * function protos
 */
/* packet management */
LOCAL	void	ctm_morepkts(int npkts, int size);
LOCAL	void	ctm_lesspkts();
LOCAL	ct_pkt *ctm_getpkt(int size);
LOCAL	void	ctm_freepkt(ct_pkt *pp);

/* MUX layer interface routines */
LOCAL	int	ctm_spaceleft(ct_unit *up);
LOCAL	u_char *ctm_putfas(ct_unit *up, int length);
LOCAL	u_char *ctm_getfas(ct_unit *up, int length);
LOCAL	u_char *ctm_putfasclr(ct_unit *up, int length);

/* FAS support routines */
LOCAL	void	ctf_procrecv(ct_unit *up);
LOCAL	void	ctf_handle_stat_chg(ct_line *ct, rsp_stat_chg *statrsp);
LOCAL	void	ctf_flush_output(ct_line *lp, int flag);
LOCAL	int	ctf_baud(ct_line *lp, struct termios *tiosp);
LOCAL	void	ctf_move_to_queue(u_char *p, circle_q *cqp, int cnt);
LOCAL	void	ctf_move_from_queue(circle_q *cqp, u_char *p, int cnt);
LOCAL	int	ctf_enable_fw(ct_line *ct);
LOCAL	int	ctf_disable_fw(ct_line *ct);
LOCAL	void	ctf_smdm(ct_line *ct, int bits, int cmd, int block);
LOCAL	void	ctf_logfasout(ct_line *ct, u_char *p);
LOCAL	void	ctf_logfasin(ct_line *ct, u_char *p);
LOCAL	void	ctf_showfascmd(char *tag, u_char *p);
LOCAL	void	ctf_display_statistics(char *p, int cnt);
LOCAL	int	ctf_iflush_wait (ct_line *lp);
LOCAL	int	ctm_flush_ioctl( dev_t arg );
LOCAL	void	ctf_suspend_input (ct_line *lp);
LOCAL	void	ctf_resume_input (ct_line *lp);

LOCAL	int	termios_differ(struct termios *a, struct termios *b);
GLOBAL	int	els_read_proc(char *page, char **start, off_t off, int count,
			      int *eof, void *data);


LOCAL	int	illegal_baud(ct_line *lp, struct termios *tiosp);
LOCAL	void	ctm_adjust_lines (ct_unit *up);
LOCAL	void	update_curr_size_fassend (ct_line *lp);
LOCAL	void	update_print_max(ct_line *lp);
LOCAL	int	ctf_int_baud (unsigned bcode);
LOCAL	void	ctm_roundtrip_adjust (ct_unit *up);


#ifdef MODULE
/*------------------------------------------------------------------
    Functions for modloading the driver
------------------------------------------------------------------*/
int init_module (void)
{
	return (els_init());
}


/*
 * cleanup involves shutting down the ctf_timer,
 * unregistering the callout & regular drivers,
 * and reclaiming the storage we used in setting up lines, 
 * units and diag buffers.
 */
void cleanup_module (void)
{
	ct_unit *up;
	ct_line *lp;
	ct_pkt  *pp;
	int i, mod, port, rc;

	kill_timer = 1;
	ct_delay(HZ*2, TASK_UNINTERRUPTIBLE);
	del_timer(&ct_timer);
	if (ct_diag) {
		ct_free(ct_diag);
		ct_diag = NULL;
	}
	for (mod = 0; mod < MAX_MODULE; mod++) {
		if ((rc=tty_unregister_driver(&els_callout_driver[mod])))
			printk("unable to unregister Digi EtherLite "
			       "callout driver: error %d\n", rc);
		if ((rc=tty_unregister_driver(&els_serial_driver[mod])))
			printk("unable to unregister Digi EtherLite "
			       "serial driver: error %d\n", rc);

		for (i = 0; i < MAX_ELS_DEVS; i++) {

			if (!(up = els_units[mod][i]))
				continue;

			for (port = 0; port < up->num_lines; port++) {
				if (!(lp = up->ct_lines[port]))
					continue;
				ct_free(lp);
				lp = NULL;
			}
			ct_free(up->ct_lines);
			
			while ((pp = up->pkt_head)) {
				up->pkt_head = pp->link;
				ctm_freepkt(pp);
			}
			ctm_freepkt(up->recv_ctpkt);
			ctm_lesspkts();
			ct_free(up);
			up = NULL;
		}
	}
	printk("Unloaded %s EtherLite(R) Driver\n", ELS_NAME);
}
#endif /* MODULE */


/*
 * els_init()
 *
 * Module initialization.	
 */
int els_init ()
{
	int	mod, rc;
	
	trc(("els_init unit_wait=~ unit_gone=~\n",unit_wait,unit_gone));
	
	printk("%s EtherLite(R) Driver, Digi International Part Number %s\n",
		   ELS_NAME, ELS_PART);
	printk("scsiTerminal Server extensions active (cook@cpoint.net)\n");

#ifdef TRACER
	printk("(Global Tracing is enabled.)\n");
#endif
	
	/* start slow utility timer */
	init_timer(&ct_timer);
	ct_timer.function = ctf_timer;
	ct_timer.expires = jiffies + HZ * 30;
	add_timer(&ct_timer);
	
	for (mod = 0; mod < MAX_MODULE ; mod++) {
		memset(&els_serial_driver[mod],0,sizeof(struct tty_driver));
		els_serial_driver[mod].magic = TTY_DRIVER_MAGIC;
		sprintf(els_name[mod], "els%2.2x", mod);
		sprintf(serial_name[mod], "ttyN%x", mod);
		els_serial_driver[mod].driver_name = els_name[mod];
		els_serial_driver[mod].name = serial_name[mod];
		els_serial_driver[mod].major = 0;
		els_serial_driver[mod].minor_start = 0;
		els_serial_driver[mod].num = NR_PORTS;
		els_serial_driver[mod].type = TTY_DRIVER_TYPE_SERIAL;
		els_serial_driver[mod].subtype = SERIAL_TYPE_NORMAL;
		els_serial_driver[mod].init_termios = tty_std_termios;
		els_serial_driver[mod].init_termios.c_cflag =
			B9600 | CS8 | CREAD | HUPCL | CLOCAL;
		els_serial_driver[mod].flags = TTY_DRIVER_REAL_RAW;
		els_serial_driver[mod].refcount = &serial_refcount[mod];
		els_serial_driver[mod].table = serial_table[mod];
		els_serial_driver[mod].termios = serial_termios[mod];
		els_serial_driver[mod].termios_locked = 
			serial_termios_locked[mod];
		
		els_serial_driver[mod].open = els_open;
		els_serial_driver[mod].close = els_close;
		els_serial_driver[mod].write = els_write;
		els_serial_driver[mod].put_char = els_put_char;
		els_serial_driver[mod].flush_chars = els_flush_chars;
		els_serial_driver[mod].write_room = els_write_room;
		els_serial_driver[mod].flush_buffer = els_flush_buffer;
		els_serial_driver[mod].ioctl = els_ioctl;
		els_serial_driver[mod].throttle = els_throttle;
		els_serial_driver[mod].unthrottle = els_unthrottle;
		els_serial_driver[mod].set_termios = els_set_termios;
		els_serial_driver[mod].stop = els_stop;
		els_serial_driver[mod].start = els_start;
		els_serial_driver[mod].hangup = els_hangup;
		els_serial_driver[mod].break_ctl = els_break_ctl;
		els_serial_driver[mod].chars_in_buffer = els_chars_in_buffer;
		els_serial_driver[mod].wait_until_sent = els_wait_until_sent;
		els_serial_driver[mod].send_xchar = els_send_xchar;
#if 0
		/* until read_proc works better */
		els_serial_driver[mod].read_proc = els_read_proc;
		els_serial_driver[mod].read_proc = NULL;
		els_serial_driver[mod].proc_entry = NULL;
#endif

		els_callout_driver[mod] = els_serial_driver[mod];
		sprintf(callout_name[mod], "cun%x", mod);
		els_callout_driver[mod].name = callout_name[mod];
		els_callout_driver[mod].major = 0;
		els_callout_driver[mod].subtype = SERIAL_TYPE_CALLOUT;
		els_callout_driver[mod].read_proc = NULL;
		els_callout_driver[mod].proc_entry = NULL;
		
		/** register this driver with the line discipline **/
		
		if ((rc = tty_register_driver(&els_serial_driver[mod])) < 0)
			printk("unable to register Digi EtherLite driver %d\n",
			       rc);
		
		if ((rc = tty_register_driver(&els_callout_driver[mod])) < 0)
			printk("unable to register Digi EtherLite callout "
			       "driver %d\n",rc);
	}
	
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
	spin_lock_init(&ct_pktlock);
#endif
	
	return (0);
}


/*
 * els_open()
 */
LOCAL int els_open (struct tty_struct *tty, struct file *filp)
{
	ct_unit *up;
	ct_line *lp;
	int pun, i;
	int port;
	int dev = MINOR(tty->device);
	int mod;
	int rc;
	int save;
	int isclosing=0;
  
	trc(("els_open % dev=% tty-count ~\n", tty, dev, tty->count ));
  
	if (!tty) {
		printk("null tty ptr in els_open\n");
		return -ENODEV;
	}
    
	/* check for global diag/util open */
	if (dev == ELS_GLOBAL_MINOR) {
		/* alloc line structure (if not already) */
		MOD_INC_USE_COUNT;
		if ((lp = ct_diag) == NULL) {
			ct_diag = lp = ct_malloc(sizeof(ct_line));
			if (lp == NULL) {
				MOD_DEC_USE_COUNT;
				return (-ENOMEM);
			}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
			init_waitqueue_head(&lp->open_wait);
			init_waitqueue_head(&lp->close_wait);
			init_waitqueue_head(&lp->mdmsig_wait);
#endif
		}
		/* diag is unreadable, unwriteable */
		set_bit(TTY_IO_ERROR, &tty->flags);
		lp->fas_histptr = lp->fas_hist;
		return (0);
	}
	
	/* If it's ELS unit, return 0 for root */
	if (ELS_UNIT(dev)) {
		if (suser()) {
			MOD_INC_USE_COUNT;
			return (0);
		}
		return -EPERM;
	}

	/*
	 * determine physical unit and module.
	 * (Up to four modules, each defining up to 7 els units.)
	 */
  
	pun = DEV2UNIT(dev);
	if (pun < 0 || pun >= MAX_ELS_DEVS)
		return (-ENODEV);
	
	mod = get_module(tty);
	if (mod < 0)
		return (-ENODEV);

	/*
	 * Wait up to 1 minute for unit ptr to become valid.  (cdetherd
	 * daemon mallocs unit ptr via ADDUNIT ioctl.  Sometime it's slow at
	 * setting up the units.)  We need to stall getty so it doesn't give
	 * the respawning messages.
	 */

	MOD_INC_USE_COUNT;
	for (i=0; !(up = els_units[mod][pun]) && i < ELS_UNIT_WAIT; i++)
		ct_delay(HZ, TASK_UNINTERRUPTIBLE);
	if (up == (ct_unit *) NULL) {
		trc(("els_open % unit DNE\n", tty));
		/*   MOD_DEC_USE_COUNT; */
		return (-ENODEV);
	}
  
	/*
	 * Wait up to 1 minute for cdetherd daemon to engage the unit interface
	 * before attempting a line open. We need to wait again so that
	 * short network glitches don't adversly affect port usability.
	 */
	for (i = 1; ((up->unit_flags & UNIT_OFFLINE) &&
		     (i <= ELS_UNIT_WAIT)) ; i++) {
		ct_delay(HZ, TASK_UNINTERRUPTIBLE);
	}
	if (up->unit_flags & UNIT_OFFLINE) {
		trc(("els_open % UNIT is OFFLINE\n", tty));
		/*  MOD_DEC_USE_COUNT; */
		return (-ENODEV);
	}
	port = DEV2LINE(dev);
	if (port < 0 || port >= up->num_lines) {
		trc(("els_open % port DNE\n", tty));
		/*  MOD_DEC_USE_COUNT; */
		return (-ENODEV);
	}
	
	if (!(lp = up->ct_lines[port])) {
		if ((lp = (ct_line *) ct_malloc(sizeof(ct_line))) == NULL) {
			/*   MOD_DEC_USE_COUNT; */
			return (-ENOMEM);
		}
    
		up->ct_lines[port] = lp;
		lp->ct_magic = CT_MAGIC;
		lp->up = up;
		lp->port = port;
		lp->error_level = STSERR_DEFAULT;
		lp->fas_histptr = lp->fas_hist;
		lp->nonloc_timeout = XC_DFLT_NONLOC_TIMEOUT;
		lp->local_timeout = XC_DFLT_LOCAL_TIMEOUT;
		lp->local_waitout = XC_DFLT_LOCAL_WAITOUT;
		lp->nonloc_waitout = XC_DFLT_NONLOC_WAITOUT;
		lp->cdbaud_ext = 0;		/* R0.12 cdstty baud extender*/
		lp->cdbaud_delta = 0;		/* R0.12  */
		lp->line_state = LINE_OFFLINE;	/* force initial inits */
		lp->normaltermios = tty_std_termios;
		lp->callouttermios = tty_std_termios;
		/*
		 * lp->flags zeroed is O.K. for Etherlites
		 * use this kluge to find the 9th parallel port on EL-8+s
		 * no ldisc for printer P2
		 */
		if ((up->num_lines == 9) && (port==8)) {
			lp->line_state |= LINE_PRINTER;
			tty->termios->c_iflag = 0;
			tty->termios->c_oflag = 0;
			tty->termios->c_lflag = 0;
		}
		lp->close_delay = (5 * HZ) / 10;
		lp->closing_wait = 30 * HZ;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
		init_waitqueue_head(&lp->open_wait);
		init_waitqueue_head(&lp->close_wait);
		init_waitqueue_head(&lp->mdmsig_wait);
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
		spin_lock_init(&lp->linelock);
#endif
	}
	LINE_LOCK(lp->linelock, save);
	lp->tty = tty;
	tty->driver_data = lp;
	lp->refcount++;
	
	if (up->unit_flags & UNIT_OFFLINE) {
		LINE_UNLOCK(lp->linelock, save);
		/*  MOD_DEC_USE_COUNT; */
		return (-ENODEV);
	}	
	/* make sure line not in offline limbo */
	if ((lp->line_state & (LINE_ENABLED | LINE_OFFLINE)) == 
	    (LINE_ENABLED | LINE_OFFLINE)) {
		LINE_UNLOCK(lp->linelock, save);
		trc(("offline limbo %\n", dev));
		/*  MOD_DEC_USE_COUNT; */
		return (-ENODEV);
	}

	if (tty_hung_up_p(filp) || !lp->tty || (isclosing=(lp->flags & ASYNC_CLOSING))) {
		LINE_UNLOCK(lp->linelock, save);
		if (isclosing)
			interruptible_sleep_on(&lp->close_wait);
		if (lp->flags & ASYNC_HUP_NOTIFY) 
			rc = -EAGAIN;
		else
			rc = -ERESTARTSYS;
		trc(("ASYNC closing\n"));
		goto error;
	}
	if (!(lp->flags & ASYNC_INITIALIZED)) {
		lp->flags |= ASYNC_INITIALIZED;
		clear_bit(TTY_IO_ERROR, &tty->flags);
		ctm_reset_circular_queue(&lp->xmit_q);
		LINE_UNLOCK(lp->linelock, save);
    
		if (lp->flags & ASYNC_CLOSING) {
			interruptible_sleep_on(&lp->close_wait);
			if (lp->flags & ASYNC_HUP_NOTIFY) 
				rc = -EAGAIN;
			else
				rc = -ERESTARTSYS;
			trc(("ASYNC closing\n"));
			goto error;
		}
		
		/* initialize the line */
		rc = ctf_init_line(lp);
		if (rc != OK) {
			trc(("ctf_init_line() failed\n"));
			/* should this be here? -Kees 
			LINE_UNLOCK(lp->linelock, save); */
			goto error;
		}
		
		/* assert DTR and RTS, and set initial paramster */
		ctf_smdm(lp, TIOCM_RTS | TIOCM_DTR, TIOCMBIS, TRUE);
		trc(("calling ctf_param after asserting DTR & RTS tty=%\n",
		     tty->termios));
		ctf_param(lp, tty->termios, FALSE);
		/* should this be here? -Kees 
		LINE_UNLOCK(lp->linelock, save); */
	}
	else
	{
		LINE_UNLOCK(lp->linelock, save);
		if (!lp->line_state  & LINE_ENABLED) {
			/* wait for 1st opener to finish enabling port */
			while (test_and_set_bit(OCSEMA,&lp->ocsema) != 0) {
				interruptible_sleep_on(&lp->open_wait);
				if (signal_pending(current))
					return -ERESTARTSYS;
			}
			if (test_and_clear_bit(OCSEMA,&lp->ocsema) == 0)
				trc(("els_open semaphore already clear\n"));

			if (!lp->line_state  & LINE_ENABLED) {
				if (lp->flags & ASYNC_HUP_NOTIFY) 
					rc = -EAGAIN;
				else
					rc = -ERESTARTSYS;
				trc(("ASYNC closing\n"));
				goto error;
			}
		}
		
	}
    
	
	if (tty->driver.subtype == SERIAL_TYPE_CALLOUT) {
		if (lp->flags & ASYNC_NORMAL_ACTIVE) {
			rc = -EBUSY;
			trc(("callout & ASYNC_NORMAL_ACTIVE"));
			goto error;
		}
		if (lp->flags & ASYNC_CALLOUT_ACTIVE) {
			if ((lp->flags & ASYNC_SESSION_LOCKOUT) &&
			    (lp->session != current->session)) {
				rc = -EBUSY;
				trc(("not current session.\n"));
				goto error;
			}
			if ((lp->flags & ASYNC_PGRP_LOCKOUT) &&
			    (lp->pgrp != current->pgrp)) {
				rc = -EBUSY;
				trc(("not current process group.\n"));
				goto error;
			}
		}
		lp->flags |= ASYNC_CALLOUT_ACTIVE;
	} else {
		if (filp->f_flags & O_NONBLOCK) {
			if (lp->flags & ASYNC_CALLOUT_ACTIVE) {
				rc = -EBUSY;
				trc(("nonblocking & ASYNC_CALLOUT_ACTIVE\n"));
				goto error;
			}
		} else {
			if ((rc = ctf_wait_for_carrier(lp, tty, filp)) != 0) {
				trc(("wait for carrier failed.\n"));
				goto error;
			}
		}
		lp->flags |= ASYNC_NORMAL_ACTIVE;
	}
  
	LINE_LOCK(lp->linelock, save);
	if ((lp->refcount == 1) && (lp->flags & ASYNC_SPLIT_TERMIOS)) {
		if (tty->driver.subtype == SERIAL_TYPE_NORMAL)
			*tty->termios = lp->normaltermios;
		else
			*tty->termios = lp->callouttermios;
		ctf_param(lp, tty->termios, FALSE);
	}
	LINE_UNLOCK(lp->linelock, save);

	lp->session = current->session;
	lp->pgrp = current->pgrp;
  
	ntrc(("els_open % els~ port ~ returns OK\n", tty, lp->up->unum,
	      lp->port));
	return (0);
  
 error:
	ntrc(("els_open % els~ port ~ returns error %\n",  tty,
	      lp->up->unum,lp->port, -rc));
	/*  MOD_DEC_USE_COUNT; */
	return (rc);
}


LOCAL int
ctf_wait_for_carrier (ct_line *lp, struct tty_struct *tty, struct file *filp)
{
	int rc, doclocal, save;
	int extracount = 0;
	
	ntrc(("ctf_wait_for_carrier\n"));
	
	rc = 0;
	doclocal = 0;
	/*
	 * If the device is in the middle of being closed, then block
	 * until it's done, and then try again.
	 */
	if (tty_hung_up_p(filp) ||
	    (lp->flags & ASYNC_CLOSING)) {
		if (lp->flags & ASYNC_CLOSING)
			interruptible_sleep_on(&lp->close_wait);
		return ((lp->flags & ASYNC_HUP_NOTIFY) ?
			-EAGAIN : -ERESTARTSYS);
	}
  
	/*
	 * If non-blocking mode is set, or the port is not enabled,
	 * then make the check up front and then exit.
	 */
	if ((filp->f_flags & O_NONBLOCK) ||
	    (tty->flags & (1 << TTY_IO_ERROR))) {
		if (lp->flags & ASYNC_CALLOUT_ACTIVE)
			return -EBUSY;
		lp->flags |= ASYNC_NORMAL_ACTIVE;
		return 0;
	}

	if (lp->flags & ASYNC_CALLOUT_ACTIVE) {
		if (lp->normaltermios.c_cflag & CLOCAL)
			doclocal++;
	} else {
		LINE_LOCK(lp->linelock, save);
		if (tty && (tty->termios->c_cflag & CLOCAL))
			doclocal++;
		LINE_UNLOCK(lp->linelock, save);
	}
	
	lp->openwaitcnt++;
	/*
	 * tty_hung_up_p(filp) returns TRUE if the current f_ops
	 * are the hung up f_ops.
	 */
	LINE_LOCK(lp->linelock, save);
	if (!tty_hung_up_p(filp)) {
		extracount=1;
		lp->refcount--;
	}
	while (1) {
		if (tty_hung_up_p(filp) ||
		    (lp->flags & ASYNC_INITIALIZED) == 0 ||
		    (lp->flags & LINE_OFFLINE)) {
			if (lp->flags & ASYNC_HUP_NOTIFY)
				rc = -EBUSY;
			else
				rc = -ERESTARTSYS;
			break;
		}
		if (((lp->flags & ASYNC_CALLOUT_ACTIVE) == 0) &&
		    ((lp->flags & ASYNC_CLOSING) == 0) &&
		    (doclocal || (lp->mdmsigs & TIOCM_CAR))) {
			break;
		}
		if (signal_pending(current)) {
			rc = -ERESTARTSYS;
			break;
		}
		LINE_UNLOCK(lp->linelock, save);
		interruptible_sleep_on(&lp->open_wait);
		LINE_LOCK(lp->linelock, save);
	}
	if (extracount)
		lp->refcount++;
	lp->openwaitcnt--;
	if (!rc)
		lp->flags |= ASYNC_NORMAL_ACTIVE;
	LINE_UNLOCK(lp->linelock, save);
	return (rc);
}


/*
 * els_close()
 */
void els_close (struct tty_struct *tty, struct file *filp)
{
	ct_line *lp;
	int save=0;
	int dev = MINOR(tty->device);

	trc(("els_close tty % dev=%\n", tty, dev));
	/* sanity checks */
	if (!tty)
		return;		/* happens if cdetherd not running yet. */
	/* handle global diag/util or ELS unit close */
	if ((dev == ELS_GLOBAL_MINOR) || ELS_UNIT(dev)) { 
		trc(("els_close % returns\n", tty));
		MOD_DEC_USE_COUNT;		/* R0.11 */
		return;
	}

	lp = (ct_line *) tty->driver_data;
	if (lp)
		LINE_LOCK(lp->linelock, save);
	/* tty_hung_up_p() tests to see if filp is the hung up set */
	if (tty_hung_up_p(filp)) {
		if (lp)
			LINE_UNLOCK(lp->linelock, save);
		trc(("els_close done -hung up filp %\n",tty));
		MOD_DEC_USE_COUNT;
		return;
	}

	if (!lp) {
		trc(("null lp in els_close %\n",tty));
		MOD_DEC_USE_COUNT;
		return;
	}

	trc(("els_close els~ port ~\n", lp->up->unum, lp->port));
	if ((tty->count == 1) && (lp->refcount != 1)) {
		printk("els%x port %x forcing refcount to 1, tty->count=%d\n", lp->up->unum, lp->port, tty->count);
		lp->refcount = 1;
	}

	if (lp->refcount-- > 1) {
		LINE_UNLOCK(lp->linelock, save); /* R0.12 */
		trc(("els_close % refcount = %\n", tty, lp->refcount));
		MOD_DEC_USE_COUNT;
		return;
	}
  
	if (!(lp->flags & ASYNC_INITIALIZED)) {
		trc(("els_close: tty ~ ASYNC not INITed\n", tty));
		LINE_UNLOCK(lp->linelock, save); /* R0.12 */
		MOD_DEC_USE_COUNT;
		return;
	}
	
	lp->refcount = 0;
	lp->flags |= ASYNC_CLOSING;

	if (lp->flags & ASYNC_NORMAL_ACTIVE)
		lp->normaltermios = *tty->termios;
	if (lp->flags & ASYNC_CALLOUT_ACTIVE)
		lp->callouttermios = *tty->termios;
  
	tty->closing = 1;
	LINE_UNLOCK(lp->linelock, save);
#if 0
	/*
	 * If transmission was throttled by the application request,
	 * just flush the xmit buffer.
	 */
#if (LINUX_VERSION_CODE >=  KERNEL_VERSION(2,1,15))
	if (tty->flow_stopped) {
		ctf_resume_output(lp);
		tty->flow_stopped = 0;
	}
#endif
#endif /* 0 */
	if (!(lp->line_state&LINE_OFFLINE) &&
	    (lp->closing_wait != ASYNC_CLOSING_WAIT_NONE)) {
		tty_wait_until_sent(tty, lp->closing_wait);
	}
	if (lp->flags & ASYNC_INITIALIZED) {
		ctf_wait_send(lp,0);
	}
	wake_up_interruptible(&lp->mdmsig_wait); /* wake up CMI waiters */

	/* if HUPCL flag is set, drop modem signals */
	if (!(lp->line_state&LINE_OFFLINE) &&
	    (tty->termios->c_cflag & HUPCL)) {
		trc(("els_close calling ctf_smdm\n"));
		ctf_smdm(lp, TIOCM_RTS | TIOCM_DTR, TIOCMBIC, TRUE);
	}
  
	if (tty->driver.flush_buffer)
		tty->driver.flush_buffer(tty);
	if (tty->ldisc.flush_buffer && (lp->line_state&LINE_ENABLED)) {
		(tty->ldisc.flush_buffer)(tty);
	}
  
	/* disable the line */
	if (!(lp->line_state&LINE_OFFLINE)) {
		ctf_deinit_line(lp);
	}
  
	LINE_LOCK(lp->linelock, save);
	lp->flags &= ~ASYNC_INITIALIZED; /* insure the circle buf gets reset */
	set_bit(TTY_IO_ERROR, &tty->flags);
	tty->closing = 0;
	trc(("els_close % els~ port ~d  clearing lp->tty\n",
	     tty, lp->up->unum, lp->port));
	lp->tty = (struct tty_struct *) NULL;
	LINE_UNLOCK(lp->linelock, save);
  
	if (lp->openwaitcnt) {
		if (lp->close_delay)
			 /*  default 0.5 seconds */
			ct_delay(lp->close_delay, TASK_INTERRUPTIBLE);
		wake_up_interruptible(&lp->open_wait);
	}
  
	LINE_LOCK(lp->linelock, save);
	lp->flags &= ~(ASYNC_CALLOUT_ACTIVE | ASYNC_NORMAL_ACTIVE |
		       ASYNC_CLOSING);
	wake_up_interruptible(&lp->close_wait);
	LINE_UNLOCK(lp->linelock, save);
	trc(("els_close done % els~ port ~d  line_state=%\n",
	     tty, lp->up->unum, lp->port, lp->line_state));
	MOD_DEC_USE_COUNT;
}


/*
 * els_write()
 */
LOCAL int els_write (struct tty_struct *tty, int from_user,
		     const u_char *buf, int count)
{
	req_send *sendcmd;
	ct_line *lp;
	ct_unit *up;
	int rc, total, count1, save;
	circle_q *cq;

	/* sanity checks */
	if (!tty || (tty->magic != TTY_MAGIC) || 
	    !(lp = (ct_line *) tty->driver_data))
		return -EIO;
  
	trc(("els_write: els~ port ~, state %, % bytes\n", 
	     lp->up->unum,lp->port, lp->line_state, count));
	if (!(cq = &lp->xmit_q)) {
		trc(("els_write, els~ port ~: null cq ptr, rc=-EIO\n",
		     lp->up->unum, lp->port));
		return -EIO;
	}

	if (((lp->line_state & (LINE_OFFLINE | LINE_ENABLED | LINE_DISABLING)) 
	     != LINE_ENABLED) || !tty || (tty->flags & (1<<TTY_IO_ERROR))) {
		return (-EIO);
}
  
	if (from_user) {
		if ((rc = verify_area(VERIFY_READ, (void *) buf, count)))
			return (rc);
	}
  
	up = lp->up;
	ELS_LOCK(up->faslock, save);
	if (lp->line_state & (LINE_OBUSY | LINE_OSTOPPED)) {
		total = 0;
    
		if (count) {
			total = ctm_circular_queue_space(cq);
			if (total > count)
				total = count;
      
			ctf_move_to_queue((u_char *) buf, cq, total);
		}
		ntrc(("els_write els~ port ~ buffer returned %\n",
		      lp->up->unum, lp->port, total));    
		ELS_UNLOCK(up->faslock, save);
		return (total);
	}

	/*
	 * limits - space left in current outgoing packet,
	 * current outgoing size adjusted by rtt, room left in
	 * circular buffer.
	 */
	total = ctm_spaceleft(up); /* how much room we got? */
	if (total > lp->curr_size_fassend)
		total = lp->curr_size_fassend;
	if (total > cq->cnt + count)
		total = cq->cnt + count;
  
	if (!total) {
		ELS_UNLOCK(up->faslock, save);
		return (0);
	}
  
	lp->line_state |= LINE_OBUSY;
	lp->icount.tx += total;
	trc(("Sending FAS_SEND count=%\n", total));
  
	/* fill out and send FAS_SEND message */
	sendcmd = (req_send *) ctm_putfasclrcmd(up);
	sendcmd->opcode = FAS_SEND;
	sendcmd->line = lp->port;
	sendcmd->cnt_lo = (char) total;
	sendcmd->cnt_hi = (u_char) (total >> 8);
  
	if (cq->cnt) {
		count1 = min(total, cq->cnt);
		ctf_move_from_queue(cq, ctm_putfas(up, count1), count1);
		total -= count1;
	}
  
	if (total) {
		if (from_user)
			COPY_FROM_USER(ctm_putfas(up, total), buf, total);
		else
			bcopy((void *) buf, ctm_putfas(up, total), total);
	}
  
	/*
	 * We were asked to write *count* bytes of data, but we only
	 * managed to write *total* bytes. Copy the remaining to the
	 * xmit_q if possible
	 */
	if (count > total) {
		count1 = ctm_circular_queue_space(cq);
		if (count1 > count - total)
			count1 = count - total;

		ctf_move_to_queue(((u_char *) buf) + total, cq, count1);
		total += count1;
	}
  
	ctf_logfasout(lp, (u_char *) sendcmd);
	ELS_UNLOCK(up->faslock, save);
  
	ctm_sendfas(up);
	
	/*
	 * note that the return value includes chars in xmit_q that
	 * will be sent out later.
	 */
	trc(("els_write els~ port ~ returned %\n", lp->up->unum,
	     lp->port, total));
	return (total);
}


LOCAL int unit_read (struct tty_struct *tty, struct file *file, 
		     __u8 *buf, size_t count)
{
	int		 count1, save, rc, mod;
	ct_unit		*up;
	circle_q	*cq;

	if (!tty) {
		printz("null tty ptr in unit_read()\n");
		return -EIO;
	}
	mod = get_module(tty);
	if (mod < 0)
		return (-ENODEV); 

	/* get unit pointer */
	up = els_units[mod][DEV2UNIT(MINOR(tty->device))];
	if (!up || (up->unit_flags & UNIT_OFFLINE))
		return (-ENODEV);

	if ((rc = verify_area(VERIFY_WRITE, (void *) buf, count)))
		return (rc);
	
	ttrc(("unit_read\n"));

	ELS_LOCK(up->faslock, save);

	/* if no data available for read, return EAGAIN */
	cq = &up->netdata_out;
	if (cq->cnt == 0) {
		ELS_UNLOCK(up->faslock, save);
		return (-EAGAIN);
	}

	/* get the length of data to copy */
	ttrc(("cq->cnt=%, count=%\n", cq->cnt, count));
	if (count > cq->cnt)
		count = cq->cnt;
  
	cq->cnt -= count;

	/* one continuous chunk of buffer of length *count* ? */
	count1 = cq->end - cq->ep;
	if (count1 >= count) {
		if (COPY_TO_USER(buf, cq->ep, count) > 0) {
			ELS_UNLOCK(up->faslock, save);
			return (-EIO);
		}

		cq->ep += count;
	} else {
		/* no, two chunks */
		ttrc(("copy two chunks to user space\n"));
		if (COPY_TO_USER(buf, cq->ep, count1) > 0) {
			ELS_UNLOCK(up->faslock, save);
			return (-EIO);
		}

		if (COPY_TO_USER(buf + count1, cq->d, count - count1) > 0) {
			ELS_UNLOCK(up->faslock, save);
			return (-EIO);
		}

		cq->ep = cq->d + count - count1;
	}
  
	ELS_UNLOCK(up->faslock, save);

	/* 
	 * When we called ctm_sendfas last time, we might have full
	 * or nearly full circular queue to prevent us from copying
	 * data from ctpkt to the queue. Now the queue is emptied a
	 * bit, and maybe we can have enough room for the data in ct_pkt.
	 * So we call ctm_sendfas again.
	 */

	ctm_sendfas(up);
  
	/* a little optimization */
	if (cq->cnt == 0)
		cq->ep = cq->fp = cq->d;

	return (count);
}


/*
 * unit_write() gets data written by the cdetherd daemon,
 * which is data to be read back to the driver. The data
 * is structured into fas messages. Each message is extracted
 * from the buffer and processed. Some of the messages are
 * input data from the els device(s), some telegraph the state
 * of the els lines and unit(s) back to this driver. 
 */
LOCAL int unit_write (struct tty_struct *tty, struct file *file, 
		      const __u8 *buf, size_t count)
{
	ct_unit *up;
	circle_q *cq;
	u_char *ptr;
	int mod, rc, moved;
	int freshdata = 0;

	if (!tty) {
		printk("null tty ptr in unit_write()\n");
		return -ENODEV;
	}
	mod = get_module(tty);
	if (mod < 0)
		return (-ENODEV); 
  
	/* get unit pointer */
	up = els_units[mod][DEV2UNIT(MINOR(tty->device))];
	if (!up || (up->unit_flags & UNIT_OFFLINE))
		return (-ENODEV);
  
	if ((rc = verify_area(VERIFY_READ, (void *) buf, count)))
		return (rc);

	ttrc(("unit_write\n"));

	/* 
	 * The circular queue data is used as a flat, unwarped buffer
	 * to simplify use of the ctf_next_fas() function, which prevents
	 * delivering partial packets to the fas layer. 
	 * At the conclusion of each unit_write, any remaining data from
	 * the middle of the buffer is shuffled down to the start for next
	 * time.
	 */
  
	cq = &up->netdata_in;
	if (count > (CQSIZE - cq->cnt)) {
		count = CQSIZE - cq->cnt;
	}
	if (COPY_FROM_USER(cq->fp, buf, count) > 0)
		return (-EIO);
	cq->cnt += count;
	cq->fp += count;
  
	/* 
	 * Copy data from receiving buffer to receiving ct_pkt.
	 * Need to make sure that no partial FAS command is copied
	 * to up->recv_ctpkt. Updating ct_pkt ptrs is only OK so long
	 * as procrecv() is only called from here and always consumes
	 * all packets in recv_pkt.
	 */
	ptr = up->recv_ctpkt->dp = up->recv_ctpkt->d;
	while ((moved = ctf_next_fas(cq->ep, cq->cnt))) {
		ntrc(("get fas cmd: ~ of length %\n", *(cq->ep), moved));
		bcopy((caddr_t) cq->ep, (caddr_t) ptr, moved);
		freshdata = TRUE;
		ptr += moved;
		cq->ep += moved;
		cq->cnt -= moved;
	}
	/* shuffle any partial fas packet to start of buffer. */
	if (cq->cnt && freshdata) {
		bcopy((caddr_t) cq->ep, (caddr_t) cq->d, cq->cnt);
		cq->fp =  cq->d + cq->cnt;
		cq->ep = cq->d;
		ntrc(("fp=%, ep=%\n",cq->fp, cq->ep ));
	}
	/* prevent ptr wrap & tidy up for next time */
	if (cq->cnt == 0)
		cq->fp = cq->ep = cq->d;
	if (freshdata) {
		*ptr = PKT_END;	
		ttrc(("calling ctf_procrecv\n"));
		ctf_procrecv(up);
	}

	return (count);
}


LOCAL int ctf_next_fas (u_char *ptr, int n)
{
	int count;
	rsp_recv *rsp;
  
	/* 
	 * All FAS commands but FAS_RECV have message length exactly
	 * FASCMDSZ (8) bytes.
	 */

	ttrc(("ctf_next_fas: 1st byte=~, n=%\n", *ptr, n));

	if (n < FASCMDSZ)
		return (0);

	if (*ptr != FAS_RECV)
		return (FASCMDSZ);

	rsp = (rsp_recv *) ptr;
	count = FASCMDSZ + (rsp->cnt_hi << 8) + rsp->cnt_lo;
	return (count <= n ? count : 0);
}


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,23)
LOCAL int unit_select (struct tty_struct *tty, struct inode *inode,
		       struct file *filp, int sel_type, select_table *wait)
{
	circle_q *cq;
	ct_unit *up;
	int mod;
  
	if (!tty) {
		printk("null tty ptr in unit_select()\n");
		return -ENODEV;
	}
	mod = get_module(tty);
	if (mod < 0)
		return (-ENODEV); 

	/* get unit pointer */
	up = els_units[mod][DEV2UNIT(MINOR(tty->device))];
	if (!up || (up->unit_flags & UNIT_OFFLINE))
		return (-ENODEV);

	/*
	 * This select routine does not check for collision due to the fact
	 * that there should ever be only one process (the daemon process)
	 * calling select(). Having more than one cdetherd daemon per unit
	 * is a very bad idea, and not just because it breaks this routine!
	 */
    
	cq = &up->netdata_out;
	switch (sel_type) {
	case SEL_IN:
		if (cq->cnt)
			return (1);
		else 
			selectwait(&up->read_wait, wait);
		break;

	case SEL_OUT:
		return (1);
	}
  
	return (0);
}
#else
LOCAL u_int 
unit_poll (struct tty_struct *tty, struct file *filp, poll_table *wait)
{
	circle_q *cq;
	ct_unit *up;
	u_int mask;
	int mod;
  
	if (!tty) {
		printk("null tty ptr in unit_poll()\n");
		return -ENODEV;
	}
	mod = get_module(tty);
	if (mod < 0)
		return (-ENODEV); 

	/* get unit pointer */
	up = els_units[mod][DEV2UNIT(MINOR(tty->device))];
	if (!up || (up->unit_flags & UNIT_OFFLINE))
		return (-ENODEV);
  
	ttrc(("unit poll\n"));
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,89)
	poll_wait(&up->read_wait, wait);
	poll_wait(&up->write_wait, wait);
#else
	poll_wait(filp, &up->read_wait, wait);
	poll_wait(filp, &up->write_wait, wait);
#endif
	cq = &up->netdata_out;
	mask = POLLWRNORM | POLLOUT;
	if (cq->cnt)
		mask |= POLLRDNORM | POLLIN;
	
	return (mask);
}
#endif


/*
 * termios_differ returns TRUE if the two termios structure args differ
 * in any way.
 */
LOCAL int termios_differ(struct termios *a, struct termios *b)
{
	int i;
	if (a==NULL || (b==NULL))
		return TRUE;
	if ((a->c_cflag) != (b->c_cflag))
		return TRUE;
	if ((a->c_iflag) != (b->c_iflag))
		return TRUE;
	if ((a->c_oflag) != (b->c_oflag))
		return TRUE;
	if ((a->c_lflag) != (b->c_lflag))
		return TRUE;
	if ((a->c_line) != (b->c_line))
		return TRUE;
	for (i=0; i < NCCS ; i++)
		if (a->c_cc[i] != b->c_cc[i])
			return TRUE;
	return FALSE;
}


/*
 * els_ioctl()
 */
LOCAL int
els_ioctl (struct tty_struct *tty, struct file *file, u_int cmd, u_long arg)
{
	ct_unit *up;
	ct_line *lp;
	stsextctl xc;
	struct termios tmp_termios;
	int err_level, tmp;
	int dev = MINOR(tty->device);
	int rc = 0;
	struct serial_struct sio;

	if (!tty) {
		printk("null tty ptr in els_ioctl()\n");
		return -ENODEV;
	}
	trc(("els_ioctl\n"));
	/* check for global */
	if (dev == ELS_GLOBAL_MINOR) {
		/* handle special global diag/util ioctl */
		rc = -EINVAL;
		switch (cmd) {
		case STSFLUSH:
			/* arg is whole device number */
			return ctm_flush_ioctl(arg);

			/* For diag, we take the target diag device's dev_t
			 * from the arg, and scan thru the unit and line
			 * structures to return a copy of the line structure
			 * back to the diag program.
			 */
		case STSLINEDIAG:	/* let cddiag see our line params */
		{
			int pun, mod, port, DUTdev;
			
			if ((rc = verify_area(VERIFY_READ, (void *) arg,
					      sizeof(int)))) 
				return (rc);
			COPY_FROM_USER(&pun, (void *) arg, sizeof(int));
			DUTdev = (unsigned short) pun;
			pun = DEV2UNIT(DUTdev);
			if (pun < 0 || pun >= MAX_ELS_DEVS)
				return (-ENODEV);
			mod = dev_to_mod(DUTdev);
			if (mod < 0)
				return (-ENODEV);
			up = els_units[mod][pun];
			if (up == (ct_unit *) NULL)
				return (-ENODEV);
      
			port = DEV2LINE(DUTdev);
			if (port < 0 || port >= up->num_lines)
				return (-ENODEV);
			if (!(lp = up->ct_lines[port])) 
				return (-ENOMEM);
			lp->next_hist = ((lp->fas_histptr - lp->fas_hist) / 
					 FASCMDSZ);
			if ((rc = verify_area(VERIFY_WRITE, (void *) arg,
					      sizeof(ct_line))))
				return (rc);
			COPY_TO_USER((void *) arg, lp, sizeof(ct_line));
			return (0);
		}
#ifdef TRACER
		case STSGETTRCBUF:
			if ((rc = verify_area(VERIFY_WRITE, 
					      (void *) arg, TRCSZ)))
				return (rc);
			COPY_TO_USER((char *) arg, trcbuf.buf, TRCSZ);
			rc = 0;
			break;
		case  STSCLRTRCBUF:
			bzero(trcbuf.buf, TRCSZ);
			trcbuf.fill = trcbuf.buf;
			*trcbuf.fill = '~';
			rc = 0;
#endif
		}
		return (rc);
	}
	/*
	 * If this is a unit node, hand off the ioctl to the routine
	 * that handles those.
	 */
	if (ELS_UNIT(dev)) {
		ntrc(("els_ioctl(%) - ELS unit\n", cmd));
		return (unit_ioctl(tty, file, cmd, arg));
	}
  
	lp = tty->driver_data;
	if (lp == (ct_line *) NULL)
		return (-ENODEV);

	/*
	 * notice we do this *AFTER* handing off the unit_ioctls because
	 * STSLINEDIAG uses ct_magic for devnum transport.
	 */
	if (lp->ct_magic  != CT_MAGIC) /* paranoia  */
		return (-ENODEV);
  
	if ((cmd != TIOCGSERIAL) && (cmd != TIOCSSERIAL) &&
	    (cmd != TIOCMIWAIT) && (cmd != TIOCGICOUNT)) {
		if (tty->flags & (1 << TTY_IO_ERROR))
			return (-EIO);
	}
  
	rc = 0;
	up = lp->up;
  
	/* check for proprietary ones first */
	switch (cmd) {
	case STSSETEXTCTL: /* handles loopback, sets timers, error level */
	{
		struct tty_struct *xc_tty;
		int save;

		if ((rc = verify_area(VERIFY_READ, (void *) arg, sizeof(xc)))) 
			return (rc);
      
		COPY_FROM_USER(&xc, (void *) arg, sizeof(xc));
      
		err_level = xc.x_flags & XC_ERRLEV;
		if (xc.x_flags & XC_LOOP) {
			if (!suser())
				return (-EPERM);
	
			if ((lp->line_state & LINE_LOOPBACKON) == 0) {
				lp->line_state |= LINE_LOOPBACKON;
				trc(("calling from STSSETEXTCTL "
				     "loopback on\n"));
				LINE_LOCK(lp->linelock, save);
				xc_tty = lp->tty;
				LINE_UNLOCK(lp->linelock, save);
				if (xc_tty)
					rc= ctf_param(lp,xc_tty->termios,TRUE);
				else 
					rc = ENODEV;
			}
		} else {
			if (lp->line_state & LINE_LOOPBACKON) {
				lp->line_state &= ~LINE_LOOPBACKON;
				trc(("calling from STSSETEXTCTL "
				     "loopback off\n"));
				LINE_LOCK(lp->linelock, save);
				xc_tty = lp->tty;
				LINE_UNLOCK(lp->linelock, save);
				if (xc_tty)
					rc =ctf_param(lp,xc_tty->termios,TRUE);
				else 
					rc = ENODEV;
			}
		}
		if (rc == OK) {
			/* looks OK, record all new vals */
			lp->error_level = err_level;
			lp->local_timeout = xc.local_timeout;
			lp->nonloc_timeout = xc.nonloc_timeout;
			lp->local_waitout = xc.local_waitout;
			lp->nonloc_waitout = xc.nonloc_waitout;
			lp->cdbaud_delta=(lp->cdbaud_ext ^ 
					  (xc.x_flags&XC_BAUDEXT)) ?
				TRUE : FALSE;
			lp->cdbaud_ext = (xc.x_flags&XC_BAUDEXT) >>
				XC_BAUDSHIFT;
		}  
		return (rc);
	}
    
	case STSGETEXTCTL:
		if ((rc = verify_area(VERIFY_WRITE, (void *) arg, sizeof(xc))))
			return (rc);
    
		bzero((caddr_t) &xc, sizeof(struct stsextctl));
		xc.x_flags |= lp->error_level;
		if (lp->line_state & LINE_LOOPBACKON)
			xc.x_flags |= XC_LOOP;
		xc.x_flags |= (XC_BAUDEXT&(lp->cdbaud_ext<< XC_BAUDSHIFT));
		xc.local_timeout = lp->local_timeout;
		xc.nonloc_timeout = lp->nonloc_timeout;
		xc.local_waitout = lp->local_waitout;
		xc.nonloc_waitout = lp->nonloc_waitout;
		strcpy(xc.product_id, up->uinfo->name);
		strncpy(xc.fw_version, up->fw_ver,VERLEN);
		xc.scsi_id = up->pun;
    
		COPY_TO_USER((void *) arg, &xc, sizeof(xc));
		return (0);

	case STSFLUSH:
	case STSGETTRCBUF:
	case STSCLRTRCBUF:
		/* if any of these are tried on a normal port, return EINVAL */
		return (-EINVAL);
	}

	/* now the termio ioctls */
	/* we rely on line discipline routines here whenever possible */
	/* first, a sanity check - are we ONLINE? */
	if (up->unit_flags & UNIT_OFFLINE)
		return -EIO;
	switch (cmd) {
	case TCGETS:
	case TCGETA:
		if (tty->ldisc.ioctl) 
			rc = (tty->ldisc.ioctl)(tty, file, cmd, arg);
		else
			rc = -ENOIOCTLCMD;
		break;
	case TCSETSF:			/* WAIT & FLUSH */
	case TCSETAF:			/* WAIT & FLUSH */
		/*
		 * we have to grab the baud rate and check it now,
		 * even though we reuse the line discipline routines
		 * for setting the termios.
		 */
		if (cmd==TCSETAF) {
			tmp_termios.c_line = 0;
			if ((rc = verify_area(VERIFY_READ, (void *) arg,
					      sizeof(struct termio)))) 
				return (rc);
			if (user_termio_to_kernel_termios(&tmp_termios,
						  (struct termio *) arg))
				return -EFAULT;
		} else {
			if ((rc = verify_area(VERIFY_READ, (void *) arg,
					      sizeof(struct termios)))) 
				return (rc);
			if (user_termios_to_kernel_termios(&tmp_termios,
						   (struct termios *) arg))
				return -EFAULT;
		}
		if ((rc = illegal_baud(lp,&tmp_termios)))
			return rc;
		ctf_wait_send(lp,0x7fffffff);
		if (signal_pending(current))
			return -EINTR;
		ctf_iflush_wait(lp);
		if (tty->ldisc.ioctl) 
			rc = (tty->ldisc.ioctl)(tty, file, cmd, arg);
		else
			rc = -ENOIOCTLCMD;
		break;
	case TCSETSW:			/* WAIT FOR OUTPUT */
	case TCSETAW:			/* WAIT */
		ctf_wait_send(lp,0x7fffffff);
		if (signal_pending(current))
			return -EINTR;
		/* drop thru */
	case TCSETS:
	case TCSETA:			/* SET RIGHT NOW */
		/* same silly story as TCSETAF above. */
		if (cmd == TCSETA || cmd==TCSETAW) {
			memcpy(&tmp_termios, tty->termios,
			       sizeof(struct termios));
			if ((rc = verify_area(VERIFY_READ, (void *) arg,
					      sizeof(struct termio)))) 
				return (rc);
			if (user_termio_to_kernel_termios(&tmp_termios,
						  (struct termio *) arg))
				return -EFAULT;
		} else {
			if ((rc = verify_area(VERIFY_WRITE, (void *) arg,
					      sizeof(struct termios)))) 
				return (rc);
			if (user_termios_to_kernel_termios(&tmp_termios,
						   (struct termios *) arg))
				return -EFAULT;
		}
		if ((rc = illegal_baud(lp,&tmp_termios)))
			return rc;
		if (tty->ldisc.ioctl) 
			rc = (tty->ldisc.ioctl)(tty, file, cmd, arg);
		else
			rc = -ENOIOCTLCMD;
		break;
	case TCFLSH:
		rc = tty_check_change(tty); /* for SIGTTOU */
		if (rc)
			return rc;
		/*
		 * The linux tty driver doesn't have a flush
		 * input routine for the driver, assuming all backed
		 * up data is in the line disc. buffers.  However,
		 * we all know that's not the case.  Here, we
		 * act on the ioctl, but then lie and say we didn't
		 * so the line discipline will process the flush
		 * also.
		 */
		switch (arg)
		{
		case TCIFLUSH:
			ctf_iflush_wait(lp);
			break;
		case TCIOFLUSH:
			ctf_iflush_wait(lp);
			/* drop thru */
		case TCOFLUSH:
			ctf_flush_output(lp, TRUE);
			break;
		default:
			trc(("bad TCFLSH param %\n", arg));
			return -EINVAL;
		}
		rc =(-ENOIOCTLCMD); /* make line disc do some work. */
		break;
	case TCXONC:
		rc = tty_check_change(tty);
		if (rc)
			return rc;
		printz("TCXONC arg %ld\n", (arg&0xffff));
		switch (arg) 
		{
		case TCOOFF:
			if (!tty->flow_stopped)
				ctf_suspend_output(lp);
			break;
		case TCOON:
			if (tty->flow_stopped)
				ctf_resume_output(lp);
			break;
		case TCIOFF:
			if (STOP_CHAR(tty) != __DISABLED_CHAR)
				ctf_suspend_input(lp);
			break;
			break;
		case TCION:
			if (START_CHAR(tty) != __DISABLED_CHAR)
				ctf_resume_input(lp);
			break;
		default:
			return -EINVAL;
		}
		rc = (-ENOIOCTLCMD); /* make line disc do some work. */
		break;
		/*	rc = 0; */
		/*
		 * TCSBRK, TCSBRKP are handled via els_break_ctl() in
		 * later kernels.
		 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,66)
	case TCSBRK:	/* SVID version: non-zero arg --> no break */
		rc = tty_check_change(tty);
		if (rc)
			return rc;
		tty_wait_until_sent(tty, 0);
		if (signal_pending(current))
			return -EINTR;
		if (!arg) {
			send_break(info, HZ/4);	/* 1/4 second */
			if (signal_pending(current))
				return -EINTR;
		}
		return 0;
	case TCSBRKP:	/* support for POSIX tcsendbreak() */
		rc = tty_check_change(tty);
		if (rc)
			return rc;
		tty_wait_until_sent(tty, 0);
		if (signal_pending(current))
			return -EINTR;
		send_break(info, arg ? arg*(HZ/10) : HZ/4);
		if (signal_pending(current))
			return -EINTR;
		return 0;
#endif
		/* end of new ioctls */
	case TIOCGSOFTCAR:
		if ((rc = verify_area(VERIFY_READ, (void *) arg,
				      sizeof(u_int))))
			return (rc);

		PUT_USER(((tty->termios->c_cflag & CLOCAL) ? 1 : 0),
			 (u_int *) arg);
		break;

	case TIOCSSOFTCAR:
		if ((rc = verify_area(VERIFY_READ, (void *) arg,
				      sizeof(u_int))))
			return (rc);
    
		GET_USER(tmp, (unsigned int *) arg);
		tty->termios->c_cflag =
			(tty->termios->c_cflag & ~CLOCAL) | (tmp ? CLOCAL : 0);
    
		break;
    
	case TIOCMGET:
		if ((rc = verify_area(VERIFY_WRITE, (void *) arg,
				      sizeof(u_int))))
			return (rc);
    
		PUT_USER(lp->mdmsigs, (u_int *) arg);
		break;
    
	case TIOCMBIS:
	case TIOCMBIC:
	case TIOCMSET:
		if ((rc = verify_area(VERIFY_READ, (void *) arg,
				      sizeof(u_int))))
			return (rc);
    
		GET_USER(tmp, (unsigned int *) arg);
		ctf_smdm(lp, tmp, cmd,  TRUE );	/* wait for response rev B */
		break;
    
	case TIOCGSERIAL:
		if ((rc = verify_area(VERIFY_WRITE, (void *) arg, 
				      sizeof(struct serial_struct))))
			return (rc);
    
		bzero(&sio, sizeof(struct serial_struct));
		sio.line = lp->port;
		sio.flags = lp->flags;
		sio.close_delay = lp->close_delay;
		sio.closing_wait = lp->closing_wait;
		sio.type = PORT_STARTECH;
		sio.xmit_fifo_size = 64;
		COPY_TO_USER((void *) arg, &sio, sizeof(struct serial_struct));
		
		break;
    
	case TIOCSSERIAL:
		if ((rc = verify_area(VERIFY_READ, (void *) arg, 
				      sizeof(struct serial_struct))))
			return (rc);
    
		COPY_FROM_USER(&sio, (void *) arg,
			       sizeof(struct serial_struct));
		
		lp->flags &= ~ASYNC_USR_MASK;
		lp->flags |= (sio.flags & ASYNC_USR_MASK);
		tty->low_latency = (sio.flags & ASYNC_LOW_LATENCY) ? 1 : 0;
		lp->close_delay = sio.close_delay;
		lp->closing_wait = sio.closing_wait;
		break;
    
	case TIOCMIWAIT:
		if ((rc = verify_area(VERIFY_READ, (void *) arg,
				      sizeof(u_int))))
			return (rc);
    
		GET_USER(lp->wait_mask, (unsigned int *) arg);
    
		while (1) {
			interruptible_sleep_on(&lp->mdmsig_wait);
			/* see if a signal did it */
			if (signal_pending(current))
				return -ERESTARTSYS;
			else {
				lp->wait_mask = 0;
				break;
			}
		}
		break;
    
	case TIOCGICOUNT:
		if ((rc = verify_area(VERIFY_WRITE, (void *) arg, 
				      sizeof(struct serial_icounter_struct))))
			return (rc);
    
		COPY_TO_USER((void *) arg, &lp->icount, 
			     sizeof(struct serial_icounter_struct));
		break;
	default:
		rc= (-ENOIOCTLCMD);		/* not -ENOTTY */
	}
	return rc;
}


LOCAL int
unit_ioctl (struct tty_struct *tty, struct file *file, u_int cmd, u_long arg)
{
	int pun;
	ct_unit *up;
	els_config els_cfg;
	ctunitinfo *uinfo;
	u_char *ptr;
	int dev = MINOR(tty->device);
	int mod, rc;
  
	/* all these require root privilege */
	if (!suser())
		return (-EPERM);
  
	mod = get_module(tty);
	if (mod < 0)
		return (-ENODEV); 
  
	pun = DEV2UNIT(dev);
	up = els_units[mod][pun];
  
	ntrc(("unit_ioctl %\n", dev));
	switch (cmd) {
	case ELSADDUNIT:			/* add/config ELS unit */
		
		ntrc(("ELSADDUNIT\n"));
		
		if ((rc = verify_area(VERIFY_READ, (void *) arg,
				      sizeof(els_config))))
			return (rc);

		COPY_FROM_USER(&els_cfg, (char *) arg, sizeof(els_cfg));

		if (up) {
			if (strcmp(up->uinfo->name, els_cfg.product_id)) {
				trc(("already configured as a different "
				     "product\n"));
				return (-EINVAL);
			}
      
			if (!(up->unit_flags & UNIT_OFFLINE))
				ctm_unit_offline(up);
      
			/*
			 * Re-initialize read and select routine for the unit
			 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,23)
			tty->ldisc.poll = unit_select;
#else
			tty->ldisc.poll = unit_poll;
#endif
			tty->ldisc.read = unit_read;
			tty->ldisc.write = unit_write;

			strcpy(up->fw_ver, els_cfg.fw_ver);
			up->netaddr = els_cfg.netaddr;
			ptr = (u_char *) &up->netaddr;
			printk("Digi EtherLite (els%2.2x): %s (%s) at IP "
			       "address %d.%d.%d.%d\n", mod * 7 + pun,
			       up->uinfo->name, up->fw_ver,
			       ptr[0], ptr[1], ptr[2], ptr[3]);
			return (0);
		}

		if ((up = (ct_unit *) ct_malloc(sizeof(ct_unit)))==NULL)
			return -ENOMEM;

		/* Save our up pointer in our global array. */
		els_units[mod][pun] = up;
		up->pun = pun;
		up->unum = mod*7+pun;	/* R0.12 */

		/* 
		 * Now the driver must make sure that the IP address
		 * of the ELS unit is valid, and that the product ID of 
		 * the unit is correct. But we can't do that right now.
		 * The ELS firmware does not support it yet. We can only
		 * assume that everything is correct.
		 */

		/* get config data */
		if (!els_cfg.product_id) {
			trc(("not given product name\n"));
			els_units[mod][pun] = NULL;
			return (-EINVAL);
		}

		uinfo = ctm_get_unitinfo(els_cfg.product_id);
		if (!uinfo) {
			trc(("no product match\n"));
			els_units[mod][pun] = NULL;
			return (-EINVAL);
		}

		if (!(uinfo->spec_func_code & SFC_ETH)) {
			trc(("not etherlite product\n"));
			els_units[mod][pun] = NULL;
			return (-EINVAL);
		}

		/* keep ptr to unitinfo in unit record */
		up->uinfo = uinfo;
		strcpy(up->fw_ver, els_cfg.fw_ver);
    
		/* record network address and allocate circular queue */
		up->netaddr = els_cfg.netaddr;
		ctm_reset_circular_queue(&up->netdata_in);
		ctm_reset_circular_queue(&up->netdata_out);
		up->avg_rtt = up->avg_rtt_dev = 0; /* R0.12 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
		spin_lock_init(&up->faslock);
#endif

		/* make packets for this unit */
		ctm_morepkts(6, MAX_PKT);
		up->recv_ctpkt = ctm_getpkt(MAX_PKT); /* R0.12 was up->max_opkt */
		
		up->fw_features = (FW_RSL | FW_RESET | FW_WAIT_SEND |
				   FW_MDM_PERSIST);
		up->num_lines = uinfo->num_lines;
		up->max_opkt   = uinfo->max_pkt;
		up->max_size_fassend = min(uinfo->max_pkt,
					   (uinfo->max_line_out));
		up->unit_flags = UNIT_OFFLINE; /* force init */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
		init_waitqueue_head(&up->read_wait);
		init_waitqueue_head(&up->write_wait);
#else
		up->read_wait = NULL;
		up->write_wait = NULL;
#endif

		/*
		 * Allocate an array of line pointers based on the maximum
		 * number of ports supported on this device type.
		 */
		up->ct_lines = (ct_line **) 
			ct_malloc(sizeof(ct_line *) * up->num_lines);
		if (up->ct_lines==NULL) {
			ct_free(els_units[mod][pun]);
			els_units[mod][pun] = NULL;
			return -ENOMEM;
		}
		/*
		 * Initialize read and select routine for the unit
		 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,23)
		tty->ldisc.poll = unit_select;
#else
		tty->ldisc.poll = unit_poll;
#endif
		tty->ldisc.read = unit_read;
		tty->ldisc.write = unit_write;

		/*
		 * Print a nice message identifing our device.
		 */
		ptr = (u_char *) &up->netaddr;
		printk("Digi els%2.2x: %s (%s) at IP address %d.%d.%d.%d\n",
		       mod * 7 + pun, uinfo->name, up->fw_ver, ptr[0], ptr[1],
		       ptr[2], ptr[3]);
		break;

	case ELSRMUNIT:				/* remove ELS unit */
		if (!up) {
			trc(("ELSRMUNIT - unit not configured\n"));
			return (-ENODEV);
		}
		trc(("ELSRMUNIT\n"));
    
		if (!(up->unit_flags & UNIT_OFFLINE))
			ctm_unit_offline(up);
    
		break;

	case ELSGETCFGDATA:		/* get config data of the ELS unit */
		if (!up) {
			trc(("ELSGETCFGDATA - unit not configured\n"));
			return (-ENODEV);
		}
		trc(("ELSGETCFGDATA\n"));

		if ((rc = verify_area(VERIFY_WRITE, (void *) arg,
				      sizeof(els_config))))
			return (rc);

		els_cfg.netaddr = up->netaddr;
		strcpy(els_cfg.product_id, up->uinfo->name);
		strcpy(els_cfg.fw_ver, up->fw_ver);
		
		COPY_TO_USER((char *) arg, &els_cfg, sizeof(els_cfg));
		break;
    
	case ELSONLINE:
		if (!up) {
			trc(("ELSONLINE - unit not configured\n"));
			return (-ENODEV);
		}
		trc(("ELSONLINE\n"));
    
		if (up->unit_flags & UNIT_OFFLINE) {
			up->active_time = ct_systime;
			up->avg_rtt = up->avg_rtt_dev = 0;
			up->unit_flags = UNIT_INITD;
			ctm_reset_circular_queue(&up->netdata_in);
			ctm_reset_circular_queue(&up->netdata_out);
			ctlogerr(STSERR_DEFAULT, up, NULL, "(%s) Unit online",
				 (int) up->uinfo->name);
		}
    
		break;

	case ELSOFFLINE:
		if (!up) {
			trc(("ELSOFFLINE - unit not configured\n"));
			return (-ENODEV);
		}
		trc(("ELSOFFLINE\n"));
    
		if (!(up->unit_flags & UNIT_OFFLINE))
			ctm_unit_offline(up);
    
		break;
    
	default:
		trc(("default\n"));
		return (-EINVAL);
	}
  
	return (0);
}



/*
 * how controls whether or not the delay is interruptible via signals.
 */
LOCAL void ct_delay (int t, int how)
{

	if (t > 0) {
		set_current_state(how);
		schedule_timeout( t);
		set_current_state(TASK_RUNNING);
	}
}


LOCAL void *ct_malloc (int size)
{
	void *ptr;

	ptr = (void *) kmalloc(size, GFP_ATOMIC); /* can be called in interrupt context */
	bzero((caddr_t) ptr, size);
	return (ptr);
}


LOCAL void ct_free (void *ptr)
{
	kfree(ptr);
}


LOCAL void ctm_reset_circular_queue (circle_q *cq)
{
	cq->fp = cq->d;
	cq->ep = cq->d;
	cq->end = &cq->d[CQSIZE]; /* 1 *BEYOND* the last byte stored */
	cq->cnt = 0;
}


LOCAL int ctm_circular_queue_space (circle_q *cq)
{
	return (CQSIZE - cq->cnt - 1);
}


/*
 * ctlogerr()
 *
 * Print error message to console identifying unit and describing condition.
 * Display standard header, unique format string, and one optional parameter.
 *
 * If the line ptr 'lp' is NULL, it doesn't display the line #, indicating it
 * was a unit level error.
 */
/*VARARGS4*/
LOCAL void ctlogerr (u_int err_level, ct_unit *up, ct_line *lp, 
		     char *err_string, int opt_arg)
{
	char fmt[160];

	if (lp) {
		if (err_level > lp->error_level)
			return;

		strcpy(fmt, "Digi els%2.2x, line %u: ");
		strcat(fmt, err_string);
		strcat(fmt, "\n");
		Printf(fmt, up->unum, lp->port, opt_arg);
	} else {
		if (err_level > ct_err_level)
			return;

		strcpy(fmt, "Digi els%2.2x: ");
		strcat(fmt, err_string);
		strcat(fmt, "\n");
		Printf(fmt, up->unum, opt_arg);
	}
}


/*
 * ctf_timer()
 *
 * Coarse-grained timer for local driver time keeping.	 Runs every
 * second to check on things that go slow.
 *
 * Duties:
 *	- keeps local system time value for offline delays
 *	- pokes all enabled lines every 30 seconds to make sure
 *	  they're alive
 */
LOCAL void ctf_timer (unsigned long unused)
{
	int mod, pun, port, save;
	ct_unit *up;
	ct_line *lp;
	req_input_ctrl *iccmd;
	static int last_scan;
	req_unit_poll *pollcmd;

	/* update our own little copy of time */
	ct_systime++;

	/* see if it's time to probe lines again */
	if (ct_systime - last_scan < CTF_TIMER_INTERVAL) 
		goto done;
	/* handle slow timer duties */
	last_scan = ct_systime;
	
	/* scan ELS units */
	for (mod = 0; mod < MAX_MODULE; mod++) {
		for (pun = 0; pun < MAX_ELS_DEVS; pun++) {
			up = els_units[mod][pun];
			if (up == NULL || (up->unit_flags & UNIT_OFFLINE))
				continue;

			ttrc(("ctf_timer - ct_systime = %, up->active_time "
			      "= %\n", ct_systime, up->active_time));

		        /* Unit has gone too long without responding */
			if (ct_systime - up->active_time > ELS_UNIT_GONE) {
				printz("timer ctf_offline\n");
				trc(("timer running ctm_unit_offline()\n"));
				ctm_unit_offline(up);
				continue;
			}

			ELS_LOCK(up->faslock, save);

			pollcmd = (req_unit_poll *) ctm_putfasclrcmd(up);
			pollcmd->opcode = FAS_UNIT_POLL;

			for (port = 0; port < up->num_lines; port++) {
				lp = up->ct_lines[port];
	
				/* see if line is active */
				if (lp == NULL) 
					continue;

				if ((lp->line_state & (LINE_OFFLINE |
						       LINE_ENABLED |
						       LINE_DISABLING))
				    != LINE_ENABLED)
					continue;

				/*
				 * basically just send a NOP out to look 
				 * for errors
				 */
				iccmd = (req_input_ctrl *)
					ctm_putfasclrcmd(up);
				iccmd->opcode = FAS_INPUT_CTRL;
				iccmd->line = lp->port;
				iccmd->flush = 0;
				iccmd->suspend = 0;
				iccmd->resume = 0;
#if 0
				/* don't log these nops */
				ctf_logfasout(lp, (u_char *) iccmd);
#endif
			}

			ELS_UNLOCK(up->faslock, save);
			up->poll_time = jiffies;
			ctm_sendfas(up);
		}
	}
	
done:
	/* and reload ourself (unless we are done) */
	if (!kill_timer) {
		ct_timer.function = ctf_timer;
		ct_timer.expires = jiffies + HZ;
		add_timer(&ct_timer);
	}
}


LOCAL void ctm_sendfas (ct_unit *up)
{
	ct_pkt *pp;
	int save;

	ttrc(("ctm_sendfas\n"));

	ELS_LOCK(up->faslock, save);
  
	/* see if there's a packet waiting */
	if ((pp = up->pkt_head) == NULL) {
		ELS_UNLOCK(up->faslock, save);
		return;
	}
  
	/* make sure something is in the ct_pkt */
	if (pp->dp == pp->d) {
		ELS_UNLOCK(up->faslock, save);
		return;
	}
  
	/* Note that we DO NOT need the PKT_END in ELS */
  
	/*
	 * If there is enough space in the circular queue, copy the data
	 * in ct_pkt to the queue, and release the ctpkt.
	 */
	if (pp->dp - pp->d <= CQSIZE - up->netdata_out.cnt) {
		ctf_move_to_queue(pp->d, &up->netdata_out, pp->dp - pp->d);
		if ((up->pkt_head = pp->link) == NULL)
			up->pkt_tail = NULL; /* show queue empty */
		ctm_freepkt(pp);
	}
  
	ttrc(("netdata_out has % bytes\n", up->netdata_out.cnt));
  
	/* wakeup up the daemon process */
	wake_up_interruptible(&up->read_wait);
	ELS_UNLOCK(up->faslock, save);
}


/*
 * ctf_init_line()
 */
LOCAL int ctf_init_line (ct_line *lp)
{
	int rc=0;

	trc(("ctf_init_line\n"));

	/* acquire open/close semaphore */
	while (test_and_set_bit(OCSEMA,&lp->ocsema) != 0) {
		interruptible_sleep_on(&lp->open_wait);
		if (signal_pending(current))
			return -ERESTARTSYS;
	}
  
	/* see if firmware already going for line */
	if ((lp->line_state & (LINE_ENABLED|LINE_DISABLING)) == LINE_ENABLED) {
		lp->line_state &= ~LINE_OFFLINE;
		if (test_and_clear_bit(OCSEMA,&lp->ocsema) == 0)
			trc(("ctf_init_line semaphore already clear\n"));
		wake_up_interruptible(&lp->open_wait);
		return (OK);
	}
  
	/*
	 * if unit is offline, delay checking again for a while. If we get
	 * interrupted, update the offline time so next time we sail thru.
	 */
	if (lp->line_state & LINE_OFFLINE) {
		while (ct_systime < lp->offline_time + CTF_OFFLINE_HOLDOFF) {
			ct_delay(HZ/2, TASK_UNINTERRUPTIBLE);
			if (signal_pending(current)) {
				lp->offline_time = ct_systime - 
					CTF_OFFLINE_HOLDOFF;
				if (test_and_clear_bit(OCSEMA,&lp->ocsema)== 0)
					trc(("ctf_init_line semaphore "
					     "already clear\n"));
				wake_up_interruptible(&lp->open_wait);
				trc(("ctf_init_line signaled in offline "
				     "hold.\n"));
				return (-EINTR);	/* bail w/ error */
			}
		}
	}
  
	/* now do firmware at line */
	lp->line_state &= ~LINE_LOOPBACKON;
	if ((rc=ctf_enable_fw(lp))) {
		if (test_and_clear_bit(OCSEMA,&lp->ocsema) == 0)
			trc(("ctf_init_line semaphore already clear\n"));
		wake_up_interruptible(&lp->open_wait);
		trc(("ctf_enable_fw failed.\n"));
		lp->line_state |= LINE_OFFLINE; /* go back to offline? -Kees */
		return (rc);			/* bail w/ error */
	}
  
	/* unit is initd, line is enabled, do bookkeeping, set init state */
	lp->line_state &= (LINE_ENABLED | LINE_PRINTER);
  
	lp->last_parms.opcode = 0xff;	/* mark as no longer valid */
	lp->last_flow.opcode = 0xff;	/* ditto */
  
	/* release sema, return success */
	if (test_and_clear_bit(OCSEMA,&lp->ocsema) == 0)
		trc(("ctf_init_line semaphore already clear\n"));
	wake_up_interruptible(&lp->open_wait);
	return (OK);
}


/*
 * wait for receipt of commands used to handle ioctl change.
 */
LOCAL int ctf_block_ioctl (ct_line *lp, int flag, int timer)
{
	int wait_time;
  
	if ((timer <= 0) || (current->state != TASK_RUNNING))
		return (OK);
  
	wait_time = jiffies + timer * HZ;
  
	/* wait for FAS response to come back */
	while ((lp->line_state & flag) && (jiffies < wait_time)) {
		schedule();
		set_current_state(TASK_RUNNING);
	}
	return ((lp->line_state & flag) ? -ENODEV : OK);
}



/*
 * ctf_startin()
 * Command the line to receive more data, but limit it to the smallest
 * of the MAX_RECV_CHUNK, room in the flip buffer, and what the line
 * discipline won't throw away.	Make request 64 bytes smaller to
 * get around a firmware problem with FAS_RECV.
 *
 * Start the input chain if no disabling factors are noted.
 * Don't bother if no room in line discipline - it will discard.
 * Unthrottle should start it as needed.
 */
LOCAL void ctf_startin (ct_line *lp)
{
	req_recv *cmd;
	int save;
	struct tty_struct *ttyp;
	int max_recv;
  
	LINE_LOCK(lp->linelock, save);
	if ((!(ttyp=lp->tty) || (ttyp->magic != TTY_MAGIC) ||
	     ttyp->closing || (ttyp->flags & (1 << TTY_IO_ERROR)) ||
	     (lp->line_state&(LINE_PRINTER | LINE_ISTOPPED | LINE_IPENDING |
			      LINE_OFFLINE))) || 
	    ((lp->line_state & (LINE_ENABLED | LINE_DISABLING)) != 
	     LINE_ENABLED)) {
		LINE_UNLOCK(lp->linelock, save);
		return;
	}
	max_recv = max(0,(MAX_RECV_CHUNK - ttyp->flip.count));
	max_recv = min(max_recv, lp->curr_size_fassend); /* R0.12 */
	if (ttyp->ldisc.receive_room)
		max_recv = min(max_recv, ttyp->ldisc.receive_room(ttyp));
	/*
	 * we have to subtract 64 from the size requested to get around
	 * a firmware bug fixed in 7.6 firmware
	 */
	max_recv = min(max_recv, (max(20,(max_recv-64))));
	max_recv = min(max_recv, 384); /* REV A fix possible timeout problem */
	LINE_UNLOCK(lp->linelock, save);
	if (!max_recv)
		return;
	/*
	 * Start a FAS_RECV for this line.
	 */
	ELS_LOCK(lp->up->faslock, save);
	lp->line_state |= LINE_IPENDING;
	cmd = (req_recv *) ctm_putfasclrcmd(lp->up);
	cmd->opcode = FAS_RECV;
	cmd->line = lp->port;
	cmd->cnt_lo = load_lo(max_recv);
	cmd->cnt_hi = load_hi(max_recv);
	ctf_logfasout(lp, (u_char *) cmd);
	ELS_UNLOCK(lp->up->faslock, save);
	ctm_sendfas(lp->up);
}


/*
 * ctf_param()
 *
 * Set parameters on port.
 *
 */
LOCAL int ctf_param (ct_line *lp, struct termios *tiosp, int flag)
{
	req_set_params parms;
	req_flow_ctrl flow;
	caddr_t parms_p;
	caddr_t flow_p;
	int baudcode;
	int save;
	int status;
	int port_type;
	ct_unit *up = lp->up;
	struct tty_struct *ttyp;

	/* determine which the port type, only process the serial port */
	port_type = ctf_port_type(up->uinfo, lp->port);
	if (port_type & PT_PARALLEL)
		return (OK);
  
	if (!(ttyp=lp->tty) || (ttyp->magic != TTY_MAGIC)  || !tiosp)
		return (OK);
  
	/*
	 * note: both set_params and flow_ctrl cmds occupy full 8 byte
	 *       block so no bzero necessary
	 */
  
	/* if the output baud rate is 0, hangup */
	if ((tiosp->c_cflag & (CBAUD^CBAUDEX)) == B0) {
		if (flag) {
			trc(("baudrate is B0, drop modem signals\n"));
			ctf_smdm(lp, TIOCM_RTS | TIOCM_DTR, TIOCMBIC, TRUE);
			return OK;
		}
		baudcode = FAS9600;
	}
	else
		baudcode = ctf_baud(lp, tiosp);
  
	/*
	 * check for legal baudrates which is complicated by units
	 * that don't support all Linux rates (e.g. B50,B134,B230400)
	 * (The first ELS units only supported 16 baud rates.)
	 */
	if (baudcode == FASNOBAUD) {
		trc(("Invalid baudrate is FASNOBAUD %\n", baudcode));
		return (-EINVAL);
	}
	if (ctf_check_baud(up->uinfo, lp->port, baudcode) != OK) {
		trc(("Invalid baudrate %\n", baudcode));
		return (-EINVAL);
	}
	lp->int_obaud = ctf_int_baud(baudcode);
	update_print_max(lp); /* update max packet based on line's properies */
	update_curr_size_fassend(lp); /* set the max fassend size */
  
	if (tiosp->c_cflag & CLOCAL) 
		lp->flags &= ~ASYNC_CHECK_CD;
	else
		lp->flags |= ASYNC_CHECK_CD;
  
	parms.line = lp->port;
	if (PT_USES_NEW_SET_PARAMS(port_type)) {
		parms.opcode = FAS_EXT_PARAMS;
		parms.baudrate_out = baudcode;
		parms.baudrate_in = baudcode;
	} else {
		parms.opcode = FAS_SET_PARAMS;
		parms.baudrate_lo = baudcode | (baudcode << 4);
		parms.baudrate_hi = 0;
	}
  
	/* set up parity */
	if (tiosp->c_cflag & PARENB) {
		if (tiosp->c_cflag & PARODD)
			parms.parity = ODD_PARITY;
		else
			parms.parity = EVEN_PARITY;
	} else
		parms.parity = NO_PARITY;

	/* set up character size */
	switch (tiosp->c_cflag & CSIZE) {
	case CS5:
		parms.c_size = C_SZ5;
		break;
	case CS6:
		parms.c_size = C_SZ6;
		break;
	case CS7:
		parms.c_size = C_SZ7;
		break;
	default:
		parms.c_size = C_SZ8;
	}
  
	/* set up number of stop bits */
	if (tiosp->c_cflag & CSTOPB)
		parms.c_frame = C_FRAME2;
	else
		parms.c_frame = C_FRAME1;
  
	/* see if receive is enabled */
	parms.flgs = (tiosp->c_cflag & CREAD) ? PF_RXENAB : 0;
  
	/* see if hardware loopback is desired */
	if (lp->line_state & LINE_LOOPBACKON) 
		parms.flgs |= PF_LOOP;
	
	/* now set up flow control cmd */
	flow.opcode = FAS_FLOW_CTRL;
	flow.line = lp->port;
	flow.xon = tiosp->c_cc[VSTART];
	flow.xoff = tiosp->c_cc[VSTOP];
  
	/*
	 * Need to check for POSIX disable. We have to disable both
	 * sides of flow control if either START or STOP is disabled.
	 */
	flow.in_mode = flow.out_mode = FC_NONE; /* revB */
	if ((tiosp->c_cc[VSTART]!=__DISABLED_CHAR) && 
	    (tiosp->c_cc[VSTOP]!=__DISABLED_CHAR)) {
		/* set up input flow control */
		if (tiosp->c_iflag & IXOFF)
			flow.in_mode = FC_XON_XOFF1;
		else
			flow.in_mode = FC_NONE;
    
		/* set up output flow control */
		switch (tiosp->c_iflag & (IXANY | IXON)) {
		case IXANY:
		case 0:
			flow.out_mode = FC_NONE;
			break;
      
		case IXON:
			flow.out_mode = FC_XON_XOFF1;
			break;
      
		case IXANY | IXON:
			flow.out_mode = FC_IXANY;
		}
	}
	/*
	 * Check for hardware flow control requested.
	 */
	if (tiosp->c_cflag & CRTSCTS) {
		flow.out_mode |= FC_RTS_CTS;
		flow.in_mode |= FC_RTS_CTS;
	}
  
	/*
	 * Need to figure out a fair chunk this line can use per pkt for
	 * outgoing data. We don't want to buffer huge amounts of stuff out
	 * at the unit or we lose time sync between host and terminal.	We
	 * want to limit to about 1/10 second worth of output at a time so a
	 * human is aware of what is actually running.
	 * 
	 * Printers don't matter, let them be more efficient and hog 3/4
	 * or more of a packet.
	 */
	if (lp->print_max) {
		flow.buf_max_lo = (u_char) lp->print_max;
		flow.buf_max_hi = (u_char) (lp->print_max >> 8);
	} else {
		flow.buf_max_lo = (u_char) lp->up->max_size_fassend;
		flow.buf_max_hi = (u_char) (lp->up->max_size_fassend >> 8);
	}
	/* now see if we need to actually send out the commands */
	if (ct_cmp((u_char *) &parms, (u_char *) &lp->last_parms, FASCMDSZ) ||
	    ct_cmp((u_char *) &flow, (u_char *) &lp->last_flow, FASCMDSZ)) {
    
		lp->line_state |= LINE_SET_PARAMS;
    
		/* put params cmd in ct_pkt */
		ELS_LOCK(lp->up->faslock, save);
		parms_p = (caddr_t) ctm_putfascmd(up);
		bcopy((caddr_t) &parms, parms_p, FASCMDSZ);
		bcopy(parms_p, (caddr_t) &lp->last_parms, FASCMDSZ);
    
		/* put flow control cmd in ct_pkt */
		flow_p = (caddr_t) ctm_putfascmd(up);
		bcopy((caddr_t) &flow, flow_p, FASCMDSZ);
		bcopy(flow_p, (caddr_t) &lp->last_flow, FASCMDSZ);
    
		ctf_logfasout(lp, (u_char *) parms_p);
		ctf_logfasout(lp, (u_char *) flow_p);
		ELS_UNLOCK(lp->up->faslock, save);
    
		/* kick off operation */
		ctm_sendfas(up);
	}
  
	status = OK;
	if (lp->line_state & LINE_SET_PARAMS) 
		status = ctf_block_ioctl(lp, LINE_SET_PARAMS, 60);
  
	return (status);
}


/*
 * ctf_handle_recv()
 *	
 * Process recv data from ELS unit. (but only if port is still open.)
 */
LOCAL void ctf_handle_recv (ct_line *lp, rsp_recv *rrsp)
{	
	int status = 0;
	u_char *fas_data;
	int len=0;
	int count, save;
	struct tty_struct *tty;
	int iflag;

	if (!lp)
		return;
	lp->line_state &= ~LINE_IPENDING;
	if (!lp->up) {
		trc(("null up in ctf_handle_recv\n"));
		return;
	}
	if (!rrsp) {
		trc(("null rrsp in ctf_handle_recv\n"));
		goto recv_end;
	}
	count = (rrsp->cnt_hi << 8) | rrsp->cnt_lo;
	fas_data = ctm_getfas(lp->up, count);
	if (lp->line_state&LINE_WAIT_IFLUSH) {
		goto recv_end;
	}
	tty = lp->tty;
#if BYPASS_FLIP
	if (tty && tty->real_raw && tty->ldisc.receive_room &&
	    !tty->flip.count && !(rrsp->status & ERR_MASK))
		len = (tty->ldisc.receive_room)(tty);
#endif
	/* don't allow tty to vanish while we work */
	LINE_LOCK(lp->linelock, save);
	/* double check tty */  
	if (((lp->line_state & (LINE_OFFLINE | LINE_ENABLED | LINE_DISABLING))
	     != LINE_ENABLED) || (lp->flags & ASYNC_CLOSING) ||
	    !(tty = lp->tty) || (tty->magic != TTY_MAGIC) ||
	    (tty->flags & (1 << TTY_IO_ERROR))) {
		LINE_UNLOCK(lp->linelock, save);
		goto recv_end;
	}

	lp->icount.rx += count;
  
	trc(("ctf_handle_recv 0x% bytes data, status 0x%\n", 
	     count, rrsp->status));
  
	/*
	 * bypass flip buffers if we can. It saves a copy or two. Requires:
	 * we have a ldisc , real_raw is set, the flip
	 * buffer is empty, no uart status on the packet, and line disc
	 * has room for the whole packet.
	 * If we didn't wait for the flip count to be 0, we might see
	 * out-of-order data delivery. -RPR
	 */
#if BYPASS_FLIP
	if (len && len >= count) {
		trc(("count=%, receive_room=%\n", count, len));
		ntrc(("bypassed flip buffer on receive % bytes\n", count));
		(tty->ldisc.receive_buf)(tty, fas_data, NULL, count);
		LINE_UNLOCK(lp->linelock, save);
		ctf_startin(lp);
		return;
	}
#endif /* BYPASS_FLIP */
	if ((tty->magic == TTY_MAGIC) &&(tty->termios))
		iflag = tty->termios->c_iflag;
	else
		iflag = 0;
	switch (rrsp->status & ERR_MASK) {
	case ERR_FRAMING:			/* normal framing error */
		if (!(iflag & INPCK))
			break;

		if (!(iflag & IGNPAR)) {
			lp->icount.frame++;
			status = TTY_FRAME;
		} else
			count = 0;
		break;
    
	case ERR_PARITY:
		if (!(iflag & INPCK))
			break;
    
		if (iflag & IGNPAR)
			count = 0;
		else {
			lp->icount.parity++;
			status = TTY_PARITY;
		}
		break;

	case ERR_BREAK:
		if (iflag & IGNBRK)
			count = 0;
		else {
			lp->icount.brk++;
			status = TTY_BREAK;
			if ((lp->flags & ASYNC_SAK) && tty)
				do_SAK(tty);
		}
		break;

	case ERR_OVERRUN:
		if (!(iflag & INPCK))
			break;
    
		if (!(iflag & IGNPAR)) {
			lp->icount.overrun++;
			status = TTY_OVERRUN;
		} else
			count = 0;
		break;

	case ERR_UNKNOWN:
	default:
		if (rrsp->status & ERR_MASK)
			trc(("handle_recv unknown error ~\n",
			     (rrsp->status & ERR_MASK)));
		break;
		/* don't know how to handle this, pretends nothing is wrong */
	case ERR_OVERFLOW:
		trc(("ERR_OVERFLOW unit ~ port ~\n",lp->up->unum,lp->port));
	}

	if (!tty || !tty->flip.char_buf_ptr) {
		len = 0;
	} else {
		/* sometimes count exceeds TTY_FLIPBUF_SIZE! */
		len = max(0,(TTY_FLIPBUF_SIZE - tty->flip.count));
	}
	
	if (count > len) {
		lp->icount.buf_overrun += count - len;
		printk("dropping %d chars out of %d on the floor\n",
		       count-len, count);
		trc(("dropping % chars out of % on floor\n",
		     count-len, count));
		count = len;
	}
  
	if (count > 0) {
		bcopy(fas_data, tty->flip.char_buf_ptr, count);
		tty->flip.char_buf_ptr += count;
		tty->flip.count += count;
    
		/*
		 * ELS units do not store a byte-by-byte account of framing,
		 * parity, overflow, break, etc, but only report the most
		 * recent error condition accumulated for each recvd message.
		 * We copy the error status of the entire message to each of
		 * the bytes in the message. It's good enough for POSIX. If
		 * you really need more status reporting, you could set the
		 * FAS_RECV max buffer size to 1 byte. I don't recommend it,
		 * though. -RPR
		 */
		if (status == 0) {
			bzero(tty->flip.flag_buf_ptr, count);
			tty->flip.flag_buf_ptr += count;
		} else {
			while (count > 0) {
				*tty->flip.flag_buf_ptr = status;
				count--;
				tty->flip.flag_buf_ptr++;
			}
		}
    		tty_schedule_flip(tty);
	}
			
	LINE_UNLOCK(lp->linelock, save);
	/* keep input coming */
 recv_end:
	ctf_startin(lp);
}


/*
 * ctf_handle_send()
 * Called on receipt of last FAS_SEND. Send more output data (and xchar)
 * as called for by current output state.
 */
LOCAL void ctf_handle_send (ct_line *lp)
{
	int save;
	struct tty_struct *hs_tty;

	LINE_LOCK(lp->linelock, save);
	lp->line_state &= ~LINE_OBUSY;

	if (((lp->line_state & (LINE_OFFLINE | LINE_ENABLED | LINE_DISABLING))
	     != LINE_ENABLED) || !(hs_tty=lp->tty) || 
	    (hs_tty->magic != TTY_MAGIC) ||
	    (hs_tty->flags & (1 << TTY_IO_ERROR))) {
		LINE_UNLOCK(lp->linelock, save);
		return;
	}
	LINE_UNLOCK(lp->linelock, save);
	if (lp->xchar)
		els_send_xchar(hs_tty, lp->xchar);
	else if (lp->xmit_q.cnt) 
		els_write(hs_tty, 0, NULL, 0);
	else {   /* pppd needs a kick here to resume output */
		if ((hs_tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
		    hs_tty->ldisc.write_wakeup)
			(hs_tty->ldisc.write_wakeup)(hs_tty);
    
		wake_up_interruptible(&hs_tty->write_wait);
	}
}


/*
 * ctf_wait_send()
 * Insures any output buffered but not queued is queued as a FAS command,
 * and then issues a FAS_WAIT_SEND command and waits for
 * FAS_WAIT_SEND completion message to return. (Normal FAS_SEND completion
 * isn't sufficient as DATA CAN REMAIN in the ELS on receipt of FAS_SEND.
 */
LOCAL int ctf_wait_send (ct_line *lp, int timer)
{
	req_wait_send	*wscmd;
	int status = OK;
	int save;
	int wait_time;
	struct tty_struct *lp_tty;

	/* If old firmware doesn't have this command, just return */
	if (!(lp->up->fw_features & FW_WAIT_SEND)) {
		trc(("ctf_wait_send: no such feature\n"));
		return (OK);
	}
  
	/* if line is belly up, skip */
	if (lp->line_state & (LINE_OFFLINE|LINE_DISABLING)) {
		trc(("ctf_wait_send: LINE_OFFLINE|LINE_DISABLING\n"));
		printz("ctf_wait_send: LINE OFFLINE|LINE_DISABLING\n");
		return (OK);
	}
 
	trc(("ctf_wait_send: timer=%\n",timer));
	if (!timer) {
		LINE_LOCK(lp->linelock, save);
		timer = (lp->tty && NONLOCAL(lp)) ? 
			lp->nonloc_waitout : lp->local_waitout;
		LINE_UNLOCK(lp->linelock, save);
	}
	if (timer <= 0)
		timer = 60;
	if (timer > 3600)
		timer = 3600;
	wait_time = jiffies + timer * HZ;
	/* If the xmit circular buffer isn't drained, drain it first! */
	while (lp->xmit_q.cnt) {
		trc(("ctf_wait_send: xmit q=%\n",lp->xmit_q.cnt));
		if (lp->line_state&LINE_OBUSY) {
			while ((lp->line_state & LINE_OBUSY) &&
			       (jiffies < wait_time)) {
				/* use ct_delay instead? -Kees
				schedule();
				set_current_state(TASK_RUNNING);
				*/
				ct_delay(HZ/10,TASK_INTERRUPTIBLE);
				if (signal_pending(current))
					goto timeout;
			}
		}
		LINE_LOCK(lp->linelock, save);
		lp_tty = lp->tty;
		LINE_UNLOCK(lp->linelock, save);
		if (lp_tty)
			els_write(lp_tty, 0, NULL, 0);
	}
	/* again, if line is belly up, skip */
	if (lp->line_state & (LINE_OFFLINE|LINE_DISABLING)) {
		trc(("ctf_wait_send: LINE_OFFLINE|LINE_DISABLING\n"));
		printz("ctf_wait_send: LINE OFFLINE|LINE_DISABLING\n");
		return (OK);
	}
	if (!(lp->line_state & LINE_WAIT_OUT)) {
		lp->line_state |= LINE_WAIT_OUT;
		ELS_LOCK(lp->up->faslock, save);
		wscmd = (req_wait_send *) ctm_putfasclrcmd(lp->up);
		wscmd->opcode = FAS_WAIT_SEND;
		wscmd->line = lp->port;
		ctf_logfasout(lp, (u_char *) wscmd);
		ELS_UNLOCK(lp->up->faslock, save);
		ctm_sendfas(lp->up);
	}
 
	/* wait for FAS response to come back */
	while ((lp->line_state & LINE_WAIT_OUT) && (jiffies < wait_time)) {
		/* again, use ct_delay instead -Kees
		schedule();
		set_current_state(TASK_RUNNING);
		*/
		ct_delay(HZ/10, TASK_INTERRUPTIBLE);
		if (signal_pending(current))
			break;
	}
 timeout:
	status = (lp->line_state & LINE_WAIT_OUT) ? !OK : OK;
	lp->line_state &= ~LINE_WAIT_OUT;
	if (lp->xmit_q.cnt)
		trc(("ctf_wait_send xmit_qcnt=%\n", lp->xmit_q.cnt));
	ntrc(("ctf_wait_send returns ~\n", status));
	return (status);
}


/*
 * ctf_iflush_wait()
 *	
 * Issues flush input command & waits for completion.
 */
LOCAL int ctf_iflush_wait (ct_line *lp)
{
	req_input_ctrl *cmd;
	int status = OK;
	int save;
	int wait_time;

	/* if line is belly up, skip */
	if (lp->line_state & (LINE_OFFLINE|LINE_DISABLING)) {
		trc(("ctf_iflush_wait: LINE_OFFLINE\n"));
		return (OK);
	}

	if (!(lp->line_state & LINE_WAIT_IFLUSH)) {
		lp->line_state |= LINE_WAIT_IFLUSH;
		/* do flush at unit */
		ELS_LOCK(lp->up->faslock, save);
		cmd = (req_input_ctrl *) ctm_putfasclrcmd(lp->up);
		cmd->opcode = FAS_INPUT_CTRL;
		cmd->line = lp->port;
		cmd->flush = TRUE;
		cmd->suspend = FALSE;
		cmd->resume = FALSE;
		/* mark this one so we can differentiate receipt of flush
		   as different than other FAS_INPUT_CTRL responses. */
		cmd->pad[0] = 1;
		ctf_logfasout(lp, (u_char *) cmd);
		ELS_UNLOCK(lp->up->faslock, save);
		ctm_sendfas(lp->up);
	}

	wait_time = jiffies + 10 * HZ; /* 10 seconds just to be safe */

	/* wait for FAS response to come back */
	while ((lp->line_state & LINE_WAIT_IFLUSH) && (jiffies < wait_time)) {
		schedule();
		set_current_state(TASK_RUNNING);
		if (signal_pending(current))
			break;
	}

	status = (lp->line_state & LINE_WAIT_IFLUSH) ? !OK : OK;
	lp->line_state &= ~LINE_WAIT_IFLUSH;
	return (status);
}


/*
 * ctf_offline()
 *	
 * Attempts to start recovery for a line that has quit responding.
 */
LOCAL void ctf_offline (ct_line *lp)
{
	int port_type;

	trc(("ctf_offline els~ port ~\n", lp->up->unum, lp->port));

	if (!lp) 
		return;
	/* log latest time reported AWOL */
	lp->offline_time = ct_systime;
  
	/* but see if we already know */
	if (lp->line_state & LINE_OFFLINE) {
		trc(("ctf_offline - already marked offline\n"));
		return;
	}
  
	/* mark line as missing in action */
	lp->line_state |= LINE_OFFLINE;
	lp->line_state &= LINE_OBUSY;	/* clear output busy state R0.12 */

	/* if line not enabled, skip further cleanup */
	if (!(lp->line_state & LINE_ENABLED)) {
		trc(("ctf_offline - not enabled\n"));
		return;
	}
  
	/* if carrier not active on serial port, fake going active */
	port_type = ctf_port_type(lp->up->uinfo, lp->port);
	if (PT_IS_SERIAL(port_type)  && 
	    !(lp->mdmsigs & TIOCM_CAR)) {
		lp->mdmsigs |= TIOCM_CAR;
		wake_up_interruptible(&lp->open_wait);
	}
  
	/* wake up any wait send that's sleeping */
	lp->line_state &= ~(LINE_OBUSY | LINE_ISTOPPED | LINE_OSTOPPED |
			    LINE_WAIT_OUT | LINE_WAIT_IFLUSH | LINE_IPENDING);
}


/*
 * ctm_unit_offline()
 */
LOCAL void ctm_unit_offline (ct_unit *up)
{
	ct_line *lp;
	int i, max, enbflag;

	trc(("ctm_unit_offline\n"));
  
	/* scan all lines */
	max = up->num_lines;
	for (i = 0; i < max; i++) {
		lp = up->ct_lines[i];
		if (lp == NULL) 
			continue;
		trc(("ctm_unit_offline- line=%x\n",lp->line_state));
		lp->flags &= ~ASYNC_INITIALIZED;
		enbflag = (lp->line_state & LINE_ENABLED);
		ctf_offline(lp);
		if (enbflag && lp->tty) {
			tty_hangup(lp->tty);
		}
	}
  
	/* if unit not already marked offline, handle fresh offline */
	if ((up->unit_flags & UNIT_OFFLINE) == 0) {
		ctlogerr(STSERR_DEFAULT, up, NULL, "(%s) Unit offline",
			 (int) up->uinfo->name);
		up->unit_flags |= UNIT_OFFLINE;
	}
  
	/* update offline time regardless */
	up->unit_offline_time = ct_systime;
}


/*
 * ctf_deinit_line()
 */
LOCAL void ctf_deinit_line (ct_line *lp)
{
	trc(("ctf_deinit_line line ~, state %\n", lp->port, lp->line_state));
  
	/* acquire open/close semaphore */
	while (test_and_set_bit(OCSEMA,&lp->ocsema) != 0) {
		interruptible_sleep_on(&lp->open_wait);
	}
	if ((lp->line_state & LINE_ENABLED)) {
		/* disable firmware for this line (if unit's still online) */
		if (lp->line_state & LINE_OFFLINE) 
			lp->line_state &= ~LINE_ENABLED;
		else
			ctf_disable_fw(lp);
	}
	if (test_and_clear_bit(OCSEMA,&lp->ocsema) == 0)
		trc(("ctf_deinit_line: semaphore already clear\n"));
	wake_up_interruptible(&lp->open_wait);
}


/*
 * ct_cmp()
 *	
 * Compares two regions of memory.
 * Returns 0 if they match, non-zero otherwise.
 */
LOCAL int ct_cmp (u_char *p1, u_char *p2, int cnt)
{
	while (cnt--) {
		if (*p1++ != *p2++)
			return (1);
	}
	return (0);
}


LOCAL void els_put_char (struct tty_struct *tty, unsigned char ch)
{
	ct_line *lp;
	int save;
  
	if ((!tty) || (tty->magic != TTY_MAGIC) || 
	    !(lp = (ct_line *) tty->driver_data))
		return;
	ELS_LOCK(lp->up->faslock, save);
	if (ctm_circular_queue_space(&lp->xmit_q) >= 1) 
		ctf_move_to_queue(&ch, &lp->xmit_q, 1);
	ELS_UNLOCK(lp->up->faslock, save);
}


/*
 * Rather than discarding (as the name implies), els_flush_chars()
 * STARTS characters to OUTPUT,  but doesn't wait for it to complete.
 */
LOCAL void els_flush_chars (struct tty_struct *tty)
{
	ct_line *lp;

	if ((!tty) || (tty->magic != TTY_MAGIC) || 
	    !(lp = (ct_line *) tty->driver_data))
		return;
	els_write(tty, 0, NULL, 0);
}


LOCAL int els_write_room (struct tty_struct *tty)
{
	ct_line *lp;
  
	if ((!tty) || (tty->magic != TTY_MAGIC) ||
	    !(lp = (ct_line *) tty->driver_data))
		return 0;
	if ((lp->line_state & (LINE_OFFLINE | LINE_ENABLED | LINE_DISABLING)) 
	    != LINE_ENABLED) 
		return (0);
  
	return (ctm_circular_queue_space(&lp->xmit_q));
}


/*
 * Disable delivery of input data.
 */
LOCAL void els_throttle (struct tty_struct *tty)
{
	ct_line *lp;

	if ((!tty) || (tty->magic != TTY_MAGIC) ||
	    !(lp = (ct_line *) tty->driver_data))
		return;
	lp->line_state |= LINE_ISTOPPED;
}


/*
 * Enable delivery of input data.
 */
LOCAL void els_unthrottle (struct tty_struct *tty)
{
	ct_line *lp;

	if ((!tty) || (tty->magic != TTY_MAGIC) ||
	    !(lp = (ct_line *) tty->driver_data))
		return;
	lp->line_state &= ~LINE_ISTOPPED;
	ctf_startin(lp);
}


/*
 * Disable delivery of output data.
 */
LOCAL void els_stop (struct tty_struct *tty)
{
	ct_line *lp;

	ntrc(("els_stop\n"));
	if ((!tty) || (tty->magic != TTY_MAGIC) ||
	    !(lp = (ct_line *) tty->driver_data))
		return;
	lp->line_state |= LINE_OSTOPPED;
}


/*
 * Enable delivery of output data.
 */
LOCAL void els_start (struct tty_struct *tty)
{
	ct_line *lp;

	ntrc(("els_start\n"));
	if ((!tty) || (tty->magic != TTY_MAGIC) ||
	    !(lp = (ct_line *) tty->driver_data))
		return;
	lp->line_state &= ~LINE_OSTOPPED;
	if (lp->xmit_q.cnt)
		els_write(tty, 0, NULL, 0);
}


/*
 * report number of characters in output buffer.  If the line is
 * enabled and has uncompleted output, we fib a bit and add one to the
 * characters in output buffer to make routines waiting for zero
 * buffered work properly.
 */
LOCAL int els_chars_in_buffer (struct tty_struct *tty)
{
	ct_line *lp;
  
	ttrc(("els_chars_in_buffer\n"));
	if ((!tty) || (tty->magic != TTY_MAGIC) ||
	    !(lp = (ct_line *) tty->driver_data))
		return 0;
	if (lp->xmit_q.cnt)
		return (lp->xmit_q.cnt);
	return ((lp->line_state & (LINE_OBUSY | LINE_ENABLED)) == 
		(LINE_OBUSY | LINE_ENABLED)) ? 1 : 0;
}


LOCAL void els_send_xchar (struct tty_struct *tty, char ch)
{
	ct_line *lp;
	req_send *sendcmd;
	int save;

	ntrc(("els_send_xchar, char ~\n", ch));
	if ((!tty) || (tty->magic != TTY_MAGIC) ||
	    !(lp = (ct_line *) tty->driver_data))
		return;

	if (((lp->line_state & (LINE_OFFLINE | LINE_ENABLED |
				LINE_DISABLING))) != LINE_ENABLED) 
		return;
  
	ELS_LOCK(lp->up->faslock, save); 

	if (lp->line_state & LINE_OBUSY) {
		lp->xchar = ch;
		ELS_UNLOCK(lp->up->faslock, save);
		return;
	} else {
    
		/* fill out and send FAS_SEND message */
		sendcmd = (req_send *) ctm_putfasclrcmd(lp->up);
		sendcmd->opcode = FAS_SEND;
		sendcmd->line = lp->port;
		sendcmd->cnt_lo = 1;
		sendcmd->cnt_hi = 0;
    
		trc(("send_xchar %\n",ch));
		*ctm_putfas(lp->up, 1) = ch;
		lp->xchar = 0;
		lp->line_state |= LINE_OBUSY;
	}
  
	ELS_UNLOCK(lp->up->faslock, save);
	ctm_sendfas(lp->up);
}


/*
 * els_flush_buffer() clears the output buffers
 * locally and on the Etherlite unit.
 */
LOCAL void els_flush_buffer (struct tty_struct *tty)
{
	ct_line *lp;
  
	if (!tty || (tty->magic != TTY_MAGIC) ||
	    !(lp = (ct_line *) tty->driver_data))
		return;
	ctf_flush_output(lp, TRUE);
	if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
	    tty->ldisc.write_wakeup)
		(tty->ldisc.write_wakeup)(tty);
}


/*
 * called by tty_hangup() on loss of carrier. 
 * Does a hachet job on the port, a close without waits
 * and no delivery of any queued data.
 */
LOCAL void els_hangup (struct tty_struct *tty)
{
	ct_line	*lp;
	int save=0;
	req_reset *reset;
 
	trc(("els_hangup %\n", tty));
	if ((!tty) || (tty->magic != TTY_MAGIC) ||
	    !(lp = (ct_line *) tty->driver_data))
		return;
  
	els_flush_buffer(tty);
	if (lp->flags & ASYNC_CLOSING)
		return;
	if (tty->termios->c_cflag & HUPCL)
		ctf_smdm(lp, TIOCM_RTS | TIOCM_DTR, TIOCMBIC, FALSE);
	ELS_LOCK(lp->up->faslock, save);
	set_bit(TTY_IO_ERROR, &tty->flags); /* bar data delivery */
	lp->flags &= ~(ASYNC_CALLOUT_ACTIVE | ASYNC_NORMAL_ACTIVE |
		       ASYNC_CLOSING|ASYNC_INITIALIZED );
	lp->line_state &= ~(LINE_ENABLED);
	lp->line_state |= LINE_RESET|LINE_DISABLING;
	reset = (req_reset *) ctm_putfasclrcmd(lp->up);
	reset->opcode = FAS_RESET;
	reset->line = lp->port;
	ctf_logfasout(lp, (u_char *) reset);
	lp->refcount = 0;
	lp->tty = (struct tty_struct *) NULL;
	ELS_UNLOCK(lp->up->faslock, save);
	ctm_sendfas(lp->up);
	wake_up_interruptible(&lp->open_wait);
}


/*
 * added test for null termios ptr
 */
LOCAL void els_set_termios (struct tty_struct *tty, struct termios *old)
{
	int mdm;
	int cflag;
	ct_line *lp;

	if ((!tty) || (tty->magic != TTY_MAGIC) ||
	    !(lp = (ct_line *) tty->driver_data))
		return;
	if (!old)
		return;
	if (lp->line_state & (LINE_OFFLINE|LINE_DISABLING))
		return;				/* prevent FAS INITD errors */

	if (tty_check_change(tty))		/* POSIX job control check */
		return;
  
	if (!lp->cdbaud_delta && !termios_differ(tty->termios, old))
		return;		/* set something once, why set it again? */
	trc(("calling ctf_param in set_termios=%\n",tty->termios));
	cflag = tty->termios->c_cflag;  
	if (ctf_param(lp, tty->termios, TRUE)) {
		/*
		 * Sorry, Charlie, that baud isn't supported --
		 * revert to old rate.
		 */
		tty->termios->c_cflag &= ~(CBAUD);
		tty->termios->c_cflag |= (old->c_cflag&(~CBAUD));
		return;
	}
  
	lp->cdbaud_delta = FALSE;
	/* Handle transition away from B0 status */
	if (old && !(old->c_cflag & (CBAUD^CBAUDEX)) &&
	    (cflag & (CBAUD^CBAUDEX))) {
		mdm = TIOCM_DTR;
		if (!(cflag & CRTSCTS) || !test_bit(TTY_THROTTLED,&tty->flags))
			mdm |= TIOCM_RTS;
    
		ctf_smdm(lp, mdm, TIOCMBIS, TRUE);
	}
}


LOCAL void els_wait_until_sent (struct tty_struct *tty, int timeout)
{
	ct_line *lp;

	ntrc(("els_wait_until_sent\n"));
	if ((!tty) || (tty->magic != TTY_MAGIC) ||
	    !(lp = (ct_line *) tty->driver_data))
		return;
	ctf_wait_send(lp, timeout);
}


LOCAL void ctf_clear_flow_control (ct_line *lp)
{
	req_flow_ctrl	*cmd;
	int save;

	if (lp->line_state & (LINE_OFFLINE|LINE_DISABLING))
		return;

	ELS_LOCK(lp->up->faslock, save);
	cmd = (req_flow_ctrl *) ctm_putfascmd(lp->up);
	cmd->opcode = FAS_FLOW_CTRL;
	cmd->line = lp->port;
	cmd->in_mode = FC_NONE;
	cmd->out_mode = FC_NONE;
	lp->last_flow.opcode = 0xff;	/* invalidate our cached flow cmd */

	ELS_UNLOCK(lp->up->faslock, save);
	ctm_sendfas(lp->up);
}


/*
 * Generate a break condition. The line protocol already waited out
 * any pending output for us.
 */
LOCAL void els_break_ctl (struct tty_struct *tty, int state)
{
	req_send_break *cmd;
	ct_line *lp;
	int save;
  
	if ((!tty) || (tty->magic != TTY_MAGIC) ||
	    !(lp = (ct_line *) tty->driver_data))
		return;
	if (!state || (lp->line_state & (LINE_OFFLINE|LINE_DISABLING)))
		return;
  
	ELS_LOCK(lp->up->faslock, save);
	lp->line_state |= LINE_BREAK;
	cmd = (req_send_break *) ctm_putfasclrcmd(lp->up);
	cmd->opcode = FAS_SEND_BREAK;
	cmd->line = lp->port;
	ctf_logfasout(lp, (u_char *) cmd);
	ELS_UNLOCK(lp->up->faslock, save);
	ctm_sendfas(lp->up);
	ctf_block_ioctl(lp, LINE_BREAK, 60);
}


LOCAL void bzero (void *ptr, int count)
{
	char *ptr0 = (char *) ptr;

	while (count > 0) {
		*ptr0++ = 0;
		count--;
	}
}


LOCAL void bcopy (void *s, void *d, int count)
{
	char *ptr0 = (char *) s;
	char *ptr1 = (char *) d;
  
	while (count > 0) {
		*ptr1 = *ptr0;
		ptr0++;
		ptr1++;
		count--;
	}
}

/*
 * return the module index, 0 thru 8, or -1 if not defined.
 */
LOCAL int get_module (struct tty_struct *tty)
{
	int major;
	int i;
  
	if (!tty)
		return -1;
	major = MAJOR(tty->device);
  
	for (i = 0; i < MAX_MODULE; i++) {
		if (els_serial_driver[i].major == major)
			return (i);
    
		if (els_callout_driver[i].major == major)
			return (i);
	}
  	return (-1);
}


/*
 * given a major, return the module index or -1 if no match.
 */
LOCAL int dev_to_mod (dev_t devno)
{
	int i, maj;

	maj = MAJOR(devno);
  
	for (i = 0; i < MAX_MODULE; i++) {
		if (els_serial_driver[i].major == maj)
			return (i);
    
		if (els_callout_driver[i].major == maj)
			return (i);
	}
  	return (-1);
}




/*************************** packet management ****************************/


/*
 * ctm_morepkts()
 */
LOCAL void ctm_morepkts (int npkts, int size)
{
	struct ct_pkt	*pktp;
	int flag;

	ELS_LOCK(ct_pktlock, flag);
  
	while (npkts-- > 0) {
		pktp = (ct_pkt *) ct_malloc(sizeof(struct ct_pkt));
		if (pktp == NULL) {
			printk("out of memory in ctm_morepkts!\n");
			panic("Digi: out of memory in ctm_morepkts!");
		}
		pktp->link = ct_pkt_free;
		ct_pkt_free = pktp;
		ct_pkt_cnt++;
	}
	ELS_UNLOCK(ct_pktlock, flag);
}


/*
 * ctm_lesspkts() - return ct_pkts from the free linked list
 * back to the O/S. Don't count on returning all of ct_pkt_cnt,
 * as some ct_pkts may no longer be in the free list. (R0.12)
 */
LOCAL void ctm_lesspkts ()
{
	struct ct_pkt	*pktp;
	int flag;
  
	ELS_LOCK(ct_pktlock, flag);
  
	while (ct_pkt_cnt > 0 && ((pktp = ct_pkt_free)!=NULL)) {
		ct_pkt_free = pktp->link;
		ct_pkt_cnt--;
		ct_free(pktp);
	}
	ELS_UNLOCK(ct_pktlock, flag);
}


/*
 * ctm_getpkt()
 * 
 * Get ct_pkt, allocate another packet if free
 * list is empty. If the free list is empty,
 * make another ct_pkt. This will happen a few times
 * with lots of port activity. Currently, no mechanism
 * for freeing ct_pkts to the kernel (other than cleanup vi rmmod.) 
 */
LOCAL ct_pkt *ctm_getpkt (int size)
{
	ct_pkt *pp;
	int flag;

 again:
	ELS_LOCK(ct_pktlock, flag);

	if ((pp = ct_pkt_free) != NULL) {
		ct_pkt_free = pp->link;
		pp->link = NULL;
		pp->dp = pp->d;			/* init filling ptr */
		pp->space_remaining = size - 1;	/* allow for terminate code */
	} else {
		ELS_UNLOCK(ct_pktlock, flag);
		++global_buy;  /* keep track how often we consume */
		ctm_morepkts(1,MAX_PKT);
		goto again;
	}
  	ELS_UNLOCK(ct_pktlock, flag);
  
	return (pp);
}


/*
 * ctm_freepkt
 *
 * free up a ct_pkt
 */
LOCAL void ctm_freepkt (ct_pkt *pp)
{
	int flag;

	ELS_LOCK(ct_pktlock, flag);
	pp->link = ct_pkt_free;
	ct_pkt_free = pp;
	ELS_UNLOCK(ct_pktlock, flag);
}




/*********************** MUX layer interface routines **********************/

/*
 * ctm_spaceleft()
 *
 * Called before doing a FAS command that has a data portion (i.e. FAS_SEND).
 * Returns the number of bytes left in current packet.	If less than
 * some nominal value (16 or so), it replenishes the packet and then
 * returns the space free in that packet.  Space free is calculated
 * assuming an 8 byte FAS command will first be put into the packet.
 * And of course, the extra byte for the packet termination code isn't
 * counted. This routine should only be called once per command.
 *
 * Note: it will never return less than 16 due to its replenishing action.
 */
LOCAL int ctm_spaceleft (ct_unit *up)
{
	ct_pkt *pp;
	int count;

	/* get ptr to current ct_pkt, replenishing if necessary */
	if ((pp = up->pkt_tail) == NULL) {
		pp = up->pkt_head = up->pkt_tail = ctm_getpkt(up->max_opkt);
	}
	/* if packet is too full, get fresh one */
	else if (pp->space_remaining < 16 + FASCMDSZ) {
		up->pkt_tail = pp->link = ctm_getpkt(up->max_opkt);
		pp = pp->link;
	}
	count = pp->space_remaining - FASCMDSZ;
  
	return (count);
}


/*
 * ctm_putfas that preclears the bytes.
 */
LOCAL u_char *ctm_putfasclr (ct_unit *up, int length)
{
	u_char *p = ctm_putfas(up, length);
	memset(p, 0, length);
	return p;
}
  

/*
 * ctm_putfas()
 * 
 * Return a pointer to a slot for FAS commands or data 'length' bytes
 * long.  If this routine is being use to write data, 'ctm_spaceleft'
 * should be called first to determine the amount of free space left
 * in the current packet.  Exceeding that amount will cause the packet to
 * get replenished and horrible splitting of commands happens.
 */
LOCAL u_char *ctm_putfas (ct_unit *up, int length)
{
	ct_pkt *pp;
	u_char *cp;

	/* get ptr to current ct_pkt, replenishing if necessary */
	if ((pp = up->pkt_tail) == NULL) {
		pp = up->pkt_head = up->pkt_tail = ctm_getpkt(up->max_opkt);
	}
	/* if packet is too full, get fresh one */
	else if (pp->space_remaining < length) {
		up->pkt_tail = pp->link = ctm_getpkt(up->max_opkt);
		pp = pp->link;
	}
   
	/* update packet, return ptr */
	cp = pp->dp;
	pp->dp += length;
	pp->space_remaining -= length;

	return (cp);
}


/*
 * ctm_getfas()
 *
 * Retrieves responses from recv'd packet. Returns pointer to next
 * FAS command or data chunk, updating emptying pointer.
 */
LOCAL u_char *ctm_getfas (ct_unit *up, int length)
{
	ct_pkt *pp;
	u_char *cp;

	pp = up->recv_ctpkt;
	cp = pp->dp;
	pp->dp += length;
	return(cp);
}


/*
 * ctf_move_to_queue()
 *
 * Move chunk of data from linear buffer to circle queue.
 */
LOCAL void ctf_move_to_queue (u_char *p, circle_q *cqp, int cnt)
{
	int chunk1, chunk2;

	if (cnt > CQSIZE - cqp->cnt)
		panic("ELS: queue overflow!");

	chunk1 = cqp->end - cqp->fp;
	chunk1 = min(chunk1, cnt);
	chunk2 = cnt - chunk1;
  
	bcopy((caddr_t)p, (caddr_t)cqp->fp, chunk1);
	if (chunk2) {
		p += chunk1;
		bcopy((caddr_t)p, (caddr_t)cqp->d, chunk2);
		cqp->fp = cqp->d + chunk2;
	}
	else
		cqp->fp += chunk1;
	cqp->cnt += cnt;
}


/*
 * ctf_move_from_queue()
 *
 * Move chunk of data from circle queue to linear buffer.
 */
LOCAL void ctf_move_from_queue (circle_q *cqp, u_char *p, int cnt)
{
	int chunk1, chunk2;

	if (cnt > cqp->cnt)
		panic("ELS: queue underflow!");
  
	chunk1 = cqp->end - cqp->ep;
	chunk1 = min(chunk1, cnt);
	chunk2 = cnt - chunk1;
	bcopy((caddr_t)cqp->ep, (caddr_t)p, chunk1);
	if (chunk2) {
		p += chunk1;
		bcopy((caddr_t)cqp->d, (caddr_t)p, chunk2);
		cqp->ep = cqp->d + chunk2;
	}
	else
		cqp->ep += chunk1;
	cqp->cnt -= cnt;
}




/*******************	 FAS support routines *********************/

/*
 * ctf_procrecv()
 *
 * Pulls FAS commands out of recv packet and processes.
 *
 */
LOCAL void ctf_procrecv (ct_unit *up)
{
	register u_char *cmdp;
	ct_line *lp;
	int count;

	while (cmdp = ctm_getfascmd(up), *cmdp != PKT_END) {
		/* make sure line index is in range */
		if (cmdp[1] >= up->num_lines) {
			/* display unexpected response */
			ctf_showfascmd("R", cmdp);
			printk("recvd bad line for els%2.2x\n", up->unum);
			ntrc(("recvd bad line #, taking down unit.\n"));
			/* craziness, take unit down */
			ctm_unit_offline(up);
			break;
		}

		/*
		 * get line # from response (always 2nd byte), look up
		 * line struct
		 */
		lp = up->ct_lines[cmdp[1]];

		/* make sure we looked up non-null line struct */
		if (lp == NULL) {
			/*
			 * check for the commands that don't return valid
			 * line #
			 */
			switch (cmdp[0]) {
			case FAS_UNIT_POLL:
				up->active_time = ct_systime;
				ctm_roundtrip_adjust(up);
				break;
			case FAS_GLOBAL:
			case FAS_UNIT_STATS:
			case FAS_MUX_STATS:
				break;

			default:
				ctlogerr(STSERR_QUIET, up, (ct_line *)NULL, 
					 "Null ct ptr on line %u",
					 (int)cmdp[1]);
				ctf_showfascmd("R", cmdp);
				ctm_unit_offline(up);
			}
		}
		else {
			/*
			 * FAS_UNIT_POLL can return what appears to be
			 * valid line
			 */
			if (cmdp[0] == FAS_UNIT_POLL) { 
				up->active_time = ct_systime;
				ctm_roundtrip_adjust(up);
			}
			else
				ctf_logfasin(lp, cmdp);
		}

		/* decode errors, common ones get ASCII interpretation */
		switch (cmdp[2] & ERR_MASK) {
		case OK:		/* not an error */
			break;
      
		case ERR_ABORTED:	/* benign error, ignore */
			break;
      
		case ERR_PARITY:
			ctlogerr(STSERR_VERBOSE, up, lp, "Parity error", 0);
			break;
      
		case ERR_OVERRUN:
			ctlogerr(STSERR_DEFAULT, up, lp, "Overrun error", 0);
			break;
      
		case ERR_FRAMING:
			ctlogerr(STSERR_VERBOSE, up, lp, "Framing error", 0);
			break;

		case ERR_BREAK:
			ctlogerr(STSERR_VERBOSE, up, lp, "Break received", 0);
			break;
			
		case ERR_MORE:		/* benign error, ignore */
			break;
			
		case ERR_OVERFLOW:
			ctlogerr(STSERR_DEFAULT, up, lp,
				 "Input buffer overflow", 0);
			break;
			
		case ERR_OFFLINE:
			ctf_offline(lp);
			break;
			
		case ERR_INITD:
			/* benign initd error expected on a few cmds */	
			if (cmdp[0] != FAS_OUTPUT_CTRL && cmdp[0] !=
			    FAS_STAT_CHG && cmdp[0] != FAS_WAIT_SEND) {
				printk("ERR_INITD cmd %d forces offline\n",
				       cmdp[0]);
				trc(("ERR_INITD cmd ~x forces offline\n",
				     cmdp[0]));
				ctf_offline(lp);
			}
			break;
		case ERR_BUSY:
			if (cmdp[0] == FAS_ENABLE)
				break;
			if (cmdp[0] == FAS_RESET)
				break;
			Printf("Digi STS: ERR_BUSY on cmd: ");

			/* display unexpected response */
			ctf_showfascmd("R", cmdp);
			break;
			
		default:
			if (cmdp[0] == FAS_STAT_CHG)
				break;	/* errors are benign for stat change */
			if (cmdp[0] == FAS_WAIT_SEND)
				break;	/* errors are benign for wait send  */
			if (cmdp[0] == FAS_OUTPUT_CTRL) 
				break;	/* benign INITD errors expected */
      
			/* display unexpected response */	
			printk("Digi: error on cmd: " );
			ctf_showfascmd("R", cmdp);
			break;
		}
    
		/* decode command type */
		switch (*cmdp) {
		case FAS_RESET:
			trc(("Got FAS_RESET\n"));
			/* show offline */
			lp->line_state &= ~(LINE_OFFLINE | LINE_DISABLING);
      
			/* return any error indication */
			lp->enab_disab_stat = cmdp[2];
			break;

		case FAS_ENABLE:
			/* tell oc_fsm */
			lp->line_state |= LINE_ENABLED;
      
			/* return any error indication */
			lp->enab_disab_stat = cmdp[2];
			break;
      
		case FAS_DISABLE:
			/*
			 * if driver leaves modem ctrl sigs active across
			 * opens, simply indicate that we no longer know
			 * the state of the incoming modem status ones.
			 */
			if (lp->up->fw_features & FW_MDM_PERSIST) 
				lp->mdmsigs &= ~(TIOCM_CTS | TIOCM_DSR |
						 TIOCM_CAR);
			else
				lp->mdmsigs = 0;
				/* this is trouble in o-c interaction */
			lp->line_state &= ~(LINE_ENABLED | LINE_DISABLING);

			/* return any error indication */
			lp->enab_disab_stat = cmdp[2];
			break;
      
		case FAS_SEND:
			/* Call the OS specific function to handle a send. */
			trc(("got FAS_SEND\n"));
			ctf_handle_send(lp);
			break;
      
		case FAS_RECV:
			/*
			 * Call the OS specific function to handle the
			 * receive data if not in the process of flushing input
			 */
			ctf_handle_recv(lp, (rsp_recv *) cmdp);
			break;

		case FAS_STAT_CHG:
			ctf_handle_stat_chg(lp, (rsp_stat_chg *) cmdp);
			break;
      
		case FAS_SEND_BREAK:
			lp->line_state &= ~LINE_BREAK;
			break;

		case FAS_INPUT_CTRL:
			/* R0.12 only clear IFLUSH if the reserved byte 5 
			 * is set which denotes receipt of the flush.*/
			trc(("got FAS_INPUT_CTRL %\n",lp->line_state));
			if (cmdp[5])
				lp->line_state &= ~LINE_WAIT_IFLUSH;
			break;
      
		case FAS_WAIT_SEND:
		{
			int save;
			struct tty_struct *lp_tty;
			
			trc(("got FAS_WAIT_SEND %\n",lp->line_state));
			lp->line_state &= ~LINE_WAIT_OUT;
			if ((lp->line_state & (LINE_OFFLINE | 
					       LINE_ENABLED | 
					       LINE_DISABLING))
			    != LINE_ENABLED)
				break;			/* Thanks, Cook! */
			LINE_LOCK(lp->linelock, save);
			lp_tty = lp->tty;
			LINE_UNLOCK(lp->linelock, save);
			if (lp_tty)
				wake_up_interruptible(&lp_tty->write_wait);
			break;
		}
      
		case FAS_UNIT_POLL:
		  /* handled already - don't do this twice.
			up->active_time = ct_systime;
			ctm_roundtrip_adjust(up); */
			break;
      
		case FAS_GLOBAL:
			break;
      
		case FAS_SET_MODEM:
			lp->line_state &= ~LINE_SET_MODEM;
			break;
      
		case FAS_SET_PARAMS:
		case FAS_EXT_PARAMS:
			lp->line_state &= ~LINE_SET_PARAMS;
			break;
      
		case FAS_IN_TIMERS:
		case FAS_FLOW_CTRL:
		case FAS_RESERVE:
		case FAS_RELEASE:
			break;
      
		case FAS_OUTPUT_CTRL:
			break;
			
		case FAS_MUX_STATS:
		case FAS_UNIT_STATS:
		case FAS_LINE_STATE:
			if (rcvrsp->status != ERR_BAD_CMD) {
				count = rcvrsp->cnt_lo + (rcvrsp->cnt_hi << 8);
				ctf_display_statistics(
					(char *) ctm_getfas(up, count), count);
			}
			/* more to come? */
			if (rcvrsp->status == ERR_MORE) {
				req_stat_chg    *statcmd;

				/* resend the command */
				statcmd = (req_stat_chg *) ctm_putfascmd(up);
				statcmd->opcode = rcvrsp->opcode;
				statcmd->line = rcvrsp->line;
			}
			break;
      
		default:
			/*
			 * display unexpected response to the console.
			 * Very bad news.
			 */
			printk("Digi els: unexpected cmd %hd recvd in "
			       "procrecv\a\n", *cmdp);
			ctf_showfascmd("R", cmdp);
			trc(("unexpected cmd ~ unit ~\n", *cmdp, up->pun));
			/* got a garbage command, take unit offline */
			ctm_unit_offline(up);
			return;
		}
	}
}


/*
 * ctf_handle_stat_chg()
 *	
 * Handles the common actions associated with processing a FAS_STAT_CHG
 * response.  Called only from 'ctf_procrecv' at interrupt time w/
 * ints off.
 *
 * As with 'ctf_smdm', all modem stuff is kept in fundamental vocabulary
 * TIOCM_xxx.	You can either define those to be what you need or
 * do the translation.
 *
 * This routine 'ctf_line_change' to handle any other processing required.
 * It has copied the old modem state to lp->old_mdmsigs and installed
 * that new state in lp->mdmsigs.  It is up to 'ctf_line_change' to
 * determine what has changed and what further actions are necessary.
 */
LOCAL void ctf_handle_stat_chg (ct_line *lp, rsp_stat_chg *statrsp)
{
	req_stat_chg	*statcmd;
	int delta, save;
	struct tty_struct *chg_tty;

	/*
	 * Save the current line settings for later use.
	 */
	lp->old_mdmsigs = lp->mdmsigs;
	lp->line_state &= ~LINE_GET_MODEM;
  
	if (((lp->line_state & (LINE_OFFLINE | LINE_ENABLED | LINE_DISABLING))
	     != LINE_ENABLED) ||  ( lp->flags & ASYNC_CLOSING))
		return;
  
	/*
	 * If ERR_INITD or ERR_ABORTED,
	 * don't process or respawn
	 */
	if (statrsp->status)
		return;
  
	/*
	 * Take care of parallel stuff.  No action required, just
	 * log state if they're interested.
	 */
	if (lp->line_state & LINE_PRINTER) {
		/* Special case:  status changes for printers */
    
		if (statrsp->flags & SC_PE) {
			ctlogerr(STSERR_VERBOSE, lp->up, lp,
				 "Paper out at printer", 0);
			lp->mdmsigs &= ~TIOCM_CTS;
		} else
			lp->mdmsigs |= TIOCM_CTS;
    
		if ((statrsp->flags & SC_SEL) == 0) {
			ctlogerr(STSERR_VERBOSE, lp->up, lp,
				 "Printer not online", 0);
			lp->mdmsigs &= ~TIOCM_CAR;
		} else
			lp->mdmsigs |= TIOCM_CAR;
    
		if (statrsp->flags & SC_ERR) {
			ctlogerr(STSERR_VERBOSE, lp->up, lp, 
				 "Error condition at printer", 0);
			lp->mdmsigs |= TIOCM_DSR;
		} else
			lp->mdmsigs &= ~TIOCM_DSR;
    
		/*
		 * See if anything changed state-- call out to OS dependent
		 * handler for further processing.
		 */
		trc(("ctf_handle_stat_chg: old=% new=%\n",
		     lp->old_mdmsigs, lp->mdmsigs));
	}
  
	/*
	 * Take care of RS232 stuff.
	 */
	else {
		/* update the state of all signals */
		if (statrsp->flags & SC_DCD) {		/* carrier */
			if (!(lp->mdmsigs & TIOCM_CAR)) {
				lp->mdmsigs |= TIOCM_CAR;
				wake_up_interruptible(&lp->open_wait);
			}
		} else if (lp->mdmsigs & TIOCM_CAR) {
			lp->mdmsigs &= ~TIOCM_CAR;
			ntrc(("DCD dropped, flags %\n", lp->flags));
			if (lp->flags & ASYNC_CHECK_CD) {
				if (!((lp->flags & ASYNC_CALLOUT_ACTIVE) &&
				      (lp->flags & ASYNC_CALLOUT_NOHUP))) {
					LINE_LOCK(lp->linelock, save);
					chg_tty = lp->tty;
					LINE_UNLOCK(lp->linelock, save);
					if (chg_tty)
						tty_hangup(chg_tty);
				}
			}
		}
    
		if (statrsp->flags & SC_DSR) 		/* data set ready */
			lp->mdmsigs |= TIOCM_DSR;
		else
			lp->mdmsigs &= ~TIOCM_DSR;
		
		if (statrsp->flags & SC_CTS) 		/* clear to send */
			lp->mdmsigs |= TIOCM_CTS;
		else
			lp->mdmsigs &= ~TIOCM_CTS;
    
		/*
		 * See if anything changed state-- call out to OS dependent
		 * handler for further processing.
		 */
		trc(("ctf_handle_stat_chg: old=% new=%\n",
		     lp->old_mdmsigs, lp->mdmsigs));
	}
  
	delta = lp->old_mdmsigs ^ lp->mdmsigs;
	if (TIOCM_CTS & delta)
		lp->icount.cts++;
	if (TIOCM_CAR & delta)
		lp->icount.dcd++;
	if (TIOCM_DSR & delta)
		lp->icount.dsr++;
	
	if (lp->wait_mask & delta)
		wake_up_interruptible(&lp->mdmsig_wait);
	
	/* respawn to keep status updates coming */
	
	ELS_LOCK(lp->up->faslock, save);
	statcmd = (req_stat_chg *) ctm_putfasclrcmd(lp->up);
	statcmd->opcode = FAS_STAT_CHG;
	statcmd->line = lp->port;
	statcmd->flags = lp->monitor_lines;
	ctf_logfasout(lp, (u_char *) statcmd);
	ELS_UNLOCK(lp->up->faslock, save);
}


#if defined(TRACER)
static char *fascmds[] = {
	"Global	 ",
	"Enable	 ",
	"Disable ",
	"Send	 ",
	"Recv	 ",
	"InTmrs	 ",
	"OutCtrl ",
	"InpCtrl ",
	"SetMdm	 ",
	"StatChg ",
	"SendBrk ",
	"SetParms",
	"FlowCtrl",
	"Reserve ",
	"Release ",
	"Reset	 ",
	"UnitStat",
	"WaitSnd ",
	"LineSt	 ",
	"LineInq ",
	"LPT_test",
	"UnitPoll",
	"UnitUnLk",
	"UnitInq ",
	"ExtParms"};
# define NF (sizeof(fascmds) / sizeof(char *))
#endif

/*
 * ctf_logfasout()
 */
LOCAL void ctf_logfasout (ct_line *lp, u_char *p)
{
	trc(("C~:$ ^\n", lp->up->unum, *p < NF ? fascmds[*p] : "????????", p));
	
	*lp->fas_histptr++ = *p++;
	*lp->fas_histptr++ = *p++;
	*lp->fas_histptr++ = *p++;
	*lp->fas_histptr++ = *p++;
	*lp->fas_histptr++ = *p++;
	*lp->fas_histptr++ = *p++;
	*lp->fas_histptr++ = *p++;
	*lp->fas_histptr++ = *p++;
	
	if (lp->fas_histptr == &lp->fas_hist[FASHISTSZ * FASCMDSZ])
		lp->fas_histptr = lp->fas_hist;
}


/*
 * ctf_logfasin()
 */
LOCAL void ctf_logfasin (ct_line *lp, u_char *p)
{
	trc(("R~:$ ^\n", lp->up->unum, *p < NF ? fascmds[*p] : "????????", p));

	*lp->fas_histptr++ = *p++ | 0x80;	/* code bit for incoming */
	*lp->fas_histptr++ = *p++;
	*lp->fas_histptr++ = *p++;
	*lp->fas_histptr++ = *p++;
	*lp->fas_histptr++ = *p++;
	*lp->fas_histptr++ = *p++;
	*lp->fas_histptr++ = *p++;
	*lp->fas_histptr++ = *p++;
  
	if (lp->fas_histptr == &lp->fas_hist[FASHISTSZ * FASCMDSZ])
		lp->fas_histptr = lp->fas_hist;
}


/*
 * ctf_showfascmd()
 *
 * Prints dump of FAS command.
 */
LOCAL void ctf_showfascmd (char *tag, u_char *p)
{
	int front, back;
	
	front = *p++;
	front <<= 8;
	front |= *p++;
	front <<= 8;
	front |= *p++;
	front <<= 8;
	front |= *p++;
	back = *p++;
	back <<= 8;
	back |= *p++;
	back <<= 8;
	back |= *p++;
	back <<= 8;
	back |= *p++;

	Printf("%s fascmd: %8x %8x\n", tag, front, back);
}


/*
 * ctf_display_statistics()
 *	
 * Just prints raw statistics dump to console.
 * Data is formatted (more or less) by the ELS unit.
 */
LOCAL void ctf_display_statistics (char *p, int cnt)
{
	char tbuf[41];
	int amt;
  
	while (cnt) {
		amt = min(cnt,40);
		strncpy(tbuf, p, amt);
		tbuf[amt] = 0;
		Printf("%s", tbuf);
		cnt -= amt;
		p += amt;
	}
}


/*
 * ctf_enable_fw()
 *	
 * Goes thru sequence to enable firmware.  If totally cold start or
 * recovering from strange error, we kick off with a FAS_RESET.
 * Otherwise, it's just a normal FAS_ENABLE followed by the initiation
 * of the FAS_STAT and FAS_RECV chains.
 *
 * Returns:
 *	OK (0)	things went well
 *	!OK !0	an error occured or a command timed out.
 */
LOCAL int ctf_enable_fw (ct_line *lp)
{
	ct_unit		*up = lp->up;
	req_reset	*reset;
	req_enable	*enab;
	req_stat_chg	*stat;
	req_recv	*recv;
	int port_type;
	int wait_time;
	int save;

	trc(("ctf_enable_fw\n"));

	/* see if we need to prep for parallel */
	port_type = ctf_port_type(up->uinfo, lp->port);
	if (port_type == PT_PARALLEL || port_type == PT_FAST_PARALLEL) {
		/* This is a parallel port */
		lp->line_state |= LINE_PRINTER;
	}
	
	/* see if FAS_RESET necessary */
	if (lp->line_state & (LINE_OFFLINE)) {
		lp->enab_disab_stat = OK;	
		
		/* make sure version of firmware/unit supports FAS_RESET */
		if (up->fw_features & FW_RESET) {
			trc(("ctf_enable_fw FAS_RESET\n"));
			ELS_LOCK(lp->up->faslock, save);
			lp->line_state &= ~LINE_DISABLING;
			reset = (req_reset *) ctm_putfasclrcmd(up);
			reset->opcode = FAS_RESET;
			reset->line = lp->port;
			ctf_logfasout(lp, (u_char *) reset);
			ELS_UNLOCK(lp->up->faslock, save);
			ctm_sendfas(up);
			
			/* Let's not spin the CPU here, instead use ct_delay 
			   Should this be UNINTERRUPTIBLE or not? -Kees */
			wait_time = jiffies + ELS_UNIT_WAIT * HZ;
			while ((lp->line_state & LINE_OFFLINE) && 
			       (jiffies < wait_time)) {
				ct_delay(HZ/10,TASK_INTERRUPTIBLE);
				/*
				schedule();
				set_current_state(TASK_RUNNING);
				*/
			}
			
			if (lp->line_state & LINE_OFFLINE) {
				ctf_offline(lp);
				return (-ENODEV);
			}
			
			/* check cmd status */
			if (lp->enab_disab_stat != OK) {
				switch(lp->enab_disab_stat)
				{
				case  ERR_BAD_CMD:	/* firmware is old */
					ctlogerr(STSERR_QUIET, up, lp,
						 "FAS_RESET error, time "
						 "to update firmware!", 0);
					lp->line_state &= ~LINE_OFFLINE;
					break;

				case  ERR_BUSY:
					/* unit already in use (other system)*/
					ctlogerr(STSERR_QUIET, up, lp,
						 "FAS_RESET failed - unit "
						 "is in use.", 0);
					ctf_offline(lp);
					return -EBUSY;
				default:
					trc(("unknown ERR value ~ calls "
					     "ctf_offline.\n"));
					ctf_offline(lp);
					return -ENODEV;
				}
			}
		}
		else {
			/* clear flags to state of freshly reset line */
			lp->line_state &= ~(LINE_ENABLED | LINE_DISABLING);
		}
	}

	/* do ENABLE */
	trc(("ctf_enable doing FAS_ENABLE.\n"));
	ELS_LOCK(lp->up->faslock, save);
	lp->line_state &= (LINE_DISABLING | LINE_ENABLED | LINE_RESET);
	enab = (req_enable *) ctm_putfasclrcmd(up);
	enab->opcode = FAS_ENABLE;
	enab->line = lp->port;
	ctf_logfasout(lp, (u_char *) enab);
	ELS_UNLOCK(lp->up->faslock, save);
	ctm_sendfas(up);

	wait_time = jiffies + ELS_UNIT_WAIT * HZ;
	while (!(lp->line_state & LINE_ENABLED) && (jiffies < wait_time)) {
		/* use ct_delay instead? -Kees
		schedule();
		set_current_state(TASK_RUNNING);
		*/
		ct_delay(HZ/10, TASK_INTERRUPTIBLE);
	}

	while (!(lp->line_state & LINE_ENABLED)) {
		ctlogerr(STSERR_QUIET, up, lp,
			 "FAS_ENABLE command timed out", 0);
		ctf_offline(lp);
		return (!OK);
	}

	/* check cmd status */
	if (lp->enab_disab_stat != OK) {
		ctlogerr(STSERR_QUIET, up, lp,
			 "error status %x on FAS_ENABLE cmd",
			 lp->enab_disab_stat);
		lp->line_state &= ~(LINE_ENABLED|LINE_DISABLING);
		ctf_offline(lp);
		return (lp->enab_disab_stat==ERR_BUSY) ? -EBUSY : -ENODEV;
	}

	update_print_max(lp);		/* R0.12 */
	update_curr_size_fassend(lp);	/* R0.12 */
	/* start status and recv chains */
	ELS_LOCK(lp->up->faslock, save);
	stat = (req_stat_chg *) ctm_putfasclrcmd(up);
	bzero((caddr_t) stat, sizeof(req_stat_chg));
	stat->opcode = FAS_STAT_CHG;
	stat->line = lp->port;

	if (lp->line_state & LINE_PRINTER) {
		/* handle printer (no input) */
		lp->monitor_lines = SC_ERR | SC_PE | SC_SEL;
		stat->flags = lp->monitor_lines | SC_IMMEDIATE;
		ctf_logfasout(lp, (u_char *) stat);
	} else {
		/* not a printer */
		lp->monitor_lines = SC_DCD;
		lp->mdmsigs = 0;
		stat->flags = lp->monitor_lines | SC_IMMEDIATE;
		ctf_logfasout(lp, (u_char *) stat);
		if (!(lp->line_state & LINE_IPENDING)) {
			recv = (req_recv *) ctm_putfasclrcmd(up);
			bzero((caddr_t) recv, sizeof(req_recv));
			recv->opcode = FAS_RECV;
			recv->line = lp->port;

			/*
			 * 64 is how many more we get (sometimes) than we
			 * asked for.  REV A changed payload to 384 bytes
			 * to prevent offline problem.
			 */
			recv->cnt_lo = (u_char) (384);
			recv->cnt_hi = (u_char) (384/256);
			lp->line_state |= LINE_IPENDING;
			ctf_logfasout(lp, (u_char *) recv);
		}
		else {
			printk("avoided extra FAS_RECV!\n");
			trc(("avoided extra FAS_RECV ~\n", lp->port));
		}
	}
	ELS_UNLOCK(lp->up->faslock, save);

	/* kick out status and recv commands */
	ctm_sendfas(up);
	return (OK);
}


/*
 * ctf_disable_fw()
 *	
 * Goes thru sequence to disable firmware.  If the line does not
 * have software carrier set, we implement a watchdog timers on the
 * disable.  If that expires, we try an output flush/resume, also
 * setting a timer.  If that also expires, we go to an unknown state
 * and indicate that a reset will be necessary.
 *
 * Watchdog timers are used depending on the values of local_timeout
 * and nonloc_timeout for the line.  If appropriate value is zero,
 * no timers are used and you're on your own.	Otherwise, the
 * appropriate value in seconds is used to watch the disable, and
 * appropriate values are used for any other commands.
 *
 * Returns:
 *	OK	things went well
 *	!OK	an error response was recv'd or a timeout occurred
 */
LOCAL int ctf_disable_fw (ct_line *lp)
{
	ct_unit *up = lp->up;
	req_disable *disab;
	int do_timer;
	int wait_time;
	int disable_timeout;
	int save;
  
	trc(("ctf_disable_fw\n"));
	if (!(lp->line_state & LINE_ENABLED))
		return (!OK);
  
	/* if dial-out or if non-local and no softcarrier, do watchdog timers*/
	if (lp->tty && NONLOCAL(lp))
		do_timer = disable_timeout = lp->nonloc_timeout;
	else {
		/*
		 * Normally, no watchdog timers are run on local lines, but
		 * if a non-zero value is provided for the local_timeout, then
		 * timers are run.  This value can be modified via the
		 * proprietary ioctl STSSETEXTCTL
		 */
		do_timer = disable_timeout = lp->local_timeout;
	}
	if (!do_timer) 
		do_timer = 3600;

	/* issue disable */
	ELS_LOCK(lp->up->faslock, save);
	disab = (req_disable *) ctm_putfasclrcmd(up);
	disab->opcode = FAS_DISABLE;
	disab->line = lp->port;
	lp->line_state |= LINE_DISABLING;
	ctf_logfasout(lp, (u_char *) disab);
	ELS_UNLOCK(lp->up->faslock, save);
	ctm_sendfas(up);
  
	wait_time = jiffies + do_timer * HZ;
	while ((lp->line_state & LINE_ENABLED) && (jiffies < wait_time)) {
		schedule();
		set_current_state(TASK_RUNNING);
		if (signal_pending(current))
			break;
	}

	if (lp->line_state & LINE_ENABLED) {
		/* try an output flush/resume to unclog things */
		ctf_flush_output(lp, TRUE);
    
		wait_time = jiffies + 10 * HZ;
		trc(("% trying a flush_output to unclog after "
		     "waiting % sec\n", jiffies,wait_time));
		while ((lp->line_state & LINE_ENABLED) &&
		       jiffies < wait_time) {
			schedule();
			set_current_state(TASK_RUNNING);
		}

		if (lp->line_state & LINE_ENABLED) {
		#if 0	/* no big deal */
			ctlogerr(STSERR_QUIET, up, lp,
				 "FAS_DISABLE timed out", 0);
		#endif
      
		/* mark as needing reset again */
		lp->line_state &= ~(LINE_ENABLED|LINE_DISABLING);
		lp->line_state |= LINE_OFFLINE;

		#if 0	/* no big deal */
		printz("disable_fw calling ctf_offline\n");
		ctf_offline(lp);
		#endif

		return (!OK);
		}
	}

	/* disable completed, check status */
	if(lp->enab_disab_stat != OK) {
		ctlogerr(STSERR_QUIET, up, lp,
			 "error status %x on FAS_DISABLE cmd",
			 lp->enab_disab_stat);
		/* mark as needing reset again */
		lp->line_state &= ~(LINE_ENABLED|LINE_DISABLING);
		printz("disable_fw3 calling ctf_offline\n");
		ctf_offline(lp);
		return (!OK);
	}
  
	return (OK);
}


/*
 * ctf_flush_output()
 *
 * Called after carrier loss on non-local line to dump output.
 * Sends FAS_OUTPUT_CTRL command to do flush/resume.
 * Also clears the local write buffer.
 *
 *  flag:	TRUE  - truely do flush
 *      	FALSE - don't do flush
 */
LOCAL void ctf_flush_output (ct_line *lp, int flag)
{
	req_output_ctrl	*occmd;
	int		 save;

  
	/* do flush at unit */
	ELS_LOCK(lp->up->faslock, save);
	ctm_reset_circular_queue(&lp->xmit_q);
	if (lp->line_state & (LINE_OFFLINE|LINE_DISABLING)) {
		ELS_UNLOCK(lp->up->faslock, save);
		return;			/* R0.12 */
	}
	occmd = (req_output_ctrl *) ctm_putfasclrcmd(lp->up);
	occmd->opcode = FAS_OUTPUT_CTRL;
	occmd->line = lp->port;
	occmd->flush = flag;
	occmd->suspend = FALSE;
	occmd->resume = TRUE;
	ctf_logfasout(lp, (u_char *) occmd);
	ELS_UNLOCK(lp->up->faslock, save);
	ctm_sendfas(lp->up);	/* R0.012 start me up */
}


/*
 * ctf_suspend_input()
 *
 */
LOCAL void ctf_suspend_input (ct_line *lp)
{
	req_input_ctrl *cmd;
	int save;
  
	trc(("ctf_suspend_input\n"));
	if (lp->line_state & (LINE_OFFLINE | LINE_DISABLING))
		return;			/* R0.12 */
  
	/* do suspend portside */
	ELS_LOCK(lp->up->faslock, save);
	cmd = (req_input_ctrl *) ctm_putfasclrcmd(lp->up);
	cmd->opcode = FAS_INPUT_CTRL;
	cmd->line = lp->port;
	cmd->flush = FALSE;
	cmd->suspend = TRUE;
	cmd->resume = FALSE;
	ctf_logfasout(lp, (u_char *) cmd);
	ELS_UNLOCK(lp->up->faslock, save);
	ctm_sendfas(lp->up);
}


/*
 * ctf_suspend_output()
 *
 */
LOCAL void ctf_suspend_output (ct_line *lp)
{
	req_output_ctrl *cmd;
	int save;
  
	trc(("ctf_suspend_output\n"));
  
	if (lp->line_state & (LINE_OFFLINE | LINE_DISABLING))
		return;			/* R0.12 */
	/* do suspend portside */
	ELS_LOCK(lp->up->faslock, save);
	cmd = (req_output_ctrl *) ctm_putfasclrcmd(lp->up);
	cmd->opcode = FAS_OUTPUT_CTRL;
	cmd->line = lp->port;
	cmd->flush = FALSE;
	cmd->suspend = TRUE;
	cmd->resume = FALSE;
	ctf_logfasout(lp, (u_char *) cmd);
	ELS_UNLOCK(lp->up->faslock, save);
	ctm_sendfas(lp->up);
}


/*
 * ctf_resume_output()
 *
 */
LOCAL void ctf_resume_output (ct_line *lp)
{
	req_output_ctrl *cmd;
	int save;
  
	trc(("ctf_resume_output\n"));
  
	if (lp->line_state & (LINE_OFFLINE | LINE_DISABLING))
		return;			/* R0.12 */
	/* do resume portside */
	ELS_LOCK(lp->up->faslock, save);
	cmd = (req_output_ctrl *) ctm_putfasclrcmd(lp->up);
	cmd->opcode = FAS_OUTPUT_CTRL;
	cmd->line = lp->port;
	cmd->flush = FALSE;
	cmd->suspend = FALSE;
	cmd->resume = TRUE;
	ctf_logfasout(lp, (u_char *) cmd);
	ELS_UNLOCK(lp->up->faslock, save);
	ctm_sendfas(lp->up);
}


/*
 * ctf_resume_input()
 *
 */
LOCAL void ctf_resume_input (ct_line *lp)
{
	req_output_ctrl *cmd;
	int save;
  
	trc(("ctf_resume_input\n"));
  
	if (lp->line_state & (LINE_OFFLINE | LINE_DISABLING))
		return;			/* R0.12 */
	/* do resume portside */
	ELS_LOCK(lp->up->faslock, save);
	cmd = (req_input_ctrl *) ctm_putfasclrcmd(lp->up);
	cmd->opcode = FAS_INPUT_CTRL;
	cmd->line = lp->port;
	cmd->flush = FALSE;
	cmd->suspend = FALSE;
	cmd->resume = TRUE;
	ctf_logfasout(lp, (u_char *) cmd);
	ELS_UNLOCK(lp->up->faslock, save);
	ctm_sendfas(lp->up);
}


/*
 * ctf_smdm()
 * 
 * Set modem signals.  Passed bits & cmd in termio format (the only modem
 * currency exchanged in this driver).
 *
 */
LOCAL void ctf_smdm (ct_line *lp, int bits, int cmd, int block)
{
	req_set_modem	*mdmcmd;
	int save;
	u_int oldsigs;
	u_char dtr_ctrl, rts_ctrl;
  
	trc(("ctf_smdm\n"));
  
	if (lp->line_state & (LINE_OFFLINE | LINE_DISABLING))
		return;			/* R0.12 */

	oldsigs = lp->mdmsigs;
	dtr_ctrl = rts_ctrl = MC_NO_CHG;
	switch (cmd) {
	case TIOCMSET:		/* set all */
		dtr_ctrl = (bits & TIOCM_DTR) ? MC_ASSERT : MC_DEASSERT;
		rts_ctrl = (bits & TIOCM_RTS) ? MC_ASSERT : MC_DEASSERT;
		lp->mdmsigs &= ~(TIOCM_DTR | TIOCM_RTS);
		lp->mdmsigs |= (bits & (TIOCM_DTR | TIOCM_RTS));
		break;
		
	case TIOCMBIS:		/* set individual */
		if (bits & TIOCM_DTR) {
			dtr_ctrl = MC_ASSERT;
			lp->mdmsigs |= TIOCM_DTR;
		}
		if (bits & TIOCM_RTS) {
			rts_ctrl = MC_ASSERT;
			lp->mdmsigs |= TIOCM_RTS;
		}
		break;
    
	case TIOCMBIC:		/* clear individual */
		if (bits & TIOCM_DTR) {
			dtr_ctrl = MC_DEASSERT;
			lp->mdmsigs &= ~TIOCM_DTR;
		}
		if (bits & TIOCM_RTS) {
			rts_ctrl = MC_DEASSERT;
			lp->mdmsigs &= ~TIOCM_RTS;
		}
		break;
    
	default:
		return;
	}
  
	/* make sure we're really requesting a change */
	if (oldsigs == lp->mdmsigs)
		return;
  
	/*
	 * See if we need modem sigs to persist after disables.
	 * Driver sets FW_MDM_PERSIST if firmware supports it AND the
	 * driver needs it.
	 */
	if (lp->up->fw_features & FW_MDM_PERSIST) {
		rts_ctrl |= MC_PERSIST;
		dtr_ctrl |= MC_PERSIST;
	}
  
	/* construct the command */
	ELS_LOCK(lp->up->faslock, save);
	mdmcmd = (req_set_modem *) ctm_putfasclrcmd(lp->up);
	mdmcmd->opcode = FAS_SET_MODEM;
	mdmcmd->line = lp->port;
	mdmcmd->rts_ctrl = rts_ctrl;
	mdmcmd->dtr_ctrl = dtr_ctrl;
	ctf_logfasout(lp, (u_char *) mdmcmd);
	ELS_UNLOCK(lp->up->faslock, save);
  
	ctm_sendfas(lp->up);
	lp->line_state |= LINE_SET_MODEM;
  
	if (block)
		ctf_block_ioctl(lp, LINE_SET_MODEM, 60);
}


/*
 * Table of Unix baud rates to FAS equivalents
 */
LOCAL int baudexttab[] = {
	FASNOBAUD,	/* 0x00 - 0 */
	FAS50,		/* 0x01 - 50 */
	FAS75,		/* 0x02 - 75 */
	FAS110,		/* 0x03 - 110 */
	FAS134,		/* 0x04 - 134.5 */
	FAS150,		/* 0x05 - 150 */
	FAS200,		/* 0x06 - 200 */
	FAS300,		/* 0x07 - 300 */
	FAS600,		/* 0x08 - 600 */
	FAS1200,	/* 0x09 - 1200 */
	FAS1800,	/* 0x0a - 1800 */
	FAS2400,	/* 0x0b - 2400 */
	FAS4800,	/* 0x0c - 4800 */
	FAS9600,	/* 0x0d - 9600 */
	FAS19200,	/* 0x0e - 19200 */
	FAS38400,	/* 0x0f - 38400 */
	FAS57600,	/* 0x10 - 57600 */
	FAS115200,	/* 0x11 - 115200 */
	FAS230400,	/* 0x12 - 230400 */
	FAS460800	/* 0x13 - 460800 */
#ifdef SPARCBAUDS
	,FAS76800,	/* 0x14 - 76800 SPARC only*/
	FAS153600,	/* 0x15 - 153600 SPARC only */
	FAS307200	/* 0x16 - 307200 SPARC only */
#endif
};
#ifdef SPARCBAUDS
#define MAXBAUDEXTN	0x16
#else
#define MAXBAUDEXTN	0x13
#endif


/*
 * Table of FAS to integer equivalents. 0 is used for undefined
 * baud rates. 134.5 is rounded up since this is used for rate
 * calculations.
 */
LOCAL int fastointtab[] = {
	75, 110, 150, 300, 600, 1200, 1800, 2000, 2400,
	4800, 9600, 19200, 38400, 57600, 76800, 115200,
	50, 100, 135, 200, 900, 3600, 7200, 0,  153600,
	0,0,0,0,0,0,0,0,0,
	230400, 
	0,0,0,0,0,0,
	307200,
	0,0,0,0,0,
	460800 };
#define MAXFASBAUDEXTN	47


/*
 * ctf_int_baud
 *
 * Lookup integer baudrate based on fasbaud code as used in
 * FAS_EXT_PARAMS (or FAS_SET_PARAMS).
 * An integer is returned. Return 0 if undefined.
 */
LOCAL int ctf_int_baud (unsigned bcode)
{
	if (bcode > MAXFASBAUDEXTN) {
		trc(("undefined baud rate %.\n", bcode));
		return(0);
	}
	return(fastointtab[bcode]);
}


/*
 * ELS units don't support all defined baud rates, and different types
 * of EL-* suport different sets of rates. This routine returns -EINVAL
 * if the baud rate is not supported by the unit, 
 * or 0 (OK) if the baud rate is supported. 
 */
LOCAL int illegal_baud (ct_line *lp, struct termios *tiosp)
{
	int baudcode;
	ct_unit *up;

	/* if the termios ptr is null or the output baud rate is 0, return OK*/
	if (!tiosp || ((tiosp->c_cflag & (CBAUD^CBAUDEX)) == B0))
		return OK;
	baudcode = ctf_baud(lp, tiosp);
	if (baudcode == FASNOBAUD) {
		trc(("Invalid baudrate is FASNOBAUD %\n", baudcode));
		return (-EINVAL);
	}
	up = lp->up;
	if (ctf_check_baud(up->uinfo, lp->port, baudcode) != OK) {
		trc(("Invalid baudrate %\n", baudcode));
		return (-EINVAL);
	}
	return OK;
}


/*
 * ctf_baud() We handle three different ways to expand baud rates:
 * 1. extended rates in c_cflag when CBAUDEX is set. (preferred)
 * 2. cdstty extensions for Digi customers still using our old method.
 * 3. setserial (ASYNC_SPD_MASK) which replaces 38400 with a higher rate.
 * The methods are mutually exclusive, with the lowest numbered method
 * winning if more than one is selected together. Methods 2 & 3 are
 * deprecated.
 *
 * The two platforms have different baud rate tables above B460800.
 * The ELS hardware can support a few oddball rates, so we will go that
 * far for SPARC only (adding B76800, B153600 & B307200 which are
 * above 460800 in B-ness). 
 */
LOCAL int ctf_baud (ct_line *lp, struct termios *tiosp)
{
	int index;
     
#ifdef SPARCBAUDS 
# define MAXINDEX (B307200 & CBAUDEX)
#else
# define MAXINDEX (B460800 & CBAUDEX)
#endif

	if (!tiosp)
		return FASNOBAUD;
	index = tiosp->c_cflag & (CBAUD);
	trc(("ctf_baud index=~\n", index));
	if (index & CBAUDEX) {
		index &= ~CBAUDEX;
		ttrc(("ext index=%\n", index));
		if ((index < 1) || (index > MAXINDEX))
			tiosp->c_cflag &= ~CBAUDEX;
		else
			index += 15;
	}
	else if ((index == B38400) && !lp->cdbaud_ext) {
		if ((lp->flags & ASYNC_SPD_MASK) == ASYNC_SPD_HI)
			index++;
		else if ((lp->flags & ASYNC_SPD_MASK) == ASYNC_SPD_VHI)
			index += 2;
		else if ((lp->flags & ASYNC_SPD_MASK) == ASYNC_SPD_SHI)
			index += 3;
		else if ((lp->flags & ASYNC_SPD_MASK) == ASYNC_SPD_WARP)
			index += 4;
	} else
		index += lp->cdbaud_ext;
	trc(("ctf_baud %, index is %, cdbaudext=%d\n",
	     tiosp->c_cflag & CBAUD,index,lp->cdbaud_ext));
	if (index < 0 || index > MAXBAUDEXTN) {
		trc(("FASNOBAUD!\n"));
		return (FASNOBAUD);
	}
	else {
		trc(("baudexttab[index.brate=%\n",baudexttab[index]));
		return (baudexttab[index]);
	}
}


/*
 * ctm_flush_ioctl(int arg)
 *	
 * Forcibly flushes a spec'd device (device AND unit)
 */
LOCAL int ctm_flush_ioctl( dev_t arg )
{
	int i, mod, pun, port, major, minor;
	ct_unit		*up;
	ct_line		*lp;

	major = MAJOR(arg);
	minor = MINOR(arg);

	if (minor == ELS_GLOBAL_MINOR)
		return EPERM;		/* can't reset diag unit! */
	pun = DEV2UNIT(arg);
	if (pun < 0 || pun >= MAX_ELS_DEVS)
		return (-ENODEV);
	
	for (i = 0, mod= -1; i < MAX_MODULE; i++) {
		if (els_serial_driver[i].major == major)
		{ mod = i ; break; }
		if (els_callout_driver[i].major == major)
		{mod = i ; break; }
	}
	if (mod < 0)
		return ENODEV;

	up = els_units[mod][pun];
	if (up == NULL) {
		trc(("ctm_flush_ioctl(%) - null unitptr\n", arg));
		return(ENODEV);
	}
	port = DEV2LINE(arg);
	if (port >= up->num_lines) {
		trc(("ctm_flush_ioctl(%) - port inval\n", arg));
		return(EINVAL);
	}
	lp = up->ct_lines[port];
	if (lp == NULL) {
		trc(("ctm_flush_ioctl(%) - null lptr\n", arg));
		return(ENOENT);
	}
	
	ntrc(("ctm_flush_ioctl(%)\n", arg ));

	/* drain any further output */

	/* make sure line is enabled */
	if (lp->line_state & LINE_ENABLED) {
    
		ctf_clear_flow_control(lp);
		/* if carrier not active on serial port, fake going active */
		if (!(lp->line_state&LINE_PRINTER) && 
		    !(lp->mdmsigs & TIOCM_CAR)) {
			lp->mdmsigs |= TIOCM_CAR;
			wake_up_interruptible(&lp->open_wait);
			ct_delay(HZ / 10, TASK_INTERRUPTIBLE);
		}
    
		/* now do a carrier drop */
		if (!(lp->line_state&LINE_PRINTER)) {
			lp->mdmsigs &= ~TIOCM_CAR;
			wake_up_interruptible(&lp->open_wait);
		}
    
		/* always send flush to firmware */
		ctf_flush_output(lp, TRUE);
	}
	else
		return(ENOTTY);
  
	return(0);
}


/*
 * ctm_roundtrip_adjust()
 * This function is called from ctf_procrecv() when a FAS_UNIT_POLL respond is 
 * received.  This function does the math to calculate new average RTT and 
 * the smoothed mean deviation of the RTT, which is used to dynamically set
 * the buffer size of the FAS_SEND for the lines on the unit. 
 */
LOCAL void ctm_roundtrip_adjust (ct_unit *up)
{
	int dtime;        /* delta time AKA 'M' in Emo's whitepaper */
	int e;            /* difference between observed RTT (dtime) and
			   * running average RTT (ctu_avg_rtt) */
	int old_rtt = up->avg_rtt;

	/*
	 * if we supported SCSI units, we'd return here since rtt is very small
	 * on the SCSI bus (if nobody is doing a tape backup).
	 */

	/*
	 * STS_GAIN_FACTOR is the number of bits to right shift the
	 * gain/loss (the e variable).  We do this to change the weight of
	 * the gain/loss when calculating the new average, such that
	 * recent measurements have some weight over the overall average,
	 * but also ensuring that temporary fluctuations don't throw our
	 * averages completely off kilter.
	 *
	 * See discussion about "Round-Trip Time Measurement" in section
	 * 21.3 of Stevens' _TCP/IP Illustrated_ pp. 299-300 for details
	 * about using this smoothing factor.
	 */
#define STS_GAIN_FACTOR 2

	if (jiffies < up->poll_time) {
		up->poll_time = 0;
		trc(("rtt: jiffies wraparound\n"));
		/*
		 * For jiffies wraparound, we'll simplify and won't recalculate
		 * packet size.  It won't kill us to miss one.
		 */
		return;
	}

	dtime = jiffies - up->poll_time;    /* elapsed time */
	
	e = dtime - up->avg_rtt;
	up->avg_rtt = up->avg_rtt + (e >> STS_GAIN_FACTOR);
	
	e = (e < 0) ? -e : e;    /* absolute value of e */

	up->avg_rtt_dev = up->avg_rtt_dev +
		((e - up->avg_rtt_dev) >> STS_GAIN_FACTOR);

#if 0
	trc(("rtt adjust: dtime=%, avg_rtt=%, smdev=% HZ=%d\n",
	     dtime,up->avg_rtt, up->avg_rtt_dev, HZ));
	printk("rtt adjust: dtime=%d, av_rtt=%d, smdev=%d old_rtt=%d\n",
	       dtime,up->avg_rtt, up->avg_rtt_dev, old_rtt);
#endif
  /*
   * clear poll_time since it makes for a handy indication to
   * let us know there are no FAS_UNIT_POLLs outstanding.
   */

	up->poll_time = 0;

	/* Now adjust the packet size for each line, if it changed. */

	if (up->avg_rtt != old_rtt) {
		ctm_adjust_lines(up);
#if 0
		printk("new avg_rtt els%2.2x=%d\n", up->unum, up->avg_rtt);
#endif
	}
	return;
}


/*
 * ctm_adjust_lines()
 *
 * Go through each open line on the unit, adjusting the packet
 * size if necessary.
 */
LOCAL void ctm_adjust_lines (ct_unit *up)
{
	ct_line *lp;
	int port;

	for (port = 0; port < (int)up->num_lines; port++) {
		lp = up->ct_lines[port];
		if (NULL == lp)
			continue;

		if (lp->line_state & LINE_PRINTER)
			continue;
		if ((lp->line_state & (LINE_OFFLINE | LINE_ENABLED |
				       LINE_DISABLING)) 
		    != LINE_ENABLED) 
			continue;
    
		update_curr_size_fassend(lp);
	}
}


/*
 * Need to figure out a fair chunk this line can use per pkt for
 * outgoing data. We don't want to buffer huge amounts of stuff out
 * at the unit or we lose time sync between host and terminal for human
 * users. For folks, we want to limit to about 1/10 second worth of 
 * output at a time so a human is aware of what is actually running.
 * 
 * Printers on the other hand, work best when sent as much data as
 * possible.
 *
 * WANs make life interesting by having round trip transit times over
 * 100 milliseconds. We allow larger packets based on transit time.
 */
LOCAL void update_curr_size_fassend (ct_line *lp)
{
	int new_curr;

	if (lp->line_state & LINE_PRINTER) {
		lp->curr_size_fassend = lp->print_max;
		return;
	}

	/* non-printer ports */
	/* Figure out a good maximum number of characters to put in each FAS
	 * packet. Technique developed by Emilio Millan, inspired by Stevens
	 * TCP/IP Illustrated TCP Timeout and re-transmission. The timer
	 * routine periodically sends a null command to the unit and marks
	 * the time in jiffies. When the response arrives back, it's jiffies
	 * are use to figure the RTT (Round Trip Time). This value, averaged,
	 * and the smoothed deviation of the value are stored per unit, and
	 * used to calculate the optimal outgoing message size for good
	 * performance.  Assume 10 bits per character.
	 */
	new_curr = (lp->up->avg_rtt + lp->up->avg_rtt_dev) * lp->int_obaud;
	new_curr /= (10*HZ);		/* to get bytes/roundtrip */

	/* Now we have our baseline. */

	/* No less than 1/10 second worth of data. */
	new_curr = max(new_curr, lp->int_obaud/100);
	/* No fewer than 16 bytes. */
	new_curr = max(new_curr, 16);
	/* No more than the unit can handle. */
	lp->curr_size_fassend = min(new_curr, lp->up->max_size_fassend-2);
}


/*
 * Set the max print packet size for this line,  based on the line properties.
 * If line isn't being used as a printer, set the print_max to 0. 
 * This needs to be called on port setup or change. This code would expand
 * greatly if we supported SCSI units.
 */
LOCAL void update_print_max(ct_line *lp)
{
	if (lp->line_state & LINE_PRINTER)
		lp->print_max = (lp->up->max_opkt * 3) / 4;
	else
		lp->print_max = 0;
}


/*
 * /proc fs routines....
 */
LOCAL inline int line_info(char *buf, ct_line *lp)
{
	char	stat_buf[30];
	int	ret;
  
	ret = sprintf(buf, "%d: %sline, flags %x",lp->port,
		      ((lp->line_state & LINE_ENABLED) ? "on" : "off"),
		      lp->line_state);
	stat_buf[0] = 0;
	stat_buf[1] = 0;
	if (lp->mdmsigs & TIOCM_RTS)
		strcat(stat_buf, "|RTS");
	if (lp->mdmsigs & TIOCM_CTS)
		strcat(stat_buf, "|CTS");
	if (lp->mdmsigs & TIOCM_DTR)
		strcat(stat_buf, "|DTR");
	if (lp->mdmsigs & TIOCM_DSR)
		strcat(stat_buf, "|DSR");
	if (lp->mdmsigs & TIOCM_CAR)
		strcat(stat_buf, "|CD");
	/* ELS units do not monitor RI */
	ret += sprintf(buf+ret, " baud:%d",lp->int_obaud);

	ret += sprintf(buf+ret, " tx:%d rx:%d",
		       lp->icount.tx, lp->icount.rx);

	if (lp->icount.frame)
		ret += sprintf(buf+ret, " fe:%d", lp->icount.frame);
  
	if (lp->icount.parity)
		ret += sprintf(buf+ret, " pe:%d", lp->icount.parity);
  
	if (lp->icount.brk)
		ret += sprintf(buf+ret, " brk:%d", lp->icount.brk);	
  
	if (lp->icount.overrun)
		ret += sprintf(buf+ret, " oe:%d", lp->icount.overrun);

	/*
	 * Last thing is the RS-232 status lines
	 */
	ret += sprintf(buf+ret, " %s\n", stat_buf+1);
	return ret;
}



int els_read_proc(char *page, char **start, off_t off, int count,
		  int *eof, void *data)
{
	int  i, mod, port, len = 0, l;
	off_t	begin = 0;
	ct_unit *up;
	ct_line *lp;
	len += sprintf(page, "elsinfo:1.0 driver:%s\n", ELS_NAME);
  
	for (mod = 0; mod < MAX_MODULE && len < 4000; mod++) {
		for (i = 0; i < MAX_ELS_DEVS && len < 4000; i++) {
			if (!(up = els_units[mod][i])) {
				continue;
			}
			for (port = 0; port < up->num_lines && 
				     len < 4000; port++) {
				if (!(lp = up->ct_lines[port]))
					continue;
				l = line_info(page + len, lp);
				len += l;
				if (len+begin > off+count)
					goto done;
				if (len+begin < off) {
					begin += len;
					len = 0;
				}
			}
		}
	}
	*eof = 1;
 done:
	if (off >= len+begin)
		return 0;
	*start = page + (off-begin);
	return ((count < begin+len-off) ? count : begin+len-off);
}


/* end of els.c */

