/*
 * TRACER.C - memory tracing module for Unix debug.
 *
 *****************************************************************************
 *
 * Copyright 1999-2000 Digi International (www.digi.com)
 *    Jeff Randall <Jeff_Randall@digi.com>
 *    Zhong Deng <Zhong_Deng@digi.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY, EXPRESS OR IMPLIED; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****************************************************************************
 *
 *
 * Tracer call recognizes the following single char format specs:
 *	tilde (~)		print %02x
 *	percent (%)		print %08lx full integer
 *	dollarsign ($)		print string
 *	carat (^)		dump 8 byte FAS command
 */


#ifndef TRCSZ 
#define	TRCSZ	0x0010000	/* 64 Kbytes */
#endif

#define TRC_MAGIC	0x54726321

typedef struct tracebuf {
	int		 magic;
	char		 buf[TRCSZ];
	unsigned char	 initd;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
	spinlock_t	 lock;
#endif
	char		 *fill;
	char		 *end;
} tracebuf;

tracebuf trcbuf;

static char		*hex = "0123456789abcdef";


#if !defined(TRACER_CONSOLE)

#define ptrc(tb,c)					\
{							\
	*tb->fill++ = (c);				\
	if(tb->fill == tb->end) tb->fill = tb->buf;	\
}

#else

#define ptrc(tb,c)					\
{							\
	*tb->fill++ = (c);				\
	if(tb->fill == tb->end) tb->fill = tb->buf;	\
}

#endif

#define pnumber(tb,n)					\
{							\
	ptrc(tb,hex[((int)(n) >> 4) & 0xf]);		\
	ptrc(tb,hex[((int)(n) & 0xf)]);			\
}

/*VARARGS1*/
void
tracer(s,a,b,c,d,e,f)
	char *s;
	int a,b,c,d,e,f;
{
	int		 argindx = 0;
	int		 x,save;
	unsigned char	*p;
	tracebuf	*tb;


	tb = &trcbuf;

	if(!tb->initd) {
		Printf("trcbuf at %x\n", (int)tb);
		tb->magic = TRC_MAGIC;
		tb->fill = tb->buf;
		tb->end = &tb->buf[TRCSZ];
		tb->initd = 1;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
		spin_lock_init(&tb->lock);
#endif
	}
	/* do trace */

	ELS_LOCK(tb->lock, save);

	while(*s) {
		if(*s == '~' || *s == '%' || *s == '^' || *s == '$') {
			switch(argindx) {
			case 0:
				x = a;
				break;
			case 1:
				x = b;
				break;
			case 2:
				x = c;
				break;
			case 3:
				x = d;
				break;
			case 4:
				x = e;
				break;
			case 5:
				x = f;
				break;
			default:
				ELS_UNLOCK(tb->lock, save);
				return;
			}
			argindx++;	
			/* do hex byte */
			if(*s == '~') {
				pnumber(tb,x);
			}
			/* do hex long word */
			else if(*s == '%') {
				ptrc(tb,hex[(x >> 28) & 0xf]);
				ptrc(tb,hex[(x >> 24) & 0xf]);
				ptrc(tb,hex[(x >> 20) & 0xf]);
				ptrc(tb,hex[(x >> 16) & 0xf]);
				ptrc(tb,hex[(x >> 12) & 0xf]);
				ptrc(tb,hex[(x >> 8) & 0xf]);
				ptrc(tb,hex[(x >> 4) & 0xf]);
				ptrc(tb,hex[x & 0xf]);
			}
			/* do string */
			else if(*s == '$') {
				p = (unsigned char *)x;
				while(*p) {
					ptrc(tb,*p);
					p++;
				}
			}
			/* display 8 byte FAS cmd (just 8 byte mem dump) */
			else {
				p = (unsigned char *)x;
				pnumber(tb,*p); p++;
				pnumber(tb,*p); p++;
				pnumber(tb,*p); p++;
				pnumber(tb,*p); p++; ptrc(tb,' ');
				pnumber(tb,*p); p++;
				pnumber(tb,*p); p++;
				pnumber(tb,*p); p++;
				pnumber(tb,*p); p++;
			}
		}
		else
			ptrc(tb,*s);

		s++;
	}

	*tb->fill = '~';		/* tilde marks end of trace entry */
	ELS_UNLOCK(tb->lock, save);
}
